// # 1 "algappr.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 14:57:47 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "algappr.mtx" 2
// $Log$


// Utiliser RngSerElt >> TODO

intrinsic PadeHermite_BeLa_fast(F,sigma,n) -> .
{
  F is a sequence of power series, sigma is the precision,
  n is the vector for degrees of the cofactors.
  It implements Beckmann & Labahn's fast algorithm. The output is the basis
  of the free module solution.
}
  S<t>:= Universe(F);
  K:=BaseRing(S);
  UP<T>:= PolynomialRing(K);
  m:=#F;
  
  // Init
  d:=n;
  P:=[ Vector(UP, [UP| j eq k select 1 else 0 : j in [1..m]]): k in [1..m] ];
  
  // main loop
  for i:=0 to sigma-1 do
    max_degree:= Max([ Max([Degree(P[l][k]): k in [1..m]]) : l in [1..m]]);
    tmpF:=[S | z+O(t^(i+1)) : z in F];
    tmpF:=[S | z-Truncate(z+O(t^(i-max_degree-1))) : z in tmpF];
    tmpR:=[&+[ S | (S!P[l][k])*tmpF[k] : k in [1..m]] : l in [1..m]];
    delete tmpF;
//    for l:=1 to m do
//      assert Valuation(tmpR[l]) ge i;
//    end for;
    c:= [Coefficient(z,i) : z in tmpR];
    delete tmpR;
    lambda:=[l : l in [1..m] | c[l] ne 0];
    
    if IsEmpty(lambda) then
      continue;
    end if;
    
    max_d:=d[lambda[1]];
    for l in lambda do
      if d[l] ge max_d then pi:=l; end if;
    end for;
    
    for l in lambda do
      if l eq pi then continue; end if;
      P[l]-:= c[l]/c[pi] * P[pi];
    end for;
    
    P[pi] *:= T;
    d[pi] -:=1;
  end for;
  return [ [UP|  P[i][j] : j in [1..m]]: i in [1..m]],d;

end intrinsic;

intrinsic PadeHermite_BeLa_superfast(F,sigma,n)->.,.
{
  F is a sequence of power series, sigma is the precision,
  n is the vector for degrees of the cofactors.
  It implements Beckmann & Labahn's super fast algorithm.
  The output is the basis of the free module solution.
}
  S<t>:= Universe(F);
  K:=BaseRing(S);
  UP<T>:= PolynomialRing(K);
  m:=#F;
  vprint AlgebraicApproximantVerbose,2: "entering at precision " cat Sprint(sigma);
  if sigma le 160 then return PadeHermite_BeLa_fast(F,sigma,n); end if;
    
  rho:=sigma div 2;
  sigma1:=sigma-rho;
  tmpF:=[S | z+O(t^(rho)) : z in F];
  P1,d1:= PadeHermite_BeLa_fast(tmpF,rho,n);
  delete tmpF;
  max_degree:= Max([ Max([Degree(P1[l][k]): k in [1..m]]) : l in [1..m]]);
  tmpF:=[S | z+O(t^sigma) : z in F];
  tmpF:=[S | z-Truncate(z+O(t^(rho-max_degree-1))) : z in tmpF];
  tmpR:=[&+[ S | (S!P1[l][k])*tmpF[k] : k in [1..m]] : l in [1..m]];

  F1:= [ S![Coefficient(z,j) : j in [rho..sigma-1]]: z in tmpR ];
  delete tmpR;
  P2,d2:=PadeHermite_BeLa_fast(F1,sigma1,d1);
  P3:= [ [ &+[P2[k][i]*P1[i][j] : i in [1..m]]  : j in [1..m] ] : k in [1..m]];
  d3:=d2;
  return P3,d3;

end intrinsic;

////////// Pade approximant
// TODO << improve
intrinsic PadeApproximant(s::RngSerElt,order::RngIntElt,n::RngIntElt,d::RngIntElt)->.
{
  s is seen as a series at order order.
  It returns a boolean b and, if b is true, a
  rational function with a numerator of delfee at most n
  and denominator at most d. 
}
// It is based on the naive extended euclidian algorithm.

  require n ge 0 and d ge 0: "n and d must be positive";

  PS:=Parent(s);
  F:=BaseRing(Parent(s));
  UP:=PolynomialAlgebra(F);
  RF:=FieldOfFractions(UP);

  if s eq 0 then return RF!0; end if;

  if n+d gt order then return RF!0; end if;

  P:=UP![Coefficient(s,i): i in [0..order-1]];

  if Degree(P) le 0 then return RF!P; end if;
 
  M:=UP.1^(order);
 
  u0:=M;
  v0:=UP!0;
  u1:=P;
  v1:=UP!1;
 
  while Degree(v1) le d do
    if Degree(u1) le n then break; end if;
    q,r:=Quotrem(u0,u1);
    u2:=u0-q*u1;
    v2:=v0-q*v1;
    u0:=u1;
    v0:=v1;
    u1:=u2;
    v1:=v2;
  end while;
  return RF!u1/RF!v1;

end intrinsic;

