// # 1 "bbp.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun 22 16:32:54 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "bbp.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "bbp.mtx" 2

//#define DEBUG

/*
   E<x,y>:= BlackboxPolynomialAlgebra(Rationals(),2);
   x:=Var(x); // Variable constructor
   y:=Var(y);

   or

   E,x,y:= BlackboxPolynomialAlgebra(Rationals(),2);
   a:=(x+y)^100;
   PrintBbp(a);
   PrintBbp(Derivative(a,1));
   Evaluate(a,[1,2]);

Creation/History:

Creation: Lutz Lehmann, Berlin, August 2000
  This implementation improves drastically GL's basic bbp.
Modified: Lutz Lehmann, working: 17th August 2000
  This will be a really radical change. After having tested the pebble
  game concept and found a fundamental cut in memory usage, I will now
  try for speed.  Experimental profiling shows, that the construction
  of the gradients is the most time consuming part. Second, there
  should be no need to carry an empty Values array around, the same
  goes for the useless Remember table and the always reconstructed
  Pebbles table. What will be globally necessary is the Pebble Game
  and the list of Terminal nodes.
Modified: Gregoire, 20th August 2000.
  Integration in Kronecker 0.166beta-1.
Modified: Gregoire, 27th May 2001.
  Update from 0.166beta-3 to 0.166beta-4.
  Added partial evaluation scheme.
*/

// #Public
// Blackbox polynomials, Lifting fibers, Copying, Top

// #Package=Blackbox polynomials

//
// This package provides a blackbox polynomial domain and is due
// to L. Lehmann. The verbose flag is BbpVerbose.
//

// #Private

///////////////////////////////////////////////////
// Basic Arithmetic Operations on dag-core-level //
///////////////////////////////////////////////////



// TODO--1 (squaring bivariate objets takes time)







opConst:= {0, 1};
opUnary:= {2, 3};
opBinary:= {4, 5, 6};

bbp_zero:= <1,1>; // constant 0
bbp_one:= <1,2>;  // constant 1

dag_include:=function(R,e)
/*
  Input:
    R, a multivariate polynomial ring created by BlackboxPolynomialAlgebra
    e
  Output:
    the position of e in R`Dag`Ops.
  Description:
    The operation e is included in R`Dag`Ops.
  Implementation: straight but implements basic arithmetic properties.
  Side Effects:
    Modifies R`Dag
*/
  op:=e[1];
  if op in opConst then
    lev := 1; // all constants and variables in niveau 1
  elif op in opUnary then 
    lev := e[2]+1; 
  elif op in opBinary then 
    lev := Max(e[2],e[4])+1;
  else
    error "bbp.m, dag_include: Unkown operation";
  end if;
  if lev gt #R`Dag`Ops then
    // creation of a new level.
    Append(~R`Dag`Ops,[]);
  end if;

  // Remove Include to have only one scan of the set in Position
  pos := Position(R`Dag`Ops[lev],e);
  if pos eq 0 then // e not found
    Append(~R`Dag`Ops[lev],e);
    pos := #R`Dag`Ops[lev];
  end if;
  return <lev,pos>;
end function;

dag_var:=function(R,x)
/*
  Variable constructor.
*/
  e:=<1,Position(R`Dag`Variables,x),0,0,0>;
  return dag_include(R,e);
end function;
  
dag_cst:=function(R,c)
/*
  Includes constants.
*/
  if c eq 0 then return bbp_zero; end if;
  if c eq 1 then return bbp_one; end if;
  pos := Position(R`Dag`Constants,c);
  if pos eq 0 then
    Append(~R`Dag`Constants,c);
    pos := #R`Dag`Constants;
  end if;
  e := <0,pos,0,0,0>;
  return(dag_include(R,e));
end function;

dag_addmul:=function(R,symb,l)
/*
   Input: R a Bbp domain
          symb is opAdd or opMul
          l a sequence of (valid) operation addresses
   Output: the address of the (last) operation
   Remark: will only build binary operations for better
           time-space optimization
*/
  OPS:=R`Dag`Ops; // It's only a view as long as not modified

  // Split the list l so that the list lc contains 
  // only addresses to constants. We have time, since 
  // those decisions will only happen once for each operation.
  // Note that constants will always reside on the first level.
  lc := [ adr: adr in l  | adr[1] eq 1 ]; // scan for entries on first level
  lc := [ adr: adr in lc | OPS[1,adr[2]][1] eq 0 ]; //separate consts from vars

  //Do only something if there is a constant
  if #lc gt 0 then
    l  := [ adr: adr in l | not adr in lc ];
    // get the actual constants and carry out the operation
    lc := R`Dag`Constants[ [ OPS[1,adr[2]][2]  : adr in lc] ]; //some Magma magic
    c  := (symb eq 4) select &+lc else &*lc;
    if symb eq 6 and IsZero(c) then return bbp_zero; end if; // on absorbe :-)
    if not((symb eq 4 and IsZero(c)) or (symb eq 6 and c eq 1)) then
      Append(~l,dag_cst(R,c));
    end if;
  end if;
  // Trivial cases
  // Is optimization here slower than the overhead while evaluation operations on constants?
  if #l eq 0 then
    return (symb eq 4) select bbp_zero else bbp_one; //empty sum is zero, empty product one
  end if;

  if #l eq 1 then return l[1]; end if;    // only one argument

  // Build a sequential iterative operation tree
  while #l gt 1 do
    Sort(~l);
    
    // Note that #l >= 2 !
    acc := dag_include(R,<symb, l[1][1], l[1][2], l[2][1], l[2][2] >);
    l[2] := acc;  // Include the result in the list by removing the second operand
    Remove(~l,1); // remove also the first operand, so that the list contains now 
                  // instead of the operands the result
  end while;
  return acc;
end function;

dag_add:=function(R,l)
  return dag_addmul(R,4,l);
end function;

dag_mul:=function(R,l)
  return dag_addmul(R,6,l);
end function;

dag_neg:=function(R,x)
  if x eq bbp_zero then return x; end if;
  e := R`Dag`Ops[x[1],x[2]];
  if e[1] eq 3 then return(<e[2],e[3]>); end if;
  if e[1] eq 0 then return(dag_cst(R,-R`Dag`Constants[ e[2] ])); end if;
  return dag_include(R,<3,x[1],x[2],0,0>);
end function;

dag_sub:=function(R,l)
/*
  Note that #l is always 2.
*/
  if l[1] eq bbp_zero then return dag_neg(R,l[2]); end if;
  if l[2] eq bbp_zero then return l[1]; end if;
  e1:=R`Dag`Ops[l[1][1],l[1][2]];
  e2:=R`Dag`Ops[l[2][1],l[2][2]];
  if e1[1] eq 0 and e2[1] eq 0 then
    return dag_cst(R,R`Dag`Constants[e1[2]]-R`Dag`Constants[e2[2]]);
  end if;
  if e2[1] eq 3 then
    return dag_add(R,[l[1],<e2[2],e2[3]>]);
  end if;
  return dag_include(R,<5,l[1][1],l[1][2],l[2][1],l[2][2]>);
end function;

dag_power:=function(R,x,n)
/*
  It computes x^n in a binary tree, the only remaining power is 2.
  It deals correctly with all the particular cases!
*/
  //trivial cases
  if n eq 0 then return(bbp_one); end if; // since x^0=1
  if x in {bbp_zero, bbp_one } then return(x); end if; // since 0^n=0, 1^n=1

  e := R`Dag`Ops[x[1],x[2]];
  if e[1] eq 0 then
    return dag_cst(R, R`Dag`Constants[e[2]]^n);
  end if;
  if n eq 2 then // end of flat recursion
    return dag_include(R,<2,x[1],x[2],0,0>);
  end if;
  if n eq 1 then return x; end if;
  // now n>2
  fac:= [];
  pow := x;
  while n gt 0 do
    n,r := Quotrem(n, 2);
    if r eq 1 then Append(~fac,pow); end if;
    if n gt 0 then pow:= $$(R,pow,2); end if; //very flat recursion
  end while;
  
  return dag_mul(R,fac);
end function;

///////////////////////////////////////////////////////////////
// Evaluation and Diffentiation procedures on dag-core-level //
///////////////////////////////////////////////////////////////

Pebble:=
/*
    Record:
    Address is the position in Dag`Ops.
    FromStart ist the longest path length to any input, 
      does not change with time
    ToEnd the longest path length to any output
      will change when new output nodes appear
*/
  recformat<Address,FromStart,ToEnd>;

PebbleCompare:=function(x,y) 
/*
  Late evaluation
  
  According to the Sort description, the result will be negative
  if x is to be evaluated before y. If there is a path from the operation at
  x`Address to the one of y`Address then x'FromStart will be less than y`FromStart
  and x`ToEnd will be bigger than y`ToEnd. And vice versa, either
  the operations are independent or both conditions hold on the same time and
  x precedes y.
*/
  // x less than y if ToEnd is bigger
  d:=y`ToEnd    -x`ToEnd;     if d ne 0 then return d; end if;
  // x less than y if FromStart is smaller
  d:=x`FromStart-y`FromStart; if d ne 0 then return d; end if;
  return x`Address[2]-y`Address[2];
end function;

PebbleCompare2:=function(x,y) 
/*
  Early evaluation

  According to the Sort description, the result will be negative
  if x is to be evaluated before y. If there is a path from the operation at
  x`Address to the one of y`Address then x'FromStart will be less than y`FromStart
  and x`ToEnd will be bigger than y`ToEnd. And vice versa, either
  the operations are independent or both conditions hold on the same time and
  x precedes y.
*/
  // x less than y if FromStart is smaller
  d:=x`FromStart-y`FromStart; if d ne 0 then return d; end if;
  // x less than y if ToEnd is bigger
  d:=y`ToEnd    -x`ToEnd;     if d ne 0 then return d; end if;
  return x`Address[2]-y`Address[2];
end function;

PebbleCompare3:=function(x,y) 
/*
  Some mixed ordering, FromStart-2*ToEnd is still strictly increasing

  According to the Sort description, the result will be negative
  if x is to be evaluated before y. If there is a path from the operation at
  x`Address to the one of y`Address then x'FromStart will be less than y`FromStart
  and x`ToEnd will be bigger than y`ToEnd. And vice versa, either
  the operations are independent or both conditions hold on the same time and
  x precedes y.
*/
  // x less than y if FromStart is smaller
  ds := x`FromStart-y`FromStart;
  de := y`ToEnd    -x`ToEnd;
  dm := ds+2*de;
  if dm ne 0 then return dm; end if;
  // x less than y if ToEnd is bigger
  if ds ne 0 then return ds; end if;
  return x`Address[2]-y`Address[2];
end function;

dag_scan_calls:=function(R,outp,discard_current_evaluation)
/*
  R, returned by BlackboxPolynomialAlgebra.
  l, a set of addresses

  Scans the DAG backwards, starting at the output nodes given in l, 
  adds a pebble for every node whose value is used.
  So it marks all nodes needed for the calculation.

  Returns the array of pebbles.
*/
  IndentPush();
  vprint BbpVerbose,4:"--> dag_scan_calls";
  OPS:=R`Dag`Ops;
  N:=#OPS;
  Pebbles:= [[ 0 : i in z] : z in OPS];
  active := 0;
  
  if #outp eq 0 then IndentPop(); return Pebbles; end if;

  for z in outp do Pebbles[z[1]][z[2]]:= 1; end for;
  for niv:=N to 1 by -1 do
    for k := 1 to #Pebbles[niv] do
      if Pebbles[niv,k] gt 0 then
        active +:= 1;
        e:=OPS[niv,k];
        //for all but the input nodes mark the dependencies
        godown:=true;
        if assigned (R`Dag)`Evaluated and not discard_current_evaluation then
	  if niv le #R`Dag`Evaluated then
	    if k le #R`Dag`Evaluated[niv] then
              if R`Dag`Evaluated[niv,k] then godown:= false; end if;
	    end if;
	  end if;
        end if;
        if godown then
          if not e[1] in opConst then
            Pebbles[e[2],e[3]] +:=1; 
            if e[1] in opBinary then
              Pebbles[e[4],e[5]] +:=1; 
	    end if;
          end if;
        end if;
      end if;
    end for;
  end for;
  vprint BbpVerbose,4:"<-- dag_scan_calls";
  IndentPop();
  return Pebbles,active;
end function;  

dag_outdegree:=function(R,outp)
/*
  R, returned by BlackboxPolynomialAlgebra.
  outp, a sequence of addresses

  Scans the DAG backwards, terminal nodes will have the outdegree 0,
  all predecessors will have a bigger outdegree. In general it will
  be the length of the longest path to any output.

  It returns the array of outdegrees. All nodes in the computation tree
  will be marked.  
*/
  IndentPush();
  vprint BbpVerbose,4:"--> scan_outdegree";
  OPS:=R`Dag`Ops;
  N:=#OPS;
  OutDegree:=  [[ -1 : i in z] : z in OPS];
  active := 0;

  if #outp eq 0 then IndentPop(); return OutDegree; end if;

  for z in outp do OutDegree[z[1]][z[2]]:=1; end for;
  for niv:=N to 1 by -1 do
    for k := 1 to #OutDegree[niv] do
      if OutDegree[niv,k] ge 0 then
        active +:= 1;
        e:=OPS[niv,k];
        toend:=OutDegree[niv,k]+1;
        //for all but the input nodes mark the dependencies
        if not e[1] in opConst then
	  OutDegree[e[2],e[3]]:=Max(OutDegree[e[2],e[3]],toend); 
	  if e[1] in opBinary then
	    OutDegree[e[4],e[5]]:=Max(OutDegree[e[4],e[5]],toend);
	  end if; 
        end if;
      end if;
    end for;
  end for;
  vprint BbpVerbose,4:"<-- scan_outdegree";
  IndentPop();
  return OutDegree, active;
end function;  

build_PebbleGame:=function(R,F)
/*
    R, returned by BlackboxPolynomialAlgebra.
    F, sequence of expressions to be evaluated.

  It extracts the pebbles, calculates the distance from the input
  and sorts the pebble-game for good time locality of values and
  computations
  
  Side effect: destroys any computed result
*/
  IndentPush();
  vprint BbpVerbose,4:"--> build_PebbleGame for ",#F," output nodes";

  OutDegree, active := dag_outdegree(R, F);
  N:=#OutDegree;
  
  // pre-allocate a sequence, appending seems quite slow
  nil_pebble:=rec<Pebble|Address:=<1,1>, FromStart:=1, ToEnd:=1>;
  pebbleGame:=[nil_pebble: i in [1..active]];
  instr:=0;
  for niv := 1 to N do
    for k := 1 to #OutDegree[niv] do
      toend := OutDegree[niv,k];
      if toend ge 0 then
        instr +:= 1;
        pebbleGame[instr] :=rec<Pebble|Address:=<niv,k>, FromStart:=niv, ToEnd:=toend>;
     end if;
    end for;
  end for;
  Sort(~pebbleGame, PebbleCompare);
  vprint BbpVerbose,4:"<-- built_PebbleGame";
  IndentPop();
  return(pebbleGame);
end function;

dag_initialize_evaluation_pebblegame:=procedure(E,F)
/*
  It builds the pebble game and the pebbles for F.
  It must update the remember table in case the global dag has grown.
*/
  E`Dag`OutputNodes:= F;
  E`Dag`PebbleGame:= build_PebbleGame(E,F);

  if assigned (E`Dag)`Values and assigned (E`Dag)`Evaluated then
    if #E`Dag`Ops ge #E`Dag`Values then
      lev:=#E`Dag`Values;
      R:=Universe(E`Dag`Values[lev]);
      for i :=1 to lev do
        if #E`Dag`Ops[i] gt #E`Dag`Values[i] then
          E`Dag`Values[i] cat:= [ R| CoerceInteger(R,-123) : z in [1..#E`Dag`Ops[i]-#E`Dag`Values[i]]];
          E`Dag`Evaluated[i] cat:= [ false : z in [1..#E`Dag`Ops[i]-#E`Dag`Values[i]]];
        end if;
      end for;
      for i := lev+1 to #E`Dag`Ops do
        Append(~E`Dag`Values, [ R| CoerceInteger(R,-123) : z in E`Dag`Ops[i]]);
        Append(~E`Dag`Evaluated, [ false : z in E`Dag`Ops[i]]);
      end for;
    end if;
  end if;
end procedure;

dag_initialize_evaluation_pebbles:=procedure(E,F)
/*
  It marks the relevant part of the DAG with the number of pebbles.
*/
  E`Dag`Pebbles := dag_scan_calls(E,F,false);
end procedure;

dag_set_evaluation_point:=procedure(E,x)
/*
  It initializes the Values together with the remember
  table E`Dag`Values and E`Dag`Evaluated.
*/
  R:=Universe(x);
  E`Dag`EvaluationPoint:=x;
  E`Dag`Values:=[ [R| CoerceInteger(R,-123) : zz in z]: z in E`Dag`Ops];
  E`Dag`Evaluated:= [ [ false : zz in z]: z in E`Dag`Ops];
end procedure;

dag_evaluate:=function(E,F)
/*
   dag_initialize_evaluation_pebblegame and dag_set_evaluation_point
   must have been called.
   F must be a subset of E`Dag`OutputNodes.
*/
  IndentPush();
  vprint BbpVerbose,4:"--> dag_evaluate: ",#F," expressions ",Explode(F);

  assert &and[ z in E`Dag`OutputNodes : z in F];

  locPebbles := dag_scan_calls(E,F,false);
  
  OPS:=E`Dag`Ops;
  N:=#OPS;
  R:=Universe(E`Dag`EvaluationPoint);

  // Note that the 1 nodes are also pebbled
  vprint BbpVerbose,4:"dag_evaluate, size of pebble game ",#E`Dag`PebbleGame;
  
  totalnbinstr:=0;
  for instr :=1 to #E`Dag`PebbleGame do




    adr:=E`Dag`PebbleGame[instr]`Address;

    if locPebbles[adr[1],adr[2]] eq 0 or E`Dag`Evaluated[adr[1],adr[2]]
    then continue;
    end if;

    e:=OPS[adr[1],adr[2]];

    case e[1]:
      when 1   : v:=E`Dag`EvaluationPoint[e[2]];
      when 0   : v:=R!CoerceInteger(R,E`Dag`Constants[e[2]]);
      when 4   : v:=E`Dag`Values[e[2],e[3]]+E`Dag`Values[e[4],e[5]];
      when 6   : v:=E`Dag`Values[e[2],e[3]]*E`Dag`Values[e[4],e[5]];
      when 3   : v:=-E`Dag`Values[e[2],e[3]];
      when 5   : v:=E`Dag`Values[e[2],e[3]]-E`Dag`Values[e[4],e[5]];
      // TODO--1 (squaring bivariate objects takes more time)
      //when 2   : v:=E`Dag`Values[e[2],e[3]]^2;
      when 2   : v:=E`Dag`Values[e[2],e[3]]*E`Dag`Values[e[2],e[3]];
      else error "dag_evaluate: Unknown operation";
    end case;



    E`Dag`Values[adr[1],adr[2]]:=v;
    E`Dag`Evaluated[adr[1],adr[2]]:=true;

    // Clean Pebbles
    if not (e[1] in opConst) then
      for i := 2 to 4 by 2 do
        rniv := e[i]; rind := e[i+1];
        E`Dag`Pebbles[rniv,rind] -:= 1;
        if E`Dag`Pebbles[rniv,rind] eq 0 then
          E`Dag`Values[rniv,rind] := R!0;
          E`Dag`Evaluated[rniv,rind] := false;
        end if;
	// do this only once for unary operations
        if e[1] in opUnary then break; end if; 
      end for;
    end if;
    totalnbinstr +:=1;




  end for;// Main loop

  vprint BbpVerbose,4:"number of operations done ", totalnbinstr;
  vprint BbpVerbose,4:"<-- dag_evaluate, maximal data size ", Ceiling(GetMaximumMemoryUsage()/1024),"kB.";
  IndentPop();
  return [E`Dag`Values[z[1],z[2]]: z in F];;
end function;

///////////////////////////////////////
// Differentiation on dag-core-level //
///////////////////////////////////////

dag_gradient_forward:=function(E,Fun)
/*
   Scans the computation tree for needed nodes by the outputs in Fun
   Generates additional nodes for the differentials of the tree,
   works in forward mode starting with the identity on the variables.
   Since the values are vectors of simple integers, nothing will be 
   thrown away.
*/
  IndentPush();
  vprint BbpVerbose,2:"--> dag_gradient_forward of ",#Fun," functions";
  Pebbles,todo := dag_scan_calls(E,Fun,true);
  
  OPS:=E`Dag`Ops;
  N:=#OPS;
  dim := #E`Dag`Variables;
  
  vprint BbpVerbose,4:"Derivation in ",todo," nodes";
  
  done := 0;
  // pre-build the derivations array, deal only with addresses
  zeroderivative := [bbp_zero: j in[1..dim] ];
  partdif:=[ [ zeroderivative: i in z] : z in OPS]; 
  
  for niv:=1 to N do
    for adr := 1 to #Pebbles[niv] do
      if Pebbles[niv, adr] eq 0 then continue; end if;
      






		 
      op:=OPS[niv,adr];
      if op[1] eq 0 then continue; end if;
      if op[1] eq 1 then partdif[niv,adr][op[2]] := bbp_one; continue; end if;

      a:=<op[2],op[3]>;
      a1:=partdif[a[1],a[2]];
      if op[1] in opBinary then
        b:=<op[4],op[5]>;
        b1:=partdif[b[1],b[2]];
      end if;
	      
      case op[1]:
        when 4 : for zz := 1 to dim do
                       partdif[niv,adr][zz] := dag_add(E,[ a1[zz] ,b1[zz] ]);
		     end for;
        when 6 : for zz := 1 to dim do
		     partdif[niv,adr][zz]:=
	               dag_add(E,[  dag_mul(E,[ a, b1[zz] ]),
			            dag_mul(E,[ b, a1[zz] ])  ]);
		     end for;
        when 3 : for zz := 1 to dim do
                       partdif[niv,adr][zz] := dag_neg(E,a1[zz]);
		     end for;
        when 5 : for zz := 1 to dim do
		       partdif[niv,adr][zz] := dag_sub(E,[a1[zz],b1[zz]]);
		     end for;
        when 2 : two_a := dag_mul(E, [ dag_cst(E,2), a]);
		     for zz := 1 to dim do
                       partdif[niv,adr][zz] := dag_mul(E, [ two_a, a1[zz] ]);
		     end for;
        else error "dag_gradient_forward: Unknown operation";
      end case;



    end for;
  end for;



  vprint BbpVerbose,2:"<-- dag_gradient_forward";
  IndentPop();
  return [  [ partdif[zz[1],zz[2]][i] : i in [1..dim] ]  :  zz in Fun  ];
end function;

dag_gradient_backward:=function(E,N)
/*
   E, RngMPolBB domain,
   N, address.
  
   The sequence of the partial derivatives of Fun.
  
   Baur-Strassen's algorithm.
*/
  vprint BbpVerbose,2:"--> dag_gradient_backward";
  pebbles:= dag_scan_calls(E,[N],true);

  // on traite le cas des variables a part.
  if N[1] eq 1 and N[2] lt 3 then
    return [ bbp_zero : i in [1..Rank(E)]];
  end if;
  if N[1] eq 1 and N[2] lt Rank(E)+2 then
    grad:=[ bbp_zero : i in [1..Rank(E)]];
    grad[N[2]-2]:= bbp_one;
    return(grad);
  end if;

  OPS:= E`Dag`Ops;

  grad:=[ [ bbp_zero : i in [1..#OPS[l]] ] : l in [1..N[1]] ];
 
  // grad[lev,i] stores the partial derivative with respect to the
  // node [lev,i].

  grad[N[1],N[2]]:= bbp_one;

  for lev:=N[1] to 1 by -1 do
  for i:=#OPS[lev] to 1 by -1 do

    if pebbles[lev,i] eq 0 then continue; end if;
    e:=OPS[lev,i];
    case e[1]:
      when 4:
        grad[e[2],e[3]]:= dag_add(E,[grad[e[2],e[3]],grad[lev,i]]);
        grad[e[4],e[5]]:= dag_add(E,[grad[e[4],e[5]],grad[lev,i]]);
 
      when 6:
        grad[e[2],e[3]]:= dag_add(E,[grad[e[2],e[3]], dag_mul(E,[ grad[lev,i], <e[4],e[5]>])]);
        grad[e[4],e[5]]:= dag_add(E,[grad[e[4],e[5]], dag_mul(E,[ grad[lev,i], <e[2],e[3]>])]);
 
      when 3:
        grad[e[2],e[3]]:= dag_sub(E,[grad[e[2],e[3]], grad[lev,i]]);

      when 5:
        grad[e[2],e[3]]:= dag_add(E,[grad[e[2],e[3]], grad[lev,i]]);
        grad[e[4],e[5]]:= dag_sub(E,[grad[e[4],e[5]], grad[lev,i]]);
 
      when 2:
        grad[e[2],e[3]]:= dag_add(E,[grad[e[2],e[3]], dag_mul(E,[dag_cst(E,2),grad[lev,i],<e[2],e[3]>])]);

      when 0: break;
      when 1: break;

      else:
        error "dag_gradient_backward: Unknown operation";        
    end case;
 
  end for; end for;
  // Ca presuppose que les variables sont dans l'ordre, arg... attention...
  vprint BbpVerbose,2:"<-- dag_gradient_backward";
  return([grad[1,i]: i in [3..Rank(E)+2]]);
 
end function;

////////////////////////////////////////////////
// Upper bound total degree on dag-core-level //
////////////////////////////////////////////////

dag_upper_bound_total_degree:=function(E,x)
/*
  Returns an upper bound on the total degree of e.
*/
  
  OPS:=E`Dag`Ops;

  e:=OPS[x[1],x[2]];

  case e[1]:
      when 1   : return 1;
      when 0   : return 0;
      when 4   : return Max($$(E,[e[2],e[3]]),$$(E,[e[4],e[5]]));
      when 6   : return $$(E,[e[2],e[3]])+$$(E,[e[4],e[5]]);
      when 3   : return $$(E,[e[2],e[3]]);
      when 5   : return Max($$(E,[e[2],e[3]]),$$(E,[e[4],e[5]]));
      // TODO--1 (squaring bivariate objects takes more time)
      when 2   : return 2*$$(E,[e[2],e[3]]);
      else error "dag_upper_bound_total_degree: Unknown operation";
  end case;

end function;

//////////////////////////////////////////
// Upper bound height on dag-core-level //
//////////////////////////////////////////

height:=function(x)
/*
  Return the maximum of the log_2 (integral part +1) of the numerator and
  denominator of x.
*/
  tmp:= Max([1,Ilog2(Max(1,Abs(Numerator(x)))), Ilog2(Max(1,Abs(Denominator(x))))])+1;
  return tmp;
end function;

dag_upper_bound_height:=function(E,x)
/*
  Returns an upper bound on the height of e, that is the
  maximum of the bit sizes of the integers of the expanded form.

  --  We do not realy compute a true upperbound, because it is too
  --  much pessimistic.
  --  This function is only to be used for heuristic choice of primes
  --  for modular solving.
*/
  OPS:=E`Dag`Ops;

  e:=OPS[x[1],x[2]];
  if Type(BaseRing(E)) ne FldRat then
    error "Base ring must be the field of the rational numbers";
  end if;
  case e[1]:
      when 1   : return 0;
      when 0   : return height(E`Dag`Constants[e[2]]);
      when 4   : return Max($$(E,[e[2],e[3]]),$$(E,[e[4],e[5]]));
      when 6   : return $$(E,[e[2],e[3]])+$$(E,[e[4],e[5]]);
      when 3   : return $$(E,[e[2],e[3]]);
      when 5   : return Max($$(E,[e[2],e[3]]),$$(E,[e[4],e[5]]));
      // TODO--1 (squaring bivariate objects takes more time)
      when 2   : return 2*$$(E,[e[2],e[3]]);
      else error "dag_upper_bound_height: Unknown operation";
  end case;

end function;

/*-----------------------------------------------------------------------*
* DAG-Manipulations outside this module can only be done by using those  *
* RngMPolBBPointers, allowing for readable code, some type safety and    *
* transparency of data structure.                                        *
*-----------------------------------------------------------------------*/

RngMPolBBPointers:=
/*
    Record:

    Parent is the Multivariate Polynomial Algebra domain where the
      RngMPolBB live. 
    Address is the position in Parent`Dag`Ops.
*/
  recformat<Parent,Address>;

IsRngMPolBBPointer:=function(e)
/*
  Returns true iff e has type recformat<Parent,Address>, else false.
*/
  return(Names(Format(e)) eq Names(RngMPolBBPointers));
end function;

// #Public
//////////////////////////////////////////////////
// #Section=Creation of a blackbox polynomial algebra
//////////////////////////////////////////////////

intrinsic BlackboxPolynomialAlgebra(R::Rng,n::RngIntElt) -> RngMPol
{
 It returns a ring of multivariate blackbox polynomials
 over R in n variables.
}
  A:=recformat <
   Ops,            // Nodes
   Variables,      // Sequence of the variables
   Constants,      // Sequence of the constants
   PebbleGame,     // Sequence of records, order of calculation
   Pebbles,        // Pebbles
   Values,         // Remember table for values
   Evaluated,      // Tells whether a node has already been evaluated
   EvaluationPoint,// Stores the current evaluation point
   OutputNodes,    // Stores the current output nodes
   Gradients       // Stores gradients [ ...,[bbpptr, grad_bbpptr],...]
 >;
 AddAttribute(RngMPol,"Dag");
 AddAttribute(RngMPol,"Configuration");
 AddAttribute(RngMPol,"MathMLsession");
 MP:=PolynomialAlgebra(R,n);

 MP`Dag:=rec<A|
   Ops:=[[ <0,1,0,0,0>,<0,2,0,0,0> ]],
   Variables:=[MP.i: i in [1..n]],
   Constants:= [R|0,1],
   Gradients:= []
 >;

 toto:=[ Var(MP.i): i in [1..n] ];
    // c'est pour etre sur que les variables sont
    // dans le bon ordre : c'est pour la fonction Gradient qui utilise
    // l'algorithme (backward) Baur-Strassen...
 toto:= [* MP *] cat SequenceToList(toto);
 return Explode(toto);
end intrinsic;

//////////////////////////////
// #Section=Creation of elements
//////////////////////////////
// #List
intrinsic Var(x::RngMPolElt) -> Rec
{
  
    It returns the blackbox polynomial element corresponding to x.
  
}
  R:=Parent(x);
  return rec<RngMPolBBPointers|Parent:=R,Address:=dag_var(R,x)>;
end intrinsic;

intrinsic Cst(E::RngMPol,c) -> Rec
{
    
       E is a domain returned by BlackboxPolynomialAlgebra.
       c is an element of the base ring of R.
    
    
      It returns the blackbox polynomial corresponding to c.
    
}
  require c in Universe(E`Dag`Constants) : "Invalid Constant";
  return rec<RngMPolBBPointers|Parent:=E,Address:=dag_cst(E,c)>;
end intrinsic;
// #EndList


// #Section=Arithmetic operations
/////////////////////////////////

// All elementary ring operations are supported.
// #List
intrinsic '+'(x::Rec) -> Rec
{
}
  require IsRngMPolBBPointer(x): "Bad Unary +";
  return(x);
end intrinsic;

intrinsic '+'(x::Rec,y) -> Rec
{
}
  require IsRngMPolBBPointer(x): "Bad Binary +";
  R:=x`Parent;
  if Type(y) ne Rec then
    res := dag_add(R,[x`Address,dag_cst(R,y)]);
  else
    require IsRngMPolBBPointer(y): "Bad Binary +";
    res := dag_add(R,[x`Address,y`Address]);
  end if;
  return rec<RngMPolBBPointers|Parent:=R,Address:=res>;    
end intrinsic;

intrinsic '+'(y,x::Rec) -> Rec
{
}
  return(x+y);
end intrinsic;

intrinsic '-'(x::Rec) -> Rec
{
}
  R:=x`Parent;
  return rec<RngMPolBBPointers|Parent:=R,Address:=dag_neg(R,x`Address)>;
end intrinsic;

intrinsic '-'(x::Rec,y) -> Rec
{
}
  require IsRngMPolBBPointer(x): "Bad Binary -";
  R:=x`Parent;
  if Type(y) ne Rec then
    res := dag_sub(R,[x`Address,dag_cst(R,y)]);
  else
    require IsRngMPolBBPointer(y): "Bad Binary -";
    res := dag_sub(R,[x`Address,y`Address]);
  end if;
  return rec<RngMPolBBPointers|Parent:=R,Address:=res>;
end intrinsic;

intrinsic '-'(y,x::Rec) -> Rec
{
}
  require IsRngMPolBBPointer(x): "Bad Binary -";
  R:=x`Parent;
  if Type(y) ne Rec then
    res := dag_sub(R,[dag_cst(R,y),x`Address]);
  else
    require IsRngMPolBBPointer(y): "Bad Binary -";
    res := dag_sub(R,[y`Address,x`Address]);
  end if;
  return rec<RngMPolBBPointers|Parent:=R,Address:=res>;
end intrinsic;

intrinsic '*'(x::Rec,y) -> Rec
{
}
  require IsRngMPolBBPointer(x): "Bad Binary *";
  R:=x`Parent;
  if Type(y) ne Rec then
    res := dag_mul(R,[x`Address,dag_cst(R,y)]);
  else
    require IsRngMPolBBPointer(y): "Bad Binary *";
    res := dag_mul(R,[x`Address,y`Address]);
  end if;
  return rec<RngMPolBBPointers|Parent:=R,Address:=res>;
end intrinsic;

intrinsic '*'(y,x::Rec) -> Rec
{
}
  return(x*y);
end intrinsic;

intrinsic '/'(x::Rec,c) -> Rec
{
 c must be invertible.
}
  require IsRngMPolBBPointer(x): "Bad Binary /"; 
  R:=x`Parent;
  require c in Universe(R`Dag`Constants) : "Invalid nonscalar Division";
  if c eq 0 then error "Division by zero in scalar Division"; end if;
  if c eq 1 then return x; end if;
  return(c^(-1)*x);
end intrinsic;

intrinsic '^'(x::Rec,n::RngIntElt) -> Rec
{
 n must be a non-negative integer.
}
  require IsRngMPolBBPointer(x): "Bad Binary ^";
  R:=x`Parent;
  return rec<RngMPolBBPointers|Parent:=R,Address:=dag_power(R,x`Address,n)>;
end intrinsic;

intrinsic Add(E::RngMPol,l::[Rec])-> Rec
{
  
      E, returned by BlackboxPolynomialAlgebra.
      l, a sequence.
  
  
    It returns the sum of the element of l.
  
}
  return rec<RngMPolBBPointers|
    Parent:=E,
    Address:=dag_add(E,[ (Type(x) eq Rec) select 
                           x`Address 
                         else 
                           dag_cst(E,x)
                       : x in l])>;
end intrinsic;

intrinsic Product(E::RngMPol,l::[Rec])-> Rec
{
  
      E, returned by BlackboxPolynomialAlgebra.
      l, a sequence.
  
  
    It returns the product of the element of l.
  
}
  return rec<RngMPolBBPointers|
    Parent:=E,
    Address:=dag_mul(E,[ (Type(x) eq Rec) select 
                            x`Address 
                          else 
                            dag_cst(E,x)
                          : x in l])>;
end intrinsic;
// #EndList

// #Section=Equality tests
//////////////////////////

// Equality test corresponds to the test on the representations:
// two blackbox polynomials are equal iff their address is the same.

// #List
intrinsic IsZero(e::Rec)-> BoolElt
{
 Tells whether the expression e is the constant 0 or not.
}
  return(e`Address eq bbp_zero);
end intrinsic;

intrinsic IsOne(e::Rec)-> BoolElt
{
 Tells whether the expression e is the constant 1 or not.
}
  return(e`Address eq bbp_one);
end intrinsic;
// #EndList

// #Section=Evaluation
///////////////////////
// #List

intrinsic ClearRememberTableValues(~E::RngMPol)
{
 It clears the remember table of the current evaluation process.
}
  delete (E`Dag)`Values;
  delete (E`Dag)`Evaluated;
  delete (E`Dag)`EvaluationPoint;
  delete (E`Dag)`PebbleGame;
  delete (E`Dag)`Pebbles;
  delete (E`Dag)`OutputNodes;
end intrinsic;

intrinsic Evaluate(F::Rec,x::[]) -> .
{
}
  return(Evaluate([F],x)[1]);
end intrinsic;

intrinsic Evaluate(F::[Rec],x::[]) -> .
{
  
     It returns the sequence of the values of the elements
     of F when the i-th variable is specialized to
     x[i], for i from 1 to #x.
     If F is empty then it returns [Universe(x)|].
  
    
     #x must equal the rank of the polynomial domain.
  
}
  begintime:=Cputime();
  IndentPush();
  vprint BbpVerbose,1:"--> Evaluate";
  U:=Universe(x);
  if IsEmpty(F) then IndentPop(); return [U|]; end if;
  R:=F[1]`Parent;
  require Rank(R) eq #x : "Evaluate: evaluation point has incorrect size";

  if assigned (R`Dag)`OutputNodes
    and assigned (R`Dag)`PebbleGame
    and assigned (R`Dag)`Pebbles 
  then
    if &or[not(z`Address in R`Dag`OutputNodes) :z in F] then 
      dag_initialize_evaluation_pebblegame(R, [ f`Address:f in F]);
      dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
    end if;
  else
    dag_initialize_evaluation_pebblegame(R, [ f`Address:f in F]);
    dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
  end if;

  if assigned (R`Dag)`EvaluationPoint
    and assigned (R`Dag)`Values
    and assigned (R`Dag)`Evaluated
  then
    if Universe(x) cmpne Universe(R`Dag`EvaluationPoint) then
      dag_set_evaluation_point(R, x);
      dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
    else
      if x ne R`Dag`EvaluationPoint then
        dag_set_evaluation_point(R, x);
        dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
      end if;
    end if;
  else
    dag_set_evaluation_point(R, x);
    dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
  end if;

  ans := dag_evaluate(R,[ f`Address:f in F]);

  vprint BbpVerbose,1:"<-- Evaluate, elapsed time", Cputime()-begintime;
  IndentPop();
  return(ans);
end intrinsic;

intrinsic InitializeEvaluation(F::[Rec])
{
  
      F, a nonempty sequence of expressions.
  
  
     It initializes the evaluation process that will serve to evaluate
     the element of F only.
  
}
  require not IsEmpty(F): "empty sequence";
  R:=F[1]`Parent;
  dag_initialize_evaluation_pebblegame(R, [ f`Address:f in F]);
  dag_initialize_evaluation_pebbles(R, [ f`Address:f in F]);
end intrinsic;

intrinsic ChangeUniverseValues(~E::RngMPol,R::Rng,f::UserProgram)
{
  
      E, a domain returned by BlackboxPolynomialAlgebra.
      R, a ring
      f, map into R.
  
  
     It applies f onto the remember table of the current evaluation
     process.
  
}
  if assigned (E`Dag)`EvaluationPoint then
    E`Dag`EvaluationPoint:= [R| f(z) : z in E`Dag`EvaluationPoint];
  end if;
  if assigned (E`Dag)`Values then
    E`Dag`Values:= [ [R| f(zz) : zz in z] : z in E`Dag`Values];
  end if;
end intrinsic;
// #EndList
 
// #SubSection=Derivatives
////////////////////////

// #private
seq_associate:=function(s,i)
  for e in s do
    if e[1] eq i then return true, e[2]; end if;
  end for;
  return false, false;
end function;

seq_associate_seq:=function(s,si)
  ans:=[];
  for i in si do
    found:=false;
    for e in s do
      if e[1] eq i then Append(~ans, e[2]); found:= true; break; end if;
    end for;
    if not found then return false, false; end if;
  end for;
  return true, ans;
end function;

// #public
intrinsic Derivative(e::Rec,i::RngIntElt:algorithm:="backward") -> Rec
{
 The derivative of e wrt its i-th variable.
}
  return(Gradient(e:algorithm:=algorithm)[i]);
end intrinsic;

intrinsic Gradient(e::Rec:algorithm:="backward") -> []
{
}
  return Gradient([e]:algorithm:=algorithm)[1];
end intrinsic;

intrinsic Gradient(F::[Rec]:algorithm:="backward") -> []
{
  Sequence of the partial derivatives.
  The parameter algorithm can be "backward" (for Baur-Strassen)
  or "forward".
}
  if #F eq 0 then return []; end if;
  MP:=F[1]`Parent;
  EP:=RngMPolBBPointers;

  // Is in remember table ?
  b,v:= seq_associate_seq(MP`Dag`Gradients,[f`Address : f in F]);
  if b then



    DF:=v;
  else
    if algorithm eq "backward" then
      DF := [ dag_gradient_backward(MP,f`Address): f in F];
    elif algorithm eq "forward" then
      DF := dag_gradient_forward(MP,[f`Address : f in F]);
    else
      error "unkown specified algorithm";
    end if;
    // Into remember table
    for i:=1 to #F do
      Append(~MP`Dag`Gradients,<F[i]`Address,DF[i]>);
    end for;
  end if;
  tmp := [ [ rec<EP|Parent:=MP,Address:=adr> : adr in Df] : Df in DF ];
  return tmp;
end intrinsic;

intrinsic ClearRememberTableGradients(~E::RngMPol)
{
 It clears the remember tables used for gradient storage.
}
  E`Dag`Gradients:=[];
end intrinsic;

// #Section=Degree, coefficients
////////////////////////////////

// #List
intrinsic ConvertToPolynomial(e::Rec) -> .
{
}
  return(ConvertToPolynomial([e])[1]);
end intrinsic;

intrinsic ConvertToPolynomial(e::[Rec]) -> .
{
  
   e is an expression or a sequence of expressions.
  
  
    It returns the multivariate polynomial elements represented by e.
  
    Implementation: uses Evaluate

    Note LL: "Dense Expansion -- a very bad idea."
}
  require &and[IsRngMPolBBPointer(z) : z in e] :"bad first agument";    
  MP:=e[1]`Parent;
  return(Evaluate(e,[MP.i : i in [1..Rank(MP)]]));
end intrinsic;

intrinsic IsHomogeneous(e::Rec) -> BoolElt
{
  It tells whether e is homogeneous (w.r.t. to grading on
  the variables of its polynomial ring). It is not probabilistic.
}
  require IsRngMPolBBPointer(e): "HomogeneousTotalDegree: bad first agument";
  p:= ConvertToPolynomial(e);
  return(IsHomogeneous(p));
end intrinsic;

intrinsic TotalDegree(e::Rec:Strategy:="UpperBound") -> RngIntElt
{
}
  return TotalDegree([e]:Strategy:= Strategy)[1];
end intrinsic;

intrinsic TotalDegree(F::[Rec]:Strategy:="UpperBound") -> .
{
  It returns the total degrees of the mulivariate polynomial
  represented by F.
  
    Strategy can be either
    "Deterministic" or "Probabilistic".
    Strategy can also be "UpperBound" in order to compute
    a deterministic upper bound only. 
  
}
  if IsEmpty(F) then return [Integers()|]; end if;
  require &and[IsRngMPolBBPointer(e) : e in F] : "bad first agument";

  if Strategy eq "Deterministic" then
    p:= ConvertToPolynomial(F);
    return [TotalDegree(z) : z in p];
  end if;
  
  if Strategy eq "Probabilistic" then
    R:=F[1]`Parent;
    UP<T>:=PolynomialAlgebra(BaseRing(R));
    n := Rank(R);
    eqs := #F;
    degs := [0 : i in [1..eqs] ];
    for count := 1 to n do
      x := [ Random(Factorial(n+1)+10)+T*(10+Random(Factorial(n+1))): i in [1..n] ];
      P := Evaluate(F,x);
      newdegs := [Degree(p): p in P];
      if newdegs ne degs then
        for i := 1 to eqs do
          degs[i] := Max(degs[i],newdegs[i]);
        end for;
      end if;
     end for;
     return(degs);
  end if;

  if Strategy eq "UpperBound" then
    R:=F[1]`Parent;
    return [Integers()|dag_upper_bound_total_degree(R,z`Address) :z in F];
  end if;
  
  error "Unknown Strategy parameter";
end intrinsic;
// #EndList

// #Private
intrinsic UpperBoundHeight(F::Rec) -> RngIntElt
{
}
  return UpperBoundHeight([F])[1];
end intrinsic;

intrinsic UpperBoundHeight(F::[Rec]) -> SeqEnum
{
  It returns more or less an upper bit-size bound for in integers
  of the mulivariate polynomial
  represented by F when expanded. The base ring must be the
  field of the rationals.
}
  /* This function is only used to determine an heuristic for 
     the choice of the prime that is used for modular solving */
  if IsEmpty(F) then return [Integers()|]; end if;
  require &and[IsRngMPolBBPointer(e) : e in F] : "bad first agument";
  R:=F[1]`Parent;
  require Type(BaseRing(R)) eq FldRat : "base ring must be the rational numbers";
  return [Integers()|dag_upper_bound_height(R,z`Address)+1 :z in F];
end intrinsic;

// #Public

// #Section=Linear algebra
//////////////////////////
// #List

// The following linear algebra functionalities are based on Berkowitz'
// algorithm.

intrinsic CharacteristicPolynomialBerkowitz(A)->.
//-----------------------------------------------
{
  
     A, sequence of sequences of expressions, viewed as a
      square matrix.
  
  
    It returns the sequence of coefficients of the characteristic polynomial
    of A, the last element being the coefficient of degree 0.
  
}
    n:=#A;
    require n gt 0: "empty square matrix";
    require &and[#e eq n: e in A]: "matrix not square";

    E:=A[1][1]`Parent;
    if n eq 1 then return([Cst(E,-1),A[1][1]]); end if;
    if n eq 2 then return([Cst(E,1),-A[2][2]-A[1][1],-A[2][1]*A[1][2]+A[2][2]*A[1][1]]); end if;

    Vect:=[ Cst(E,0): i in [1..n+1]];
    Vect[1]:=Cst(E,1);
    Vect[2]:=-A[2][2]-A[1][1];
    Vect[3]:=-A[2][1]*A[1][2]+A[2][2]*A[1][1];
    C:=[ [Cst(E,0): j in [1..n]]: i in [1..n+1]];
    R:=[ Cst(E,0): i in[1..n-1]];
    Q:=[ Cst(E,0): i in[1..n-1]];
    temp:=[ Cst(E,0): i in[1..n+1]];
    for r:=3 to n do
      for i:=1 to r-1 do R[i]:=A[r][i]; end for;
      for i:=1 to r-1 do Q[i]:=A[i][r]; end for;
      C[1][1]:=Cst(E,-1);
      C[2][1]:=A[r][r];
      C[3][1]:=Add(E,[R[i]*Q[i]: i in [1..r-1]]);
      for i:=1 to r-1 do temp[i]:=Add(E,[A[i][j]*Q[j]:j in [1..r-1]]); end for;
      for i:=1 to r-1 do Q[i]:=temp[i]; end for;
      C[4][1]:=Add(E,[R[i]*Q[i]: i in [1..r-1]]);
      for i:=3 to r-1 do
        for j:=1 to r-1 do temp[j]:=Add(E,[A[j,k]*Q[k]:k in [1..r-1]]); end for;
        for j:=1 to r-1 do Q[j]:=temp[j]; end for;
        C[i+2][1]:=Add(E,[R[j]*Q[j]: j in [1..r-1]]);
      end for;
      for i:=2 to r do
        for j:=0 to r+1-i do
          C[i+j][i]:=C[j+1][1];
        end for;
      end for;
      for i:=1 to r+1 do temp[i]:=Add(E,[C[i][j]*Vect[j]: j in [1..r]]); end for;
      for i:=1 to r+1 do Vect[i]:=temp[i]; end for;
    end for;
    return(Vect);
end intrinsic;

intrinsic DeterminantBerkowitz(A)->.
//----------------------------------
{
  
     A, sequence of sequences of Expressions, viewed as a
      square matrix.
  
  
    It returns the determinant of A.
  
}
  cp:=CharacteristicPolynomialBerkowitz(A);
  return(cp[#cp]);
end intrinsic;

// #EndList

// #Section=Printing
////////////////////

forward SubPrintBbp;

SubPrintSeqOp:=procedure(~dag,t,u,v)

  SubPrintBbp(~dag,u);
  printf "%o",t;
  SubPrintBbp(~dag,v);
end procedure;


SubPrintBbp:=procedure(~dag,x)

  e:=dag`Ops[x[1],x[2]];
  case e[1] :
    when 1 : printf "%o",dag`Variables[e[2]];
    when 0 : printf "%o",dag`Constants[e[2]];
    when 4 : printf "(";
                 SubPrintSeqOp(~dag,"+",[e[2],e[3]],[e[4],e[5]]);
                 printf ")";
    when 6 : printf "(";
                 SubPrintSeqOp(~dag,"*",[e[2],e[3]],[e[4],e[5]]);
                 printf ")";
    when 3 : printf "(-(";
                 $$(~dag,[e[2],e[3]]);
                 printf "))";
    when 5 : printf "(";
                 SubPrintSeqOp(~dag,"-",[e[2],e[3]],[e[4],e[5]]);
                 printf ")";
    when 2 : if not dag`Ops[e[2],e[3]][1] in [1,0] then
                   printf "(";
		 end if;
                 $$(~dag,[e[2],e[3]]);
                 if not dag`Ops[e[2],e[3]][1] in [1,0] then
                   printf ")";
		 end if;
                 printf "^2";
  end case;

end procedure;

intrinsic PrintBbp(s::[])
{
}
  printf "[";
  IndentPush();
  first:=true;
  for e in s do
    if not first then printf ", "; end if;
    PrintBbp(e);
    first:=false;
  end for;
  IndentPop();
  printf "]";
end intrinsic;

intrinsic PrintBbp(e::Rec)
{
 It prints the expression represented by e.
}
  require IsRngMPolBBPointer(e): "bad agument";
  if Type(e) eq Rec then
    dag:=e`Parent`Dag;
    x:=e`Address;
    SubPrintBbp(~dag,x);
  else
    print(e);
  end if;

end intrinsic;
