// # 1 "equidec.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 16:35:03 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "equidec.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "equidec.mtx" 2


// #Public

// Equidimensional Decomposition, Geometric Solve, Hensel Lifting, Top

// #Package=Equidimensional decomposition
//////////////////////////////////////

// This section provides functionalities to handle equidimensional
// decomposition. The verbose flag is GeometricSolveVerbose.

// #Section=Inclusion of components
////////////////////////////////

intrinsic IsIncludedIrreducibleLF(lf1,l,lf2::Rec)-> BoolElt
{
 
    lf1 is a fiber.
    l is an integer.
    lf2 is an isolated lifting fiber.
 
 
   The function tells whether the lth irreducible factor of lf1
   is included in lf2 or not. If it is detected that lf2
   is not an isolated lifting fiber then a string is returned.
 
}
  IndentPush();
  vprint GeometricSolveVerbose,1:"--> IsIncludedIrreducibleLF in degrees ",Degree(lf1`MinimalPolynomial[l])," \\ ",DegreeLF(lf2);
  // Stamp equality prevents from inclusions
  if lf1`Stamp eq lf2`Stamp then
    vprint GeometricSolveVerbose,3: "Using stamps";
    vprint GeometricSolveVerbose,1: "--> IsIncludedIrreducibleLF";
    IndentPop();
    return false;
  end if;

  // Non-trivial cases
  F:=lf1`ResolutionField;
  q:=lf1`MinimalPolynomial[l];
  n:=NumberOfVariablesLF(lf1);
  UP<T>:=Parent(q);
  K<a>:=quo<UP|q>;
  r1:=DimensionLF(lf1);
  r2:=DimensionLF(lf2);
  vprint GeometricSolveVerbose,2: "Inclusion of a component of dimension",r1,"in one of dimension",r2;
  // Components have same dimension
  if r1 eq r2 then
    assert lf1`ChangeOfVariables`LinearPart eq lf2`ChangeOfVariables`LinearPart
      and lf1`ChangeOfVariables`AffinePart eq lf2`ChangeOfVariables`AffinePart
      and lf1`MagicPoint eq lf2`MagicPoint
      and lf1`PrimitiveElement eq Parent(lf1`PrimitiveElement)!lf2`PrimitiveElement;
    IndentPop();
    for z in lf2`MinimalPolynomial do
      if K!z eq 0 then return true; end if;
    end for; 
    return false;
  end if;
  // lf1 is simple, we use the Jacobian criterion
  if not IsMultipleLF(lf1) and
    { z`Address : z in lf2`LiftingSystem} subset { z`Address : z in lf1`LiftingSystem }
  then
    vprint GeometricSolveVerbose,3: "Using Jacobian criterion";
    point:= ChangeUniverse(lf1`MagicPoint,K) cat ChangeUniverse(lf1`Parametrization[l],K);
    jac:= [Gradient(z): z in lf1`LiftingSystem];
    M:=Matrix(K,#lf1`LiftingSystem,n,EvaluateLF(lf1,&cat jac,point));
    tmp:=Rank(M);
    vprint GeometricSolveVerbose,1: "--> IsIncludedIrreducibleLF";
    IndentPop();
    return tmp lt n-r1;
  end if;
  // Generic case
  loclf2:=lf2;
  ChangeResolutionFieldLF(~loclf2,K);
  pgen:=ChangeUniverse(lf1`MagicPoint,K) cat ChangeUniverse(lf1`Parametrization[l],K);
  pgen:=[K| &+[K|K!lf1`ChangeOfVariables`LinearPart[i,j]*pgen[j] : j in [1..n]]
    + K!lf1`ChangeOfVariables`AffinePart[i] : i in [1..n]];
  M:=(lf2`ChangeOfVariables`LinearPart)^(-1);
  pgen:=[K| &+[K!M[i,j]*(pgen[j]-K!lf2`ChangeOfVariables`AffinePart[j])
    : j in [1..n]] : i in [1..n]];
  magicpoint:=[pgen[i]: i in [1..r2]];
  if r2 gt 0 then
    ChangeMagicPointLF(~loclf2,magicpoint);
    if Type(loclf2) eq MonStgElt then
      IndentPop();
      return loclf2;
    end if;
  end if;
  u:=Evaluate(loclf2`PrimitiveElement,[K |0 : i in [1..r2]]
     cat [ K | pgen[i] : i in [r2+1..n]]);
  IndentPop();
  vprint GeometricSolveVerbose,1: "--> IsIncludedIrreducibleLF";
  for z in loclf2`MinimalPolynomial do
    if Evaluate(z,u) eq 0 then
      return true;
    end if;
  end for;
  return false;
end intrinsic;

intrinsic IsIncludedIrreducibleLF(lf,l,llf::SeqEnum) -> BoolElt
{
 The function iterates IsIncludedIrreducibleLF over each element of
 llf and tells whether lf[l] is included in llf.
}
  for lf2 in llf do
    b:=IsIncludedIrreducibleLF(lf,l,lf2);
    if Type(b) eq MonStgElt then return b; end if;
    if b then return(true); end if;
  end for;
  return(false);
end intrinsic;

// #Section=Set difference
///////////////////////

intrinsic DifferenceLF(~lf1,lf2::Rec)
{
 
    lf1 is a fiber.
    lf2 is an isolated lifting fiber.
 
 
   In return lf1 contains its only components that are not
   included in lf2.
 
 
   If it is detected that lf2 is not an isolated lifting
   fiber then lf1 contains a string in return.
 
}
  IndentPush();
  vprint GeometricSolveVerbose,1: "--> DifferenceLF in degrees ",DegreeLF(lf1)," \\ ",DegreeLF(lf2);
  UP:=Universe(lf1`MinimalPolynomial);
  s:=#lf1`MinimalPolynomial;
  tokeep:={};
  for l:=1 to s do
    b:= IsIncludedIrreducibleLF(lf1,l,lf2);
    if Type(b) eq MonStgElt then lf1:=b; IndentPop(); return; end if;
    if not b then Include(~tokeep,l); end if;
  end for;
  if #tokeep ne s then
    lf1`MinimalPolynomial:=[UP| lf1`MinimalPolynomial[l] : l in tokeep] ;
    lf1`Denominator:=[ UP |lf1`Denominator[l] : l in tokeep ] ;
    lf1`Parametrization:=[ lf1`Parametrization[l] : l in tokeep ] ;
  end if;
  vprint GeometricSolveVerbose,1: "<-- DifferenceLF";
  IndentPop();
end intrinsic;

intrinsic DifferenceLF(~lf,llf::SeqEnum)
{
  The procedure iterates DifferenceLF over each element of llf,
  so that lf only contains its components that are not included
  in llf in return.
}
  for lf1 in llf do
    DifferenceLF(~lf,lf1);
    if Type(lf) eq MonStgElt then return; end if;
  end for;
end intrinsic;

// #Section=Minimization
/////////////////////

intrinsic MinimizeLLF(~llf:RemoveMultipleComponents:=false)
{
 llf is an equidimenstional decomposition that may be redundant.
 In return llf does not contain redundant components.
 It is important that llf[i] contains the components
 of dimension i+1.
 In case of bugs llf may contain a string in return.
}
  IndentPush();
  vprint GeometricSolveVerbose,1: "--> MinimizeLLF";
  n:=#llf-1;
  for i:=n to 1 by -1 do
    dim_i:=[];
    bigger:= &cat[ llf[j] : j in [i+1..n] ];
    for j:=1 to #llf[i] do
      todo:=llf[i][j];
      if RemoveMultipleComponents and IsMultipleLF(todo) then continue; end if;
      DifferenceLF(~todo, dim_i cat bigger);
      if Type(todo) eq MonStgElt then return; end if;
      if not IsEmptyLF(todo) then dim_i cat:= LiftSplitLF(todo); end if;
      delete todo;
    end for;
    llf[i]:=dim_i;
  end for;
  vprint GeometricSolveVerbose,1: "<-- MinimizeLLF";
  IndentPop();
end intrinsic;

// #Section=Intersection
/////////////////////

intrinsic IntersectLF(llf::SeqEnum,f::Rec,h::[Rec]:RemoveMultipleComponents:=false) -> SeqEnum
{
 
    llf, a minimal sequence of sequences lifting fibers
     encoding an equidimensional decomposition of an algebraic
     variety.
    f, a blackbox polynomial.
    h, a sequence of blackbox polynomials.
 
 
   It returns the minimal equidimensional decomposition
   for the intersection of the input variety given by llf 
   with the hypersurface defined by f=0 and outside
   h=0.
 
 
   In case of problems (bug or unlucky choices) a string is returned.
 
}
  _:=Gradient(f); // To be sure it is included in the remember table.
  E:=f`Parent;
  n:=Rank(E);
  answer:= [ [] : i in [1..n+1] ];
  for r:=n+1 to 1 by -1 do
    todo:=[];
    easy:=[];
    for lf in llf[r] do
      tmpZ,tmpNZ:=SplitLF(lf,f);
      if not IsEmptyLF(tmpZ) then
        Append(~tmpZ`LiftingSystem,f);
	delete tmpZ`GenericTrace; // Because tmpZ may become regular
	                          // thanks to the new equation.
        Append(~answer[r],tmpZ);
      end if;
      delete tmpZ;
      if not IsEmptyLF(tmpNZ) then
        Append(~todo,tmpNZ);
      end if;
      delete tmpNZ;
    end for;
    if r eq 1 then continue; end if;
    for i:=1 to #todo do
      tmp:=todo[i];
      if DimensionLF(tmp) gt 0 and not IsEmptyLF(tmp) then
        LiftLastFreeVariableLF(~tmp);
	if Type(tmp) eq MonStgElt then return tmp; end if;
        KroneckerIntersect(~tmp,f
	  :RemoveMultipleComponents:=RemoveMultipleComponents);
        if Type(tmp) eq MonStgElt then return tmp; end if;
        CleanLF(~tmp,h);
        if Type(tmp) eq MonStgElt then return tmp; end if;
        if not IsEmptyLF(tmp) then
          answer[r-1] cat:=tmp;
        end if;
      end if;
      delete tmp;
    end for;
  end for;
  MinimizeLLF(~answer:RemoveMultipleComponents:=RemoveMultipleComponents);
  return answer;
end intrinsic;

