// # 1 "fda.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: lun jun 10 21:18:55 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "fda.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "fda.mtx" 2



debug:=false;
//debug:=true;

/*
 * Fast Deflation Algorithm
 *
 * Creation: G. Lecerf, Sept. 2000.
 * 
 * Integration to Kronecker 0.166beta-3, November 2000.
 * May 2001, GL.
 *   o Revert order of the variables.
 *   o Built demo for thesis.
 * June 2001, GL.
 *   o Integration to 0.166beta-4: now separate file for series (see mseries).
 * December 2001, GL.
 *   o Improved interface and evaluation.
 * January 2001, GL.
 *   o Comments added.
 */

intrinsic FdAlgebra(R::Rng,m::SeqEnum:PrettyPrint:=false, VariableName:="dx") -> AlgAss
{
  FdAlgebra over R and multiplicity sequence m.
}
  AddAttribute(AlgAss,"MultiplicitySequence");
  supp:= MpsSupportProduct(m);
  dsupp:= MpsIdealGradient(supp);
  vprint FdaVerbose,2: "FdAlgebra: cardinal of the support is", #dsupp;
  A:=MultivariatePowerSeriesAlgebra(R,dsupp:
                                     PrettyPrint:=PrettyPrint,
                                     VariableName:=VariableName);
  A`MultiplicitySequence:=m;
  return(A);
end intrinsic;

intrinsic FdChangeRing(A::AlgAss,R) -> AlgAss,.
{
  It returns B,h: h is the coercion map from A to B.
}
  B:= FdAlgebra(R,A`MultiplicitySequence:
                PrettyPrint:=A`PrettyPrint,
                VariableName:=A`VariableName);
  n:= #B`Support;
  Bzero:= B!0;
  h:= function(x)
        y:=Bzero;
        for i:=1 to n do y:=y+R!x[i]*B.i; end for;
        return(y);
      end function;
  return B,h;
end intrinsic;

intrinsic FdChangeRing(A::AlgAss,R,f) -> AlgAss,.
{
  It returns B,h: h is the coercion map from A to B.
}
  B:= FdAlgebra(R,A`MultiplicitySequence:
                PrettyPrint:=A`PrettyPrint,
                VariableName:=A`VariableName);
  n:= #B`Support;
  Bzero:= B!0;
  h:= function(x)
        y:=Bzero;
        for i:=1 to n do y:=y+f(x[i])*B.i; end for;
        return(y);
      end function;
  return B,h;
end intrinsic;

/**************************************************************
*              Change multiplicity sequence                   *
**************************************************************/

intrinsic FdChangeMultiplicitySequence(A::AlgAss,m::SeqEnum)-> AlgAss,.
{
  It returns B,h: h is the coercion map from A to B.
}
  B:= FdAlgebra(BaseRing(A),m:
                PrettyPrint:=A`PrettyPrint,
                VariableName:=A`VariableName);
  n:= MpsRank(B);
  Bzero:= B!0;
  Bsupport:=B`Support;
  ASupport:=A`Support;
  h:= function(x)
        y:=Bzero;
        for i:=1 to #Bsupport do
          j:=Position(ASupport,Bsupport[i]);
          if j eq 0 then continue; end if;
          y:=y+x[j]*B.i; end for;
        return(y);
      end function;
  return B,h;

end intrinsic;

/////////////MISC
intrinsic MyIsUnit(x::RngUPolResElt) -> BoolElt, RngUPolResElt
{
}
  A:=Parent(x);
  UP:= PreimageRing(A);
  q:=Modulus(A);
  d,ix:=Xgcd(UP!x,q);
  b:= IsUnit(d);
  if b then
    return true, A!ix*A!(Coefficient(d,0)^(-1));
  else
    return false, A!d;
  end if;

end intrinsic;

intrinsic MyIsUnit(x::FldPrElt) -> BoolElt, .
{
}
  if x eq 0 then
    return false, x;
  else
    return true, x^(-1);
  end if;
end intrinsic;

intrinsic MyIsUnit(x) -> BoolElt, .
{
  Provides a uniform interface.
  Magma's IsUnit function should be improved this way.
}
  b:= IsUnit(x);
  if b then
    return true, x^(-1);
  else
    return false, x;
  end if;
end intrinsic;

/**************************************************************
*                    Deflation algorithm                      *
**************************************************************/

// Internal Messages
DEFLATE:="DEFLATE";


// # 161 "fda.mtx"















improve_inverse:=function(x,invx)
/*
  One step of Newton's iterator to compute the inverse of an element.
*/
  A:=Parent(x);
  return(invx+invx*( MpsOne(A) - x*invx));
end function;	


evaluate_omega:=function(pre_evaluate,F,trace,vblock,b,Y)
/*
  Evaluate omega on Y.
*/
  eq_idx:=trace[1];   // Index of the input polynomial concerned.
  eq_coeff:=trace[2]; // Coefficient to pick up.
  eq_defl:=trace[3];  // Deflation number.
  vprint FdaVerbose,2: "Evaluation...";
  ans:= Evaluate(F[eq_idx],pre_evaluate(Y));
  ans:= &*[Integers()| Factorial(z) : z in eq_coeff]
                * MpsCoefficient(ans,
                &cat[vblock[z] : z in [1..b-1]],
                eq_coeff);
  ans:= MpsDerivative(ans,vblock[b][1],eq_defl);
  return ans;
end function;
  
first_order_substitution:=function(z,vblock,dxv)
/*
  Substitution dx=dxv, for the variables contained in vblock only and at first order.
*/
  vprint FdaVerbose,2:"First order subsitution...";
  return MpsCoefficient(z,vblock,[ 0 : j in [1..#vblock]])
      + &+[ MpsCoefficient(z,vblock,[ vblock[j] eq k select 1 else 0: j in [1..#vblock]])*dxv[k]
        : k in vblock];
end function;

///////////// MAIN INTERNAL FUNCTIONS

fda_find_block:=procedure(pre_evaluate,F,~Y,~vblock,~ipivot,~trace,~val,~mesg)
/*
  F is a sequence of straight-line polynomials.
  Y is the sequence of the nested coordinates.
  vblock is the sequence of the block of variables.
  ipivot is the sequence of the inverses of the encountered pivots.
  coordinates must be generic enough.  
*/
  IndentPush();
  vprint FdaVerbose,1:"--> fda_find_block entering level",#vblock;
  A:= Universe(Y);
  R:=BaseRing(A);
  m:= A`MultiplicitySequence;
  n:= MpsRank(A);
  fvar:=Setseq({1..n} diff Seqset(&cat vblock)); // free variables
  dvar:=&cat vblock; // dependent variables
  if debug then printf "Input\n"; printf "  Y = "; MpsPrint(Y); printf "  vblock = %o\n",vblock; printf "  ipivot = "; MpsPrint(ipivot); printf "  trace = %o\n",trace; printf "  m = %o\n",m; end if;;
  
  dx:= MpsVariables(A);
  dxidx:= MpsIndexVariables(A);
  S:= A`Support;

  vprint FdaVerbose,2:"Evaluation...";
  Phi:= Evaluate(F,pre_evaluate(Y));
  
  if debug then printf "Phi = "; MpsPrint(Phi); end if;;
  locsupp:= MpsSupportProduct([m[i] : i in dvar]);
  // stores the solution
  dxv:= dx; // free variables
  dxi:= [ ]; // Les variables qui deviennent dependentes.
  if debug then printf "dx[fvar[1]] = "; MpsPrint(dx[fvar[1]]); end if;;
  // The val first derivations have already be examined.
  for j:=0 to val-1 do
    for k:=1 to #Phi do
      Phi[k]:= MpsDerivative(Phi[k],fvar[1]);
    end for;
  end for;
  for j:=val to m[fvar[1]]-1 do
    vprint FdaVerbose,3:"Studying Phi at derivation",j;
    //jacobian_matrix:=[R|];
    for k:=1 to #Phi do
      for e in locsupp do
        if debug then
          printf "Picking up coefficient %o^%o\n",Sprint(dvar),Sprint(e);
        end if;
        equation:= &*[Integers()| Factorial(z) : z in e]
                   *MpsCoefficient(Phi[k],dvar,e);
        if debug then
          printf "Examining equation: "; MpsPrint(equation);
        end if;
        //assert equation[1] eq 0; // doit etre la pendant la trace.
	// sauf avec les flottants.
        // Rcupre la valeur du gradient
        gradient:= A!0;
        for l in fvar do
          //Append(~jacobian_matrix,MpsCoefficient(equation,S[dxidx[l]]));
          gradient:= gradient + MpsCoefficient(equation,S[dxidx[l]])*dxv[l];
        end for;
        for l in fvar do
          dxlidx:= MpsIndexVariable(A,l);
          pivot:= gradient[dxlidx];
          if pivot ne 0 then // Ici il faut scinder
            b,locipivot:= MyIsUnit(pivot);
            if not b then
              vprint FdaVerbose,2:"Find_Block: split!";
              mesg:= locipivot; // Dans ce cas c'est le gcd.
              IndentPop();
              return;
            end if;
            Append(~dxi,l);
            trace[l]:=<k,e,j>; // On stocke l'equation, le coeff et l'ordre de derivation.
            ipivot[l]:= locipivot*MpsOne(A);
            if debug then
              printf "Initialization ipivot[%o] = ",l; MpsPrint(ipivot[l]);
            end if;
            gradient[dxlidx]:=0;
            dxv[l]:= -ipivot[l]*gradient;
            if debug then
              printf "Found %o%o = ",A`VariableName,l ; MpsPrint(dxv[l]);
            end if;
            dxv:= [z+z[dxlidx]*dxv[l]-z[dxlidx]*A.dxlidx : z in dxv];
            break;
          end if;
        end for;
      end for;
      Phi[k]:= MpsDerivative(Phi[k],fvar[1]);
    end for;
    if not IsEmpty(dxi) then val:=j+1; break; end if;
  end for;
  if IsEmpty(dxi) then
    vprint FdaVerbose,2:"fda_find_block: deflate more!";
    val:=m[fvar[1]];
    mesg:=DEFLATE;
  else
    for j in dxi do
      Y:=[ z + MpsDerivative(z,j)*dxv[j] : z in Y];
    end for;
    Append(~vblock,dxi);
    mesg:="";
  end if;
  if debug then printf "Output\n"; printf "  Y = "; MpsPrint(Y); printf "  vblock = %o\n",vblock; printf "  ipivot = "; MpsPrint(ipivot); printf "  trace = %o\n",trace; end if;;
  vprint FdaVerbose,1:"<-- fda_find_block";
  IndentPop();
end procedure;

fda_lift_block:=procedure(pre_evaluate,F,~Y,vblock,b,~ipivot,trace,~mesg)
/*
  Improves the precision wrt the bth block of variables.
  mesg may contain an non invertible pivot in return.
*/
  A:=Universe(Y);
  m:=A`MultiplicitySequence;
  n:=MpsRank(A);
  IndentPush();
  vprint FdaVerbose,1:"--> fda_lift_block level",b;
  if debug then printf "Input\n"; printf "  Y = "; MpsPrint(Y); printf "  vblock = %o\n",vblock; printf "  ipivot = "; MpsPrint(ipivot); printf "  trace = %o\n",trace; printf "  m = %o\n",m; end if;;

  InitializeEvaluation(F);
  // Place to store the solutions.
  dxv:= MpsVariables(A);
  equations:= [evaluate_omega(pre_evaluate,F,trace[l],vblock,b,Y) : l in vblock[b]];  
  for ll:=1 to #vblock[b] do
    l:=vblock[b][ll];
    vprint FdaVerbose,2:"Dealing with variable", l;
    equation:=equations[ll];
    if debug then printf "Dealing with equation   "; MpsPrint(equation);  end if;
    equation:= first_order_substitution(equation,vblock[b],dxv);
    if debug then printf "Equation after subsback "; MpsPrint(equation); end if;
    value:= MpsCoefficient(equation,[l],[0]);
    pivot:= MpsCoefficient(equation,[l],[1]);
    if IsZero(ipivot[l]) then
      tmp0,tmp1:= MyIsUnit(MpsConstantCoefficient(pivot));
      if not tmp0 then
         vprint FdaVerbose,2:"pivot not invertible";
          mesg:= tmp1;
          IndentPop();
          return;
      else
        ipivot[l]:=tmp1*MpsOne(A);
	delete tmp1;
      end if;
    else
      ipivot[l]:= improve_inverse(pivot,ipivot[l]);
    end if;
    if debug then printf "ipivot[l] = "; MpsPrint(ipivot[l]); end if;;
    dxv[l]:= -ipivot[l]*value;
    if debug then
      printf "Gaussian elimination done: %o%o = ",A`VariableName,l; MpsPrint(dxv[l]);
    end if;
    // substitution.
    for k in vblock[b] do
      if k ne l then
        dxv[k]:= MpsCoefficient(dxv[k],[l],[0]) + MpsCoefficient(dxv[k],[l],[1])*dxv[l];
      end if;
    end for;
  end for;
  for j in vblock[b] do
    Y:=[ z+ MpsDerivative(z,j)*dxv[j] : z in Y];
  end for;
  mesg:="";
  if debug then printf "Output\n"; printf "  Y = "; MpsPrint(Y); printf "  vblock = %o\n",vblock; printf "  ipivot = "; MpsPrint(ipivot); printf "  trace = %o\n",trace; end if;;
  vprint FdaVerbose,1:"<-- fda_lift_block";
  IndentPop();
end procedure;

fda_lift_all_blocks:=procedure(pre_evaluate,F,~Y,vblock,~ipivot,trace)
/*
  Improves the precision with respect to all the given vblock's.
*/
  if #vblock eq 0 then return; end if;
  IndentPush();
  vprint FdaVerbose,1:"--> fda_lift_all_blocks with vblock = ",Sprint(vblock);
  for b:=1 to #vblock do
    fda_lift_block(pre_evaluate,F,~Y,vblock,b,~ipivot,trace,~mesg);
    // mesg should be empty.
    assert mesg cmpeq "";
  end for;
  vprint FdaVerbose,1:"<-- fda_lift_all_blocks";
  IndentPop();
end procedure;

fda_find_all_blocks:=procedure(pre_evaluate,F,~Y,~vblock,~ipivot,~trace,~mesg)
/*
  Find all the blocks, the trace and the inverse of the pivots.
*/
  IndentPush();
  vprint FdaVerbose,1:"--> fda_find_all_blocks";

  A:= Universe(Y);
  m:= A`MultiplicitySequence;
  n:= MpsRank(A);
  mesg:="";
  while #&cat vblock lt n do
    fvar:=Setseq({1..n} diff Seqset(&cat vblock)); // free variables
    val:=0; // Contains the valuation estimate
    repeat
      // Don't increase multiplicity at the very beginning
      if not (#fvar eq n and mesg eq "") then
        vprint FdaVerbose,2:"Increasing multiplicity sequence";
        m[fvar[1]]*:=2;  // TODO: how to fix?
        vprint FdaVerbose,2:"New multiplicity sequence:",m;
        A,h:= FdChangeMultiplicitySequence(A,m);
        ipivot := [ h(z) : z in ipivot ];
        Y := [ h(z) : z in Y ];
	// TODO ???
        fda_lift_all_blocks(pre_evaluate,F,~Y,vblock,~ipivot,trace);
	if m[fvar[1]] ge 3 then
          fda_lift_all_blocks(pre_evaluate,F,~Y,vblock,~ipivot,trace);
	end if;
      end if;
      fda_find_block(pre_evaluate,F,~Y,~vblock,~ipivot,~trace,~val,~mesg);
    until mesg cmpne DEFLATE;
    // Update the deflation algebra and continue
    if mesg cmpeq "" then
      m[vblock[#vblock][1]]:=val;
      A,h:= FdChangeMultiplicitySequence(A,m);
      ipivot := [ h(z) : z in ipivot ];
      Y := [ h(z) : z in Y ];
    else     // A splitting is detected, break the loop.
      IndentPop();
      return;
    end if;
  end while;
  vprint FdaVerbose,1:"<-- fda_find_all_block";
  mesg:="";
  IndentPop();
end procedure;

fda_nested_coordinates_with_trace:=procedure(pre_evaluate,F,~Y,vblock,~ipivot,trace,~mesg)
/*
  Compute the nested coordinates and the inverses of the pivots, according
  to the given blocks of variables and the trace.
  The coordinates must be generic enough.
*/
  IndentPush();
  vprint FdaVerbose,1:"--> fda_nested_coordinates_with_trace";

  A:= Universe(Y);
  m:= A`MultiplicitySequence;
  n:= MpsRank(A);

  for b:=1 to #vblock do
    xd:=vblock[b][1]; // The deflation variable.
    vprint FdaVerbose,2:"Deflation variables:",xd;
    while m[xd] le trace[xd][3] do
      m[xd] :=Min(2*m[xd],trace[xd][3]+1);
      vprint FdaVerbose,2:"New multiplicity sequence:",m;
      A,h:= FdChangeMultiplicitySequence(A,m);
      ipivot := [ h(z) : z in ipivot ];
      Y := [ h(z) : z in Y ];
      fda_lift_all_blocks(pre_evaluate,F,~Y,[vblock[l]: l in [1..b-1]],~ipivot,trace);
      if m[xd] ge 3 and m[xd] eq trace[xd][3]+1 then
        fda_lift_all_blocks(pre_evaluate,F,~Y,[vblock[l]: l in [1..b-1]],~ipivot,trace);
      end if;
    end while;
    fda_lift_block(pre_evaluate,F,~Y,vblock,b,~ipivot,trace,~mesg);
    if Type(mesg) ne MonStgElt then IndentPop(); mesg:="Failed"; return; end if;
    if mesg cmpne "" then IndentPop(); return; end if;
  end for;
  vprint FdaVerbose,1:"<-- fda_nested_coordinates_with_trace";
  mesg:="";
  IndentPop();
end procedure;

fda_nested_coordinates:=procedure(pre_evaluate,F,~Y,~vblock,~ipivot,~trace,~mesg)
/*
*/
  IndentPush();
  vprint FdaVerbose,1:"--> fda_nested_coordinates";
  A:= Universe(Y);
  n:= MpsRank(A);
  FDvars:= MpsVariables(A);
  Y:=[A| Y[i]+FDvars[i] : i in [1..n]];
  ipivot:=[A| A!0: i in [1..n] ];
  if not assigned vblock then
    vblock:=[];
    trace:=[ <0,[ 0 ],0> : i in [1..n] ];
    fda_find_all_blocks(pre_evaluate,F,~Y,~vblock,~ipivot,~trace,~mesg);
  else
    fda_nested_coordinates_with_trace(pre_evaluate,F,~Y,vblock,~ipivot,trace,~mesg);
  end if;
  vprint FdaVerbose,1:"<-- fda_nested_coordinates";
  IndentPop();
end procedure;

/*******************************************************************************
                                  Exports
*******************************************************************************/

intrinsic FdTrace() -> Cat
{
  Returns the recformat used for storing traces.
}
  return( recformat< BlocksOfVariables, Trace> );
end intrinsic;

intrinsic FdIsSmoothTrace(trace) -> BoolElt
{
 Return true if no true deflation occurs.
}
   return &and [ z[3] eq 0: z in trace`Trace];
end intrinsic;

intrinsic FdNestedCoordinates(pre_evaluate,F,~Y,~trace,~ipivot,~mesg)
{
  
   F, a sequence of straight-line polynomials.
   Y, a sequence of ring elements.
   trace, ipivot and mesg are not assigned.
  
  
   Y contains the sequence of the nested coordinates.
   trace contains the trace of the deflation process.
   mesg may contain an element of the base ring .
  
  
   If the coordinates are not generic enough (lack of Weierstrass postion) the computation may not finish.
   If the root Y is not isolated then the computation may not finish.
   If the base ring is not a field, mesg can contain an element than in not zero and
    not invertible: this element prevents from finishing the computation.
  
}
  fda_nested_coordinates(pre_evaluate,F,~Y,~tmpvblock,~ipivot,~tmptrace,~mesg);
  trace:=rec <FdTrace() |
    BlocksOfVariables:=tmpvblock,
    Trace:=tmptrace
  >;
end intrinsic;

intrinsic FdNestedCoordinatesWithTrace(pre_evaluate,F,~Y,trace,~ipivot,~mesg)
{
  
   F, a sequence of straight-line polynomials.
   Y, a sequence of ring elements.
   trace, the trace of the deflation.
   ipivot and mesg are not assigned.
  
  
   Y contains the sequence of the nested coordinates.
   ipivot contains the inverses of some elements required by the deflation.
   mesg may contain an element of the base ring or a string.
  
  
   The coordinates are assumed to be generic enough.
   Y is assumed to be an isolated root.
   If the computation can not be performed according to the trace then mesg
    contains a non invertible element in return.
  
}
  fda_nested_coordinates(pre_evaluate,F,~Y,~trace`BlocksOfVariables,~ipivot,~trace`Trace,~mesg);
end intrinsic;

intrinsic FdLiftNestedCoordinates(pre_evaluate,F,~Y,trace,~ipivot)
{
  
   F, a sequence of straight-line polynomials.
   Y, the nested coordinates.
   trace, the trace of the deflation.
   ipivot, the sequence of the inverses of the pivots.
  
  
   Y contains the nested coordinates with precision twice as much as in input.
   ipivot contains the refined inverses.
  
  
   The coordinates are assumed to be generic enough.
   Y, trace and ipivot are assumed to come from one the deflation
    fonctions that computes them.
  
}
  fda_lift_all_blocks(pre_evaluate,F,~Y,trace`BlocksOfVariables,~ipivot,trace`Trace);
end intrinsic;
