// # 1 "fiber.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 18:36:56 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "fiber.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "fiber.mtx" 2

import "random.m": random_RandomInvertibleMatrix, random_RandomInvertibleUpperTriangularMatrix, random_RandomVector, random_RandomPrimeNumber, random_RandomSequence;// #import random.m

//#define DEBUG

// Lifting fibers, Hensel Lifting, Blackbox polynomials, Top

// #Package=Lifting fiber

// #Public

// This package provides functions for manipulating lifting
// libers. Lifting fibers encode equidimensional algebraic varieties.@*
// 
// Before all, a
// BlackboxPolynomialAlgebra domain must have been built this way:@*
//  E,x[1],...,x[n]:= BlackboxPolynomialAlgebra(F,n);
// @*
// where F is either the field of the rational numbers or
// a prime field. Note that it may work with other fields as soon as
// the characteristic is big enough and a factorization
// function exists in F[T].
// 
// A fiber encoding an r-equidimensional algebraic
// variety V is the following record:
//  
//  ResolutionField, a field K over which the resolution
//   is defined.
//   There must exist a coercion from F to K.
//  LiftingSystem, a lifting system F for V, it
//   is a sequence of blackbox polynomials.
//  PrimitiveElement, a linear form u in the
//   x[i]'s over K, it is a multivariate polynomial
//   in variables.
//  MinimalPolynomial, a sequence q of monic
//   univariate polynomials in K[T], of size s.
//  Denominator, a sequence p of univariate
//   polynomials in K[T] of size s.
//  Parametrization, a sequence w of size s of
//   sequences of elements of K[T] of size n-r.
//  MagicPoint, a sequence P of elements in K
//   of size r.
//  ChangeOfVariables, a record <LinearPart,AffinePart>,
//   LinearPart
//   is an invertible square matrix M of size and
//   AffinePart is a vector b of size n.
//   Booth have entries in K.
//  IsMultiple, boolean flag telling if V is
//   multiple as a solution of F=0.
//  GenericTrace, generic trace of the deflation process.
//  ParentBbp points to E.
//  itemize
//
// 
// Let y be the new variables defined by x=M.y+b,
// the following properties hold:
//  
//  V is a subvariety of the set of roots of F=0.
//  The variables y are in projective Noether position
//   with respect to V.
//   y[1]=P[1],,y[r]=P[r] defines a finite fiber
//   V' of V.
//  The primitive element u separates the points of this
//   fiber V'.
//  The elements of q are monic and squarefree.
//  The fiber V' is the union of the set of points described
//   by the parametrizations:@*
//  q[l](T)=0, p[l](T)y[r+1]=v[l](T), , p[l](T)y[n]=v[n](T), for l in [1,,s].
//  itemize
// We say the fiber is isolated if V is an isolated subvariety 
// in the set of roots of the system F=0.
// We say that the magic point is a lifting point if it satisfies the smoothess
// hypothesis of the fast deflation algorithm.
// The parametrization of is in the Kronecker presentation
// if p[l]=q'[l] for all l and is in the Shape-lemma if
// p[l]=1 for all l.

///////////////////////////////////
// #Section=Creation of lifting fibers
///////////////////////////////////

intrinsic LiftingFiber() -> Cat
{
  
     It returns the record format used to store lifting fibers.
  
}
  return(recformat<
    LiftingSystem,
    ResolutionField,
    PrimitiveElement,
    MinimalPolynomial,
    IsMultiple,
    GenericTrace,
    Denominator,
    Parametrization,
    MagicPoint,
    ChangeOfVariables,
    ParentBbp,
    Stamp // For internal use only.
  >);
end intrinsic;

intrinsic WholeSpaceLF(E,K:GenericLinearChangeOfVariables:=true) -> .
{
  
     E, a domain created by BlackboxPolynomialAlgebra.
     K, a field.
  
  
    It returns a lifting fiber of the ambient space with
    resolution field K.
  
  
    If GenericLinearChangeOfVariables is set then
     a generic affine change of the coordinates is performed.
  
}
  n:=Rank(E);
  UP<T>:=PolynomialAlgebra(K);
  MP:=PolynomialAlgebra(K,n);
  lf:=rec<LiftingFiber()|
    LiftingSystem:=[],
    ResolutionField:=K,
    PrimitiveElement:=MP!0,
    MinimalPolynomial:=[UP.1],
    IsMultiple:= false,
    Denominator:=[UP!1],
    Parametrization:=[[UP|]],
    MagicPoint:=[K| 0 : z in [1..n]],
    ChangeOfVariables:=rec<recformat<LinearPart,AffinePart>|
       LinearPart:=ScalarMatrix(n,K!1), AffinePart:=Vector([K| K!0: i in [1..n]])>,
    ParentBbp:=E,
    GenericTrace:= rec< FdTrace() | BlocksOfVariables:=[], Trace:=[] >,
    Stamp:= "" >;
  if GenericLinearChangeOfVariables then
    lf`ChangeOfVariables`LinearPart:= random_RandomInvertibleUpperTriangularMatrix(K,n);
    lf`ChangeOfVariables`AffinePart:= random_RandomVector(K,n);
  end if;
  return(lf);
end intrinsic;

// #Section=Getting properties
///////////////////////////

// lf denotes a lifting fiber and llf a sequence of lifting fibers.

intrinsic CodimensionLF(lf::Rec) -> RngIntElt
{
 
   It returns the codimension of the variety encoded by lf.
 
}
  return(NumberOfVariablesLF(lf)-DimensionLF(lf));
end intrinsic;

intrinsic CodimensionLF(llf::[]) -> RngIntElt
{
 
   It returns the minimum of the codimensions of the elements of llf.
 
 
   llf must not be empty.
 
}
  require not IsEmpty(llf): "Argument is empty";
  return(Min([CodimensionLF(z): z in llf]));
end intrinsic;


intrinsic DegreeLF(lf::Rec) -> RngIntElt
{
  
    It returns the degree of the variety represented by lf.
  
}
  return(&+[Integers()| Degree(z) : z in lf`MinimalPolynomial]);
end intrinsic;

intrinsic DegreeLF(llf::[]) -> RngIntElt
{
  
    It returns the sum of the degrees of the element of llf.
  
}
  return(&+[Integers()| DegreeLF(z) : z in llf]);
end intrinsic;

intrinsic DimensionLF(lf) -> RngIntElt
{
 See RankLF.
}
  return(RankLF(lf));
end intrinsic;

intrinsic HasKroneckerParametrizationLF(lf::Rec)-> BoolElt
{
  
     It returns true if lf has a Kronecker parametrization.
  
}
  // empty and whole space cases
  if DegreeLF(lf) le 0 then return(true); end if;
  if &or[Booleans() | z eq 0: z in lf`Denominator] then return(false); end if;
  for i:=1 to #lf`MinimalPolynomial do
    qp:=Derivative(lf`MinimalPolynomial[i]);
    p:=lf`Denominator[i];
    lcqp:=LeadingCoefficient(qp);
    lcp:=LeadingCoefficient(p);
    if p*(lcqp/lcp) ne qp then return(false); end if;
  end for;
  return true;
end intrinsic;

intrinsic HasKroneckerParametrizationLF(llf::[])-> BoolElt
{
  
     It returns true if all the elements of llf have
     a Kronecker parametrization.
  
}
  for lf in llf do
    if not HasKroneckerParametrizationLF(lf) then return(false); end if;
  end for;
  return(true);
end intrinsic;

intrinsic HasShapeLemmaParametrizationLF(lf::Rec)-> BoolElt
{
  
     It returns true if lf has a shape lemma parametrization.
  
}
  F:=lf`ResolutionField;
  for p in lf`Denominator do
    if Degree(p) gt 0 then return(false); end if;
    p0:=Coefficient(p,0);
    if Type(F) eq RngSerLaur then // special case for series
      if not IsWeaklyEqual(F!1, p0) then return false; end if;
    else
      if not IsOne(p0) then return false; end if;
    end if;
   end for;
  return(true);
end intrinsic;

intrinsic HasShapeLemmaParametrizationLF(llf::[])-> BoolElt
{
  
     It returns true if all the elements of lf have
     a shape lemma parametrization.
  
}
  for lf in llf do
    if not HasShapeLemmaParametrizationLF(lf) then return(false); end if;
  end for;
  return(true);
end intrinsic;

intrinsic IsEmptyLF(lf::Rec) -> BoolElt
{
  
    It returns a boolean telling whether the variety represented by lf is empty or not.
  
}
  return(DegreeLF(lf) eq 0);
end intrinsic;

intrinsic IsEmptyLF(llf::[]) -> BoolElt
{
  
    It returns a boolean telling whether all the elements of llf
    represent the empty variety or not.
  
}
  return(&and[Booleans()| IsEmptyLF(z): z in llf]);
end intrinsic;

intrinsic IsMultipleLF(lf::Rec) -> BoolElt
{
  
    It returns a boolean telling whether lf
    is a multiple component or not.
  
}
  return(lf`IsMultiple);
end intrinsic;

intrinsic IsWholeSpaceLF(lf::Rec) -> BoolElt
{
 
   It returns whether lf represents the ambient space or not.
 
}
  return(DimensionLF(lf) eq NumberOfVariablesLF(lf));
end intrinsic;

intrinsic IsWholeSpaceLF(llf::[]) -> BoolElt
{
 
   It returns whether at least one of the elements of llf
   represents the ambient space or not.
 
}
  return(&or[IsWholeSpaceLF(z) : z in llf]);
end intrinsic;

intrinsic NumberOfVariablesLF(lf::Rec) -> RngIntElt
{
  
    It returns the dimension of the ambient space in which lf lives.
  
}
  return(Rank(ParentBbpLF(lf)));
end intrinsic;

intrinsic NumberOfFactorsLF(lf::Rec) -> RngIntElt
{
  
    It returns the number of factors lf, that is
    the cardinal of lf`MinimalPolynomial.
  
}
  return(#lf`MinimalPolynomial);
end intrinsic;

intrinsic ParentBbpLF(lf::Rec) -> Cat
{
 
   It returns the Multivariate Polynomial Algebra associated to
   the blackbox polynomial domain of the lifting system of lf.
 
}
  return(lf`ParentBbp);
end intrinsic;

intrinsic RankLF(lf::Rec) -> RngIntElt
{
 
   It returns the dimension of the variety encoded by lf.
 
}
  return(#lf`MagicPoint);
end intrinsic;

intrinsic RankLF(llf::[]) -> RngIntElt
{
 
   It returns the maximum of the dimensions of the elements of llf.
 
 
   llf must not be empty.
 
}
  require not IsEmpty(llf): "Argument is empty";
  return(Max([RankLF(z): z in llf]));
end intrinsic;

// #Section=Basic operations
/////////////////////////

// lf denotes a lifting fiber and llf a sequence
// of lifting fibers.

intrinsic ChangeResolutionFieldLF(~lf::Rec,K)
{
  
      lf, a LiftingFiber.
      K, a field.
  
  
    If there exists a coercion from lf`ResolutionField
    to K,
    then it modifies lf to be a resolution over K.
  
  
    There must exist a coercion from lf`ResolutionField
    to K.
  
}
  lf`ResolutionField:=K;
  UP:=Universe(lf`MinimalPolynomial);
  UP:=ChangeRing(UP,K);
  ChangeUniverse(~lf`MinimalPolynomial,UP);
  ChangeUniverse(~lf`Denominator,UP);
  lf`Parametrization:=[ChangeUniverse(z,UP): z in lf`Parametrization];
  lf`MagicPoint:=[K| K!z : z in lf`MagicPoint];
  MyChangeRing(~lf`ChangeOfVariables`LinearPart,K);
  MyChangeRing(~lf`ChangeOfVariables`AffinePart,K);
  ChangeRing(~lf`PrimitiveElement,K);
end intrinsic;

intrinsic ChangeResolutionFieldLF(~llf::[],K)
{
 It iterates ChangeResolutionFieldLF over llf.
}
  for i:=1 to #llf do
    ChangeResolutionFieldLF(~llf[i],K);
  end for;
end intrinsic;

intrinsic ChangeResolutionFieldLF(~lf::Rec,K,f)
{
  
      lf, a LiftingFiber.
      K, a field.
      f, a homomorphism from lf`ResolutionField
        to K.
  
  
    Applies f on lf`ResolutionField to coerce it 
    to K via f, so that in return lf is a resolution over
    K.
  
}

  lf`ResolutionField:=K;
  UP:=Universe(lf`MinimalPolynomial);
  UPK<T>,F:=ChangeRing(UP,K,f);
  ChangeUniverse(~lf`MinimalPolynomial,UPK,F);
  ChangeUniverse(~lf`Denominator,UPK,F);
  lf`Parametrization:=[ChangeUniverse(z,UPK,F): z in lf`Parametrization];
  lf`MagicPoint:=[K| f(z) : z in lf`MagicPoint];
  MyChangeRing(~lf`ChangeOfVariables`LinearPart,K,f);
  MyChangeRing(~lf`ChangeOfVariables`AffinePart,K,f);
  ChangeRing(~lf`PrimitiveElement,K,f);
end intrinsic;

intrinsic ChangeResolutionFieldLF(~llf::[],K,f)
{
 It iterates ChangeResolutionFieldLF over llf.
}
  for i:=1 to #llf do
    ChangeResolutionFieldLF(~llf[i],K,f);
  end for;
end intrinsic;

intrinsic MakeKroneckerParametrizationLF(~lf::Rec)
{
  
     lf is changed to a Kronecker parametrization.
  
}
  if HasKroneckerParametrizationLF(lf) then return; end if;
  if not HasShapeLemmaParametrizationLF(lf) then
    MakeShapeLemmaParametrizationLF(~lf);
  end if;
  UP<T>:=Universe(lf`MinimalPolynomial);
  lf`Denominator:=[ UP| Derivative(q): q in lf`MinimalPolynomial];
  lf`Parametrization:=[ [UP| z*lf`Denominator[i] mod lf`MinimalPolynomial[i]
            : z in lf`Parametrization[i]]: i in [1..#lf`MinimalPolynomial]];
end intrinsic;

intrinsic MakeKroneckerParametrizationLF(~llf::[])
{
  It iterates MakeKroneckerParametrizationLF over each element
  of llf.
}
  for i:=1 to #llf do
    MakeKroneckerParametrizationLF(~llf[i]);
  end for;
end intrinsic;

intrinsic MakeMonicLF(~lf::Rec)
{
  
     It makes lf`MinimalPolynomial monic.
  
}
  for i:=1 to #lf`MinimalPolynomial do
    lc:=LeadingCoefficient(lf`MinimalPolynomial[i]);
    lf`MinimalPolynomial[i]:=lf`MinimalPolynomial[i]/lc;
  end for;
end intrinsic;

intrinsic MakeMonicLF(~llf::[])
{
  It iterates MakeMonicLF over each element of llf.
}
  for i:=1 to #llf do
    MakeMonicLF(~llf[i]);
  end for;
end intrinsic;

intrinsic MakeShapeLemmaParametrizationLF(~lf::Rec)
{
  
     It changes the parametrization of lf into a Shape Lemma form.
  
  
     If lf`Denominator
     is not invertible modulo lf`MinimalPolynomial.
     then lf contains a string in return.
  
}
  UP<T>:=Universe(lf`MinimalPolynomial);
  for l:=1 to #lf`MinimalPolynomial do
    q:=lf`MinimalPolynomial[l];
    if IsZero(q) then continue; end if;
    p:=lf`Denominator[l];
    multfact, ip := Xgcd(p,q);
    if Degree(multfact) ge 1 and Type(multfact) ne RngSerLaur then
      if multfact ne 1 then
          lf:="Failed";
	  return;
      end if;
    end if;
    A:=quo<UP|q>;
    ip:=A!ip;
    lf`Parametrization[l]:=[UP| UP!(A!z*ip) : z in lf`Parametrization[l]];
    lf`Denominator[l]:=UP!1;
  end for;
end intrinsic;

intrinsic MakeShapeLemmaParametrizationLF(~llf::[])
{
  It iterates MakeShapeLemmaParametrizationLF over each element
  of llf.
}
  for i:=1 to #llf do
    lf:=llf[i];
    MakeShapeLemmaParametrizationLF(~lf);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    llf[i]:=lf;
  end for;
end intrinsic;

// #Section=Changes of fibers
//////////////////////////

intrinsic ChangeAlgebraicVariablesLF(~lf::Rec,N,c)
{
  
      lf, a LiftingFiber
       of dimension r in a n-dimensional space.
      N, a square matrix of size n-r with rational numbers
       entries.
      c, a n-r-vector with rational numbers entries.
  
  
     Let M and b be lf`ChangeOfVariables, such that
     x=My+b. The procedure changes the coordinates of
     lf in the following way: M:=M.(Idr|N) and
     b:=M.(0|c) + b. This change of
     algebraic variables is applied in consequence to the
     parametrization and the primitive element so that lf remains
     consistent.
  
  
     The procedure raises an error if N is not invertible.
  
}

  r:=RankLF(lf);
  n:=NumberOfVariablesLF(lf);
  M:=lf`ChangeOfVariables`LinearPart;
  b:=lf`ChangeOfVariables`AffinePart;
  F:=BaseRing(M);
  
  require Type(N) eq AlgMatElt : "ChangeAlgebraicVariables: second argument has wrong type";
  require Ncols(N) eq n-r : "ChangeAlgebraicVariables: second argument has wrong size";
  require Dimension(Parent(c)) eq n-r : "ChangeAlgebraicVariables: third argument has wrong size";

  bigN:=InsertBlock(ScalarMatrix(F,n,1),N,r+1,r+1);
  bigC:=Vector([ F| 0 : i in [1..r]] cat Eltseq(c));

  lf`ChangeOfVariables`LinearPart:=M*bigN;
  lf`ChangeOfVariables`AffinePart:=bigC*Transpose(M)+b;

  invBigN:=bigN^(-1);
  F:=lf`ResolutionField;
  u:=lf`PrimitiveElement;

  for l:=1 to #lf`MinimalPolynomial do
    v:=lf`Parametrization[l];
    p:=lf`Denominator[l];   
    lf`Parametrization[l]:=[ &+ [ F!(F!invBigN[i+r,j+r])*(v[j]-F!(c[j])*p) : j in [1..n-r]]: i in [1..n-r]];
  end for;
  MP:=Parent(u);
  lf`PrimitiveElement:= Evaluate(u,[MP| &+[MP|bigN[i,j]*MP.j : j in [1..n]] + bigC[i] : i in [1..n]]);
end intrinsic;

intrinsic ChangeAlgebraicVariablesLF(~llf::[],N,c)
{
 It applies ChangeAlgebraicVariablesLF to each element of llf.
}
  for i:=1 to #llf do
    ChangeAlgebraicVariablesLF(~llf[i],N,c);
  end for;
end intrinsic;

intrinsic ChangeBackAlgebraicVariablesLF(~lf::Rec)
{
  
      lf, a LiftingFiber.
  
  
    In return lf has the identity for the part of the
    change of variables corresponding to the algebraic variables.
  
}
  n:=NumberOfVariablesLF(lf);
  r:=RankLF(lf);
  M0:=lf`ChangeOfVariables`LinearPart;
  M:=Submatrix(M0,r+1,r+1,n-r,n-r);
  b:=lf`ChangeOfVariables`AffinePart;
  b:=Vector(BaseRing(Parent(b)),n-r,[b[i] : i in [r+1..n]]);
  ChangeAlgebraicVariablesLF(~lf,M^(-1),-b*Transpose(M^(-1)));
end intrinsic;

intrinsic ChangeBackAlgebraicVariablesLF(~llf::[])
{
 It applies ChangeBackAlgebraicVariablesLF to each
 element of llf.
}
  for i:=1 to #llf do
    ChangeBackAlgebraicVariablesLF(~llf[i]);
  end for;
end intrinsic;

intrinsic ChangeBackFreeVariablesLF(~lf::Rec)
{
  
      lf, a LiftingFiber.
  
  
    In return lf has the identity in the part of its
    change of variables corresponding to the free variables.
  
}
  n:=NumberOfVariablesLF(lf);
  r:=RankLF(lf);
  M:=lf`ChangeOfVariables`LinearPart;
  MA:=Parent(M);
  m:=MA!1;
  for i:=1 to r do for j:=1 to r do
    m[i,j]:=M[i,j];
  end for; end for;
  lf`ChangeOfVariables`LinearPart:=M*m^(-1);
  mp:=lf`MagicPoint;
  lf`MagicPoint:=[ Universe(mp) |&+[m[i,j]*mp[j] : j in [1..r]] : i in [1..r]];
end intrinsic;

intrinsic ChangeBackFreeVariablesLF(~llf::[])
{
 It iterates ChangeBackFreeVariablesLF on each element of llf.
}
  for i:=1 to #llf do
    ChangeBackFreeVariablesLF(~llf[i]);
  end for;
end intrinsic;

intrinsic ChangePrimitiveElementLF(~lf::Rec,u)
{
  
      lf, a LiftingFiber
       of dimension r in a n-dimensional space.
      u, a linear form given as a multivariate polynomial.
  
  
     It Changes lf`PrimitiveElement to u,
     if it is actually a primitive element.
  
  
     The procedure returns "Bad primitive element"
     in lf if u is not a primitive element.
  
}

  IndentPush();
  vprint FiberVerbose,1:"--> ChangePrimitiveElementLF";
  begintime:=Cputime();

  if IsWholeSpaceLF(lf) then IndentPop(); return; end if;

  MakeShapeLemmaParametrizationLF(~lf);

  UP<T>:=Universe(lf`MinimalPolynomial);
  F:=BaseRing(UP);

  new_MinimalPolynomial:=[UP|];
  new_Parametrization:=[];
  dim:=DimensionLF(lf);
  
  for l:=1 to #lf`MinimalPolynomial do
    q:=lf`MinimalPolynomial[l];
    v:=lf`Parametrization[l];
    rz:=[Universe(v)| 0 : i in [1..dim]];

    delta:=Degree(q);
    I:=[F| j : j in [1..delta+2]];
    Q:=[F| 0 : j in [1..delta+2]];
    A:=[F| 0 : j in [1..delta+2]];
    B:=[F| 0 : j in [1..delta+2]];

    tmp:=UP!Evaluate(u, rz cat v);
    for j:=1 to delta+2 do
      r,sr:= SubResultant(q,UP!j-tmp);
      Q[j]:= F!r;
      A[j]:= Coefficient(sr,1);
      B[j]:= -Coefficient(sr,0);
    end for;
    Q:=Interpolation(I,Q);
    A:=Interpolation(I,A);
    B:=Interpolation(I,B);
    if Degree(Q) ne delta or Discriminant(Q) eq 0
      then lf:="Bad primitive element"; IndentPop(); return;
    end if;

    for ll:=1 to l-1 do
      if Degree(Gcd(new_MinimalPolynomial[ll],Q)) ne 0
      then lf:="Bad primitive element"; IndentPop(); return;
      end if;
    end for;
    
    // Now u is really a primitive element.
    // It remains to deduce the new parametrization.
    // Q(u)=0, A(u)T=B(u)

    // invert A modulo Q
    m,iA:=Xgcd(A,Q);
    assert m eq 1;
    QUO<a>:=quo<UP|Q>;
    B:=QUO!(iA*B);
    // TODO << Baby step/Giant Step
    powers_of_T:= [QUO| i eq 0 select 1 else Self(i)*B : i in [0..delta]];
    powers_of_T:= [UP| UP!z : z in powers_of_T];

    subs:=function(P)
      return &+[UP|Coefficient(P,i)*powers_of_T[i+1] : i in [0..delta]];
    end function;

    Append(~new_MinimalPolynomial,Q);
    Append(~new_Parametrization,[ UP |subs(z) : z in v]);

  end for;

  lf`MinimalPolynomial:= new_MinimalPolynomial;
  lf`Denominator:=[UP| 1 : l in [1..#new_MinimalPolynomial]];
  lf`Parametrization:= new_Parametrization;
  lf`PrimitiveElement:=u;
  //assert VerifyLF(lf);  
  vprint FiberVerbose,1:"<-- ChangePrimitiveElementLF, elapsed time: ", Cputime()-begintime;
  IndentPop();
end intrinsic;

intrinsic ChangePrimitiveElementLF(~llf::[],u)
{
 It applies ChangePrimitiveElementLF to each element of llf.
}
  for i:=1 to #llf do
    lf:=llf[i]; ChangePrimitiveElementLF(~lf,u);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    llf[i]:=lf;
  end for;
end intrinsic;

intrinsic TranslateKthFreeVariableToZeroLF(~lf::Rec,k)
{
  
     lf, a LiftingFiber.
  
  
    Let y[k] be the kth free variable of lf,
    the procedure modifies
    lf`ChangeOfVariables replacing y[k] by 
    y[k]+lf`MagicPoint[k].
    Then lf`MagicPoint[k] is set to 0.
  
}
  if k gt RankLF(lf) then error "k greater than dimension"; end if;
  for i:=1 to NumberOfVariablesLF(lf) do
    lf`ChangeOfVariables`AffinePart[i]+:=
      lf`ChangeOfVariables`LinearPart[i,k]*lf`MagicPoint[k];
  end for;
  lf`MagicPoint[k]:=0;
end intrinsic;

intrinsic TranslateKthFreeVariableToZeroLF(~llf::[],k)
{
 It iterates TranslateKthFreeVariableToZeroLF on each
 element of llf.
}
  for i:=1 to #llf do
    lf:=llf[i]; TranslateKthFreeVariableToZeroLF(~lf,k); llf[i]:=lf;
  end for;
end intrinsic;

intrinsic TranslateAllFreeVariablesToZeroLF(~lf::Rec)
{
 It applies TranslateKthFreeVariableToZeroLF to all the
 free variables of lf.
}
  for x in [1..RankLF(lf)] do
    TranslateKthFreeVariableToZeroLF(~lf,x);
  end for;
end intrinsic;

intrinsic TranslateAllFreeVariablesToZeroLF(~llf::[])
{
 It applies TranslateAllFreeVariablesToZeroLF to each element
 of llf.
}
  for i:=1 to #llf do
    lf:=llf[i]; TranslateAllFreeVariablesToZeroLF(~lf); llf[i]:=lf;
  end for;
end intrinsic;

intrinsic TranslateLastFreeVariableToZeroLF(~lf::Rec)
{
 It applies TranslateKthFreeVariableToZeroLF to all the
 last free variables of lf.
}
  TranslateKthFreeVariableToZeroLF(~lf,RankLF(lf));
end intrinsic;

intrinsic TranslateLastFreeVariableToZeroLF(~llf::[])
{
 It applies TranslateLastFreeVariableToZeroLF to each
 element of llf.
}
  for i:=1 to #llf do
    lf:=llf[i]; TranslateLastFreeVariableToZeroLF(~lf); llf[i]:=lf;
  end for;
end intrinsic;


// #Section=Evaluation
///////////////////

// lf denotes a lifting fiber and llf a sequence of lifting fibers.

intrinsic EvaluateLF(lf::Rec,f::Rec,x) -> .
{
}
  n:=NumberOfVariablesLF(lf);
  require #x eq n :"third argument has bad cardinality";
//  require f`Parent eq ParentBbpLF(lf) : "second argument has not the right Parent";
  return(EvaluateLF(lf,[f],x)[1]);
end intrinsic;

// TODO: suppress these functions.
intrinsic EvaluateLF(lf::Rec,f::[Rec],x::[]) -> .
{
  
     lf, a LiftingFiber.
     f, a sequence of black box polynomials.
     x, a sequence of values.
  
  
    It returns the sequence of the values of f
    evaluated on the point x.
  
  
     #x must be equal to NumberOfVariablesLF(lf).
     BaseRing(ParentBbpLF(lf)) must be coercible to Universe(x).
  
}
  n:=NumberOfVariablesLF(lf);
  require #x eq n :"third argument has bad cardinality";
  R:=Universe(x);
  if not IsEmpty(f) then
//    require Universe(f) eq ParentBbpLF(lf) : "second argument has not the right Parent";
  else return([Universe(x)|]);
  end if;

  // Special case we optimize
  if Type(R) eq RngUPolRes then
    if Degree(Modulus(R)) eq 1 then
      F:=BaseRing(R);
      xx:=[F | Coefficient(z,0) : z in x];
      yy:=[ &+[R| R!lf`ChangeOfVariables`LinearPart[i,j]*xx[j] : j in [1..n]] +
        R!lf`ChangeOfVariables`AffinePart[i] : i in [1..n]];
      return(ChangeUniverse(Evaluate(f,yy),R));
    end if;
  end if;

  y:=[ &+[R| R!lf`ChangeOfVariables`LinearPart[i,j]*x[j] : j in [1..n]] +
    R!lf`ChangeOfVariables`AffinePart[i] : i in [1..n]];
  return(Evaluate(f,y));
end intrinsic;

intrinsic VerifyLF(lf::Rec:Strategy:="Probabilistic") -> BoolElt
{
 
   It returns true iff the parametrization of lf
   satisfies its lifting system modulo its MinimalPolynomial.
 
 
   Strategy, string:
      
        "", the verification is perfomed over
          lf`ResolutionField.
        "Probabilistic", the verification is probabilistic.
          Computations are done modulo a random prime number if the
          resolution field is the field of the rational numbers.
          If the resolution field is a rational function field, the variables
          are specialized at random.
      itemize
 
}
  if Type(lf) eq MonStgElt then return(false); end if;
  F:=lf`ResolutionField;

  if Strategy eq "Probabilistic" then
    if Type(F) eq FldRat then
      F:=GF(random_RandomPrimeNumber());
      loclf:=lf;
      ChangeResolutionFieldLF(~loclf,F);
      return(VerifyLF(loclf));
    elif Type(F!1) eq FldFunRatMElt then
      K:=BaseRing(F);
      point:=ChangeUniverse(random_RandomSequence(K,Rank(F)),K);
      f:=hom<F->K | x:-> Evaluate(Numerator(x),point)/Evaluate(Denominator(x),point)>;
      loclf:=lf;
      ChangeResolutionFieldLF(~loclf,K,f);
      return(VerifyLF(loclf));
    elif Type(F!1) eq FldFunRatUElt then
      K:=BaseRing(F);
      point:=ChangeUniverse(random_RandomSequence(K,Rank(F)),K);
      point:=point[1];
      f:=hom<F->K | x:-> Evaluate(Numerator(x),point)/Evaluate(Denominator(x),point)>;
      loclf:=lf;
      ChangeResolutionFieldLF(~loclf,K,f);
      return(VerifyLF(loclf));
    else return(VerifyLF(lf:Strategy:=""));
    end if;
  end if;

  UP:=Universe(lf`MinimalPolynomial);

  for l:=1 to #lf`MinimalPolynomial do

    q:=lf`MinimalPolynomial[l];

    if q eq 0 and #lf`LiftingSystem gt 0 then return(false); end if;
    if q eq 0 and #lf`LiftingSystem eq 0 then continue; end if;

    p:=lf`Denominator[l];

    A<a>:=quo<UP|q>;
    h:=hom<UP -> A |a>;
    w:=[h(z):z in lf`Parametrization[l]];

    if p eq 1 then
      ip:=1;
    else
      multfact,ip:=Xgcd(p,q);
      if not IsUnit(multfact) then return false; end if;
      ip:=ip/multfact;
    end if;

    ip:=h(ip);
    v:=[z*ip : z in w];

    point:=ChangeUniverse(lf`MagicPoint,A) cat v;

    if Evaluate(lf`PrimitiveElement,point) ne a then return(false); end if;

    if #lf`LiftingSystem gt 0 then
      InitializeEvaluation(lf`LiftingSystem);
    end if;

    for f in lf`LiftingSystem do
        if EvaluateLF(lf,f,point) ne 0 then
          return(false);
        end if;
    end for;
  end for;
  return(true);
end intrinsic;

intrinsic VerifyLF(llf::[]:Strategy:="Probabilistic") -> BoolElt
{
 It tells whether all the elements of llf satisfies their lifting systems.
}
  for i:=1 to #llf do
    if not VerifyLF(llf[i]:Strategy:=Strategy) then return false; end if;
  end for;
  return(true);
end intrinsic;

///////////////////
// #Section=Splittings
///////////////////

SplitLF_with_option:=function(lf,f,option)
/*
  option can be "Zero", "NonZero", "Both"
*/

  MakeShapeLemmaParametrizationLF(~lf);
  UP<T>:=Universe(lf`MinimalPolynomial);

  lff:=lf; lfnf:=lf;

  lff`MinimalPolynomial:=[UP|];
  lff`Denominator:=[UP|];
  lff`Parametrization:=[];

  lfnf`MinimalPolynomial:=[UP|];
  lfnf`Denominator:=[UP|];
  lfnf`Parametrization:=[];

  for l:=1 to #lf`MinimalPolynomial do
    q:=lf`MinimalPolynomial[l];
    p:=lf`Denominator[l];
    A<a>:=quo<UP|q>;

    point:=ChangeUniverse(lf`MagicPoint,A) cat ChangeUniverse(lf`Parametrization[l],A);

    if Type(f) eq SeqEnum then // special case for sequences
      InitializeEvaluation(f);
      d:=UP!1;
      for e in f do
        p:=UP!EvaluateLF(lf,e,point);
        d*:=Gcd(p,q);
      end for;
    else
      p:=UP!EvaluateLF(lf,f,point);
      d:=Gcd(p,q);
    end if;

    if Degree(d) gt 0 and option in ["Zero","Both"] then // here f=0
      Append(~lff`MinimalPolynomial,d);
      Append(~lff`Denominator,UP!1);
      Append(~lff`Parametrization,[UP| z mod d : z in lf`Parametrization[l]]);
    end if;
    if (Degree(q) - Degree(d) gt 0) and option in ["NonZero","Both"] then // here f<>0
      dd:= q div d;
      Append(~lfnf`MinimalPolynomial,dd);
      Append(~lfnf`Denominator,UP!1);
      Append(~lfnf`Parametrization,[UP| z mod dd : z in lf`Parametrization[l]]);
    end if;
  end for;

  lff`Stamp cat:= "@[" cat Sprint(f`Address[1]) cat "," cat Sprint(f`Address[2]) cat "]";

  case option:
  when "Zero": return lff;
  when "NonZero": return lfnf;
  when "Both": return lff, lfnf;
  else error "unkown option in SplitLF_with_option";
  end case;

end function;

intrinsic SplitLF(lf::Rec,f::[Rec])-> Rec,Rec
{
}
  return SplitLF_with_option(lf,f,"Both");
end intrinsic;

intrinsic SplitLF(lf::Rec,f::Rec)-> Rec,Rec
{
  
     lf, a lifting fiber.
     f, an element of BlackboxPolynomialAlgebra.
  
  
    It returns two fibers lfz, lfnz. The first one represents
    the points of lf satisfying f=0 and the second one
    the other points.
  
}
  return SplitLF_with_option(lf,f,"Both");
end intrinsic;

intrinsic CleanLF(~lf::Rec, ineqs::Rec)
{
 
   Removes the points of lf satisfying ineqs=0.
 
}
  lf:=SplitLF_with_option(lf,ineqs,"NonZero");
end intrinsic;

intrinsic CleanLF(~lf::Rec, ineqs::[])
{
 
   Removes the points of lf satisfying ineqs=0.
 
}
  for e in ineqs do
    CleanLF(~lf,e);
  end for;
end intrinsic;

intrinsic CleanLF(~llf::[], ineqs)
{
 It iterates CleanLF over each element of llf.
}
  for i:=1 to #llf do
    CleanLF(~llf[i],ineqs);
  end for;
end intrinsic;

// #Private
FactorizationLF_ShapeLemma:=procedure(~lf)
/*
//  
//     The ShapeLemma is required.
//  
*/

  if IsEmptyLF(lf) then return; end if;

  UP<T>:=Universe(lf`MinimalPolynomial);
  ans:=lf;
  ans`MinimalPolynomial:=[UP|];
  ans`Denominator:=[UP|];
  ans`Parametrization:=[];

  for l:=1 to #lf`MinimalPolynomial do
    fq:=Factorization(lf`MinimalPolynomial[l]);
    assert lf`Denominator[l] eq UP!1;
    fq:=[ z[1]^z[2] : z in fq];
    for q in fq do
      ans`MinimalPolynomial cat:= fq;
      ans`Denominator cat:= [ UP!1 : z in [1..#fq] ];
      ans`Parametrization cat:= SimultaneousModulo(lf`Parametrization[l],fq);
    end for;
  end for;
  lf:=ans;
end procedure;

// #Public

intrinsic FactorizationLF(~lf::Rec)
{
  
     lf is factorized.
  
  
     The parametrization must be shape lemma.
  
}
  if HasShapeLemmaParametrizationLF(lf) then FactorizationLF_ShapeLemma(~lf);
  else error "Parametrization must be shape lemma";
  end if;
end intrinsic;

CombineLF_Kronecker:=procedure(~lf)
/*
  @var{lf} must have a Kronecker parametrization.
*/
  if IsWholeSpaceLF(lf) then return; end if;
  UP<T>:=Universe(lf`MinimalPolynomial);
  q:=lf`MinimalPolynomial;
  s:=#q;
  c:= CodimensionLF(lf);
  fanin:=Fanin(q:THRESHOLD:=1);
  lf`MinimalPolynomial:=[fanin[#fanin][1]];
  lf`Denominator:=[Derivative(fanin[#fanin][1])];
  lf`Parametrization:= [ [UP| FaninUp([lf`Parametrization[j][i] : j in [1..s]],fanin):
                         i in [1..c]] ];
end procedure;

CombineLF_ShapeLemma:=procedure(~lf)
/*
  @var{lf} must have a shapelemma parametrization.
*/
  if IsWholeSpaceLF(lf) then return; end if;
  UP<T>:=Universe(lf`MinimalPolynomial);
  q:=lf`MinimalPolynomial;
  s:=#q;
  c:= CodimensionLF(lf);
  fanin:=Fanin(q:THRESHOLD:=1);
  lf`MinimalPolynomial:=[fanin[#fanin][1]];
  lf`Denominator:= [UP!1];
  lf`Parametrization:= [
    [UP| FastChineseRemainderTheorem([lf`Parametrization[j][i] :j in [1..s] ],q):
    i in [1..c]] ];
end procedure;

intrinsic CombineLF(~lf::Rec: Parametrization:="Unknown")
{
 It combines the factors of lf.
 The parameter can be "Unknown" (default),
 "ShapeLemma" or "Kronecker" according to the properties
 known about lf.
}
  type:=Parametrization;
  if type eq "Unknown" then
    if HasShapeLemmaParametrizationLF(lf) then
      type:="ShapeLemma";
    elif HasKroneckerParametrizationLF(lf) then
      type:="Kronecker";
    else
      error "Can not determine the shape of the parametrization";
    end if;
  end if;
  case type:
  when "Kronecker": CombineLF_Kronecker(~lf);
  when "ShapeLemma": CombineLF_ShapeLemma(~lf);
  else: error "Bad shape of the parametrization";
  end case;
end intrinsic;

intrinsic MergeLF(llf::[])->Rec
{
  
    It merges the elements of llf into one.
  
  
     The elements of llf must share the same
      ResolutionField, LiftingSystem,
      PrimitiveElement, MagicPoint, ChangeOfVariables,
      GenericTrace and ParentBbp.
     llf must not be empty.
  
}
  require #llf gt 0 : "First argument must not be empty.";
  for i :=2 to #llf do
    require llf[i]`ResolutionField eq llf[1]`ResolutionField : "resolution fields differ";
    require llf[i]`LiftingSystem eq llf[1]`LiftingSystem : "lifting systems differ";
    require llf[i]`PrimitiveElement eq llf[1]`PrimitiveElement : "primitive elements differ";
    require llf[i]`MagicPoint eq llf[1]`MagicPoint : "magic points differ";
    require llf[i]`ChangeOfVariables eq llf[1]`ChangeOfVariables : "change of variables differ";
    require llf[i]`GenericTrace eq llf[1]`GenericTrace : "generic traces differ";
    require llf[i]`ParentBbp eq llf[1]`ParentBbp : "parents bbp differ";
  end for;
  lf:=llf[1];
  lf`MinimalPolynomial:= &cat[ z`MinimalPolynomial : z in llf ];
  lf`Denominator:= &cat[ z`Denominator : z in llf ];
  lf`Parametrization:= &cat[ z`Parametrization : z in llf ];
  return(lf);
end intrinsic;

// #Section=Printing
/////////////////

intrinsic PrintLF(lf::Rec)
{
 
   It prints lf on stdout.
 
}
  if Type(lf) eq MonStgElt then print(lf); return; end if;
  print("LF Stamp: " cat lf`Stamp);
  print("Lifting system");
  PrintBbp(lf`LiftingSystem);
  print("");
  print("Resolution Field");
    print(lf`ResolutionField);
  print("");
  print("Primitive Element");
    print(lf`PrimitiveElement);
  print("Multiplicity");
    print(IsMultipleLF(lf));
  print("Minimal Polynomial");
    print(lf`MinimalPolynomial);
  print("Denominator");
    print(lf`Denominator);
  print("Parametrization");
    print(lf`Parametrization);
  print("Magic Point");
    print(lf`MagicPoint);
  print("Change Of Variables");
    print( lf`ChangeOfVariables`LinearPart);
    print( lf`ChangeOfVariables`AffinePart);
end intrinsic;

intrinsic PrintLF(llf::[])
{
 
   Prints llf on stdout.
 
}
  if Type(llf) eq MonStgElt then print(llf); return; end if;
  for z in llf do
    PrintLF(z);
  end for;
end intrinsic;

