// # 1 "gausshom.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 14:57:48 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "gausshom.mtx" 2
// $Log$


import "random.m": random_RandomInvertibleMatrix, random_RandomInvertibleUpperTriangularMatrix, random_RandomVector, random_RandomPrimeNumber, random_RandomSequence;// #import random.m

intrinsic GaussHomotopySolve(equations::SeqEnum,K) -> .
{
}
   if IsEmpty(equations) then return []; end if;
   E:=equations[1]`Parent;
   n:=Rank(E);
   F:=BaseRing(E);

   // Dirty but quick way to introduce the new variable t used
   // for the homotopy.
   // t is the last variable.
   Et:=BlackboxPolynomialAlgebra(F,n+1);
   Et`Dag`Ops:=E`Dag`Ops;
   Et`Dag`Constants:=E`Dag`Constants;
   
   // Cration de l'homotopie
   homotopy:=equations;
   initpoint:=[Random(1000): i in [1..n]];
   InitializeEvaluation(equations);
   for i:=1 to #equations do
     homotopy[i]`Parent:=Et;
     homotopy[i]:=homotopy[i]-(Cst(Et,1)-Var(Et.(n+1)))*Evaluate(equations[i],initpoint);
   end for;
   PrintBbp(homotopy);
   
   // Build up the solution at t=0.
   // We change the variables so that t becomes the first one.
   lf:= WholeSpaceLF(Et,K:GenericLinearChangeOfVariables:=false);
   delete lf`GenericTrace;
   Q:=Rationals();
   UP<T>:=PolynomialRing(K);
   MP_t:=Parent(lf`PrimitiveElement);
   u:= &+[ MP_t| Random(1000)*MP_t.i : i in [2..n+1]];
//   u:= MP_t.2+ 2*MP_t.3+3*MP_t.4;
   val_u:=Evaluate(u, [0] cat initpoint);
   q:= T-val_u;
   A:= quo<UP|q>;

   lf`MinimalPolynomial:=[UP| q];
   lf`Denominator:=[UP| 1];
   lf`Parametrization:=[ ChangeUniverse(initpoint,UP) ];
   lf`PrimitiveElement:=u;
   lf`MagicPoint:=[Q | 0];
   for i:=1 to n do
     lf`ChangeOfVariables`LinearPart[i,i]:=0;
     lf`ChangeOfVariables`LinearPart[i,i+1]:=1;
   end for;
   lf`ChangeOfVariables`LinearPart[n+1,1]:=1;
   lf`ChangeOfVariables`LinearPart[n+1,n+1]:=0;
   lf`LiftingSystem:=homotopy;
   for i:=1 to n do
     _:=Gradient(homotopy[i]);
   end for;
  
   // Testing and initialisation
   assert VerifyLF(lf);
   lf:=LiftSplitLF(lf);
   assert #lf eq 1;
   lf:=lf[1];
   assert not IsMultipleLF(lf);

   // Lifting
   deg_T:=70;
   deg_t:=70;
   precision:=(deg_T+1)*(deg_t+1)+1;
   LiftCurveLF(~lf,[1],precision);

   // Reconstruction
   S<t>:=BaseRing(Universe(lf`MinimalPolynomial));
   UP<T>:= PolynomialRing(K);
   mp:=ChangeUniverse(lf`MagicPoint,S);
   par:=[S| Coefficient(z,0) : z in lf`Parametrization[1]];
   val_u:=Evaluate(u, mp cat par);
   power_u:= [ S| i eq 0 select S!1 else val_u*Self(i) : i in [0..deg_T] ];

   valuation:= function(a)
     for j:=1 to #a do
       if a[j] ne 0 then return j-1; end if;
     end for;
     return #a;
   end function;   
   // Reconstructing the minimal polynomial
   vprint NonUniversalVerbose,1: "Reconstruction of the minimal polynomial";
   Q,d:=PadeHermite_BeLa_superfast(power_u,precision,[deg_t : i in [0..deg_T]]);
   i_max:=1; d_max:=d[1]; v:=valuation(Q[1]);
   for j:=1 to #d do
     vj:= valuation(Q[j]);
     if d[j] gt d_max or vj gt v then i_max:=j; d_max:=d[j]; v:=vj; end if;
   end for;
   Q:=Q[i_max];

   q:= UP![ Evaluate(z,1) : z in Q];
   q:= q div T^v;
   deg_T:=Degree(q);

   vprint NonUniversalVerbose,1: "Found degree",deg_T;
   lc_q:=LeadingCoefficient(q);
   q:=q/lc_q;

   val_Qp:= &+[S| j*(S!Q[j+1+v])*power_u[j] : j in [1..deg_T] ];
   val_Qp:= val_Qp/lc_q;
   w:= [UP| 0 : l in [1..n]];
   for l:=1 to n do
     // Reconstructing the lth parametrization
     vprint NonUniversalVerbose,1: "Reconstruction of the",l,"th parametrization";
     precision:=(deg_T+1)*(deg_t+1)+1;
     W,d:=PadeHermite_BeLa_superfast(Append([power_u[i]: i in [1..deg_T]],val_Qp*par[l]),precision,[deg_t : i in [1..deg_T]] cat [0]);
     i_max:=1; d_max:=d[1];
     for j:=1 to #d do
       if d[j] gt d_max then i_max:=j; d_max:=d[j]; end if;
     end for;
     W:=W[i_max];
     c:=Evaluate(W[deg_T+1],1);
     assert c ne 0;
     w[l]:= -UP![ Evaluate(W[j],1) : j in [1..deg_T] ]/c;
   end for;

   // Build up the solution at t=1
   lf:= WholeSpaceLF(E,K:GenericLinearChangeOfVariables:=false);
   delete lf`GenericTrace;
   Q:=Rationals();
   MP:=Parent(lf`PrimitiveElement);

   lf`MinimalPolynomial:=[UP| q];
   lf`Denominator:=[UP| Derivative(q)];
   lf`Parametrization:=[ w ];
   lf`PrimitiveElement:= &+[ MP| Evaluate(Derivative(u,1,i+1),[0 : j in [1..n+1]])*MP.i : i in [1..n]];
   lf`MagicPoint:=[Q | ];
   lf`LiftingSystem:=equations;
   for i:=1 to n do
     _:=Gradient(equations[i]);
   end for;
  
   // Testing and initialisation
   lf;
   assert VerifyLF(lf);
   MakeShapeLemmaParametrizationLF(~lf);
   lf:=LiftSplitLF(lf);
   return lf;
   
end intrinsic;

intrinsic GaussHomotopySolveForRationals(equations::SeqEnum) -> .
{
}
  K:=GF(random_RandomPrimeNumber());
  lf:=GaussHomotopySolve(equations,K);
  FindSmallPrimitiveElementLF(~lf);
  LiftIntegersLF(~lf);
  return lf;
end intrinsic;

intrinsic GaussHomotopySolve(equations::SeqEnum) -> .
{
}
  if IsEmpty(equations) then return []; end if;
  R:=equations[1]`Parent;
  F:=BaseRing(R);

  case Type(F):

    when FldRat: return(GaussHomotopySolveForRationals(equations));

    else : return(GaussHomotopySolve(equations,F));

  end case; 
end intrinsic;
