// # 1 "geosolve.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: dim jun 23 10:25:48 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "geosolve.mtx" 2
// $Log$

import "random.m": random_RandomInvertibleMatrix, random_RandomInvertibleUpperTriangularMatrix, random_RandomVector, random_RandomPrimeNumber, random_RandomSequence;// #import random.m
// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 7 "geosolve.mtx" 2

// #Public

// Geometric Solve, Function Index, Equidimensional Decomposition, Top

// #Package=Geometric Solve

/////////////////////
// #Section=Introduction
/////////////////////

// This package provides the main functionalities of Kronecker.
// The verbose flag is GeometricSolveVerbose.

//////////////////
// #Section=Verbosity
//////////////////

intrinsic KroneckerInformations()
{
 It displays informations about Kronecker package:
 version, author, date...
}

  version:="0.166-9";
  author:="Grgoire LECERF";
  date:="dim jun 23 10:25:48 UTC 2002";
  doc:="http://kronecker.medicis.polytechnique.fr/";
  bugreport:="Gregoire.Lecerf@math.uvsq.fr";
  copyright:="1999--2002, G. Lecerf, Centre National de la Recherche Scientifique";

  print "---------------------------------------------------------------------";
  print "";
  print "                       Kronecker version", version;
  print "";
  print "---------------------------------------------------------------------";
  print "";
  print "                          by ",author;
  print "";
  print "                   ",date;
  print "";
  print "            Copyright ",copyright;
  print "";
  print "                           Documentation at";
  print "";
  print "         ",doc;
  print "";
  print "         Bug report at ", bugreport;
  print "";
  print "---------------------------------------------------------------------";
end intrinsic;

intrinsic KroneckerVersion()->MonStgElt
{
}
  return "0.166-9";
end intrinsic;

intrinsic KroneckerSetVerbose(l::RngIntElt)
{
  Make Kronecker verbose. Argument l is an integer between 0 and 5.
}
  SetIndent(2);
  requirerange l, 0, 5;  
  SetVerbose("BbpVerbose",GlobalBbpVerbose[l+1]);
  SetVerbose("FdaVerbose",GlobalFdaVerboseVerbose[l+1]);
  SetVerbose("FiberVerbose",GlobalFiberVerbose[l+1]);
  SetVerbose("GeometricSolveVerbose",GlobalGeometricSolveVerbose[l+1]);
  SetVerbose("HenselVerbose",GlobalHenselVerboseVerbose[l+1]);
  SetVerbose("IntersectionVerbose",GlobalIntersectionVerbose[l+1]);
end intrinsic;

intrinsic KroneckerSetVerbose()
{
  Set to default verbosity 2.
}
  KroneckerSetVerbose(2);
end intrinsic;

intrinsic KroneckerSetMathMLVerbose()
{
  Make Kronecker verbose via MathML.
}
  SetVerbose("KroneckerMathMLVerbose",2);
end intrinsic;

intrinsic KroneckerUnsetMathMLVerbose()
{
  Stop Kronecker verbose via MathML.
}
  SetVerbose("KroneckerMathMLVerbose",0);
end intrinsic;

// #Private
intrinsic KroneckerBug(file,line)
{
  Print bug report message and tell what to do.
}
  print "BUG in Kronecker! Please report to Gregoire.Lecerf@math.uvsq.fr:";
  print "-> Send the following informations:";
  print "file:", file;
  printf "lineno: %o\n", line;
  printf "seed: %o\n",GetSeed();
  a,b,c:=GetVersion();
  printf "Magma version: %o.%o-%o\n",a,b,c;
  printf "Kronecker version: %o\n",KroneckerVersion();
  print "-> Send the whole input file.";
  print "-> If possible, send the whole output of the session.";
  error "Thanks for reporting.";
end intrinsic;
// #Public

////////////////////////////////////////
// #Section=Geometric solve functionalities
////////////////////////////////////////

intrinsic GeometricSolve(equations::SeqEnum, inequations::SeqEnum, K: GenericLinearChangeOfVariables:= true, RemoveMultipleComponents:=false) -> Rec
{
  
     equations.
     inequations.
     K, field.
  
  
    It returns a sequence of sequences llf of lifting fibers.
  
  
     llf describes the variety defined by the equations
     outside the inequations, if everything goes right.
     The computations are performed over K.
  
  
    Since the algorithm is probabilistic an error can be raised.
  
}
  local i;
  s:=#equations;
  if s eq 0 then return([]); end if;
  R:=equations[1]`Parent;
  n:=Rank(R);
  if assigned R`MathMLsession then alreadyinsession:=true; session:=R`MathMLsession; else alreadyinsession:=false; end if;
  
  if IsVerbose("KroneckerMathMLVerbose") and not alreadyinsession then
    R`MathMLsession:= OpenMathMLSession();

    MathML_open_page(~R`MathMLsession, "index.html", "Kronecker session");
    if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "<h1> Kronecker: geometric resolution algorithm </h1> \n"); end if;
    if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "<HR>\n"); end if;
  end if;

  answer:= [ [] : i in [1..n+1] ];
  answer[n+1]:=[WholeSpaceLF(R,K: GenericLinearChangeOfVariables:= GenericLinearChangeOfVariables)];

  for i:=1 to #equations do
    f:=equations[i];
    answer:=IntersectLF(answer,f,inequations
      :RemoveMultipleComponents:=RemoveMultipleComponents);
    if Type(answer) eq MonStgElt then KroneckerBug("geosolve.mtx",164); end if;

    vprint GeometricSolveVerbose,1:"GeometricSolve: ",i," equations solved. Degree is ",DegreeLF(answer);
    vprint GeometricSolveVerbose,2:"                      *   *";
    vprint GeometricSolveVerbose,2:"                        *  ";
    vprint GeometricSolveVerbose,2:"GeometricSolve: Checking resolution ...";
    vprint GeometricSolveVerbose,5:Sprint(answer);
    if not VerifyLF(answer) then
      PrintLF(answer);
      KroneckerBug("geosolve.mtx",173);
    end if;
    vprint GeometricSolveVerbose,2:"GeometricSolve: resolution OK";
    vprint GeometricSolveVerbose,2:"Degrees of the components\n", [[ DegreeLF(z) : z in zz]: zz in answer];

    if IsVerbose("KroneckerMathMLVerbose") and not IsEmpty(answer) then
      FlushMathMLSession(~R`MathMLsession);
      MathML_tag_1(~R`MathMLsession,"hr","");
      MathML_tag_1(~R`MathMLsession,"p","");
      if i eq 1 then
        MathML_tag_1(~R`MathMLsession,"b",Sprint(i) cat " equation solved:": attributes:=rec<MathMLAttributes()|color:="red">);
      else
        MathML_tag_1(~R`MathMLsession,"b",Sprint(i) cat " equations solved:": attributes:=rec<MathMLAttributes()|color:="red">);
      end if;
      MathMLLLF(~R`MathMLsession,answer);
    end if;
  end for;

  if IsVerbose("KroneckerMathMLVerbose") and not alreadyinsession then
       MathML_tag_1(~R`MathMLsession,"hr","");
       MathML_close_page(~R`MathMLsession);
       CloseMathMLSession(~R`MathMLsession);
  end if;

  return(answer);
end intrinsic;

// #Private
intrinsic FactorizedGeometricSolve(equations::SeqEnum, inequations::SeqEnum, K: GenericLinearChangeOfVariables:= true, RemoveMultipleComponents:=false) -> Rec
{
  
     equations.
     inequations.
     K, field.
  
  
    It returns a sequence of sequences llf of lifting fibers.
  
  
     llf describes the variety defined by the equations
     outside the inequations, if everything goes right.
     The computations are performed over K.
     Intermediate varieties computed during the resolution are factorized.
  
  
    Since the algorithm is probabilistic an error can be raised.
  
}
  local i;
  s:=#equations;
  if s eq 0 then return([]); end if;
  R:=equations[1]`Parent;
  n:=Rank(R);

  answer:= [ [] : i in [1..n+1] ];
  answer[n+1]:=[WholeSpaceLF(R,K: GenericLinearChangeOfVariables:= GenericLinearChangeOfVariables)];

  for i:=1 to #equations do
    f:=equations[i];
    answer:=IntersectLF(answer,f,inequations
      :RemoveMultipleComponents:=RemoveMultipleComponents);
    if Type(answer) eq MonStgElt then KroneckerBug("geosolve.mtx",234); end if;

    vprint GeometricSolveVerbose,1:"GeometricSolve: ",i," equations solved. Degree is ",DegreeLF(answer);
    vprint GeometricSolveVerbose,2:"                      *   *";
    vprint GeometricSolveVerbose,2:"                        *  ";
    vprint GeometricSolveVerbose,2:"GeometricSolve: Checking resolution ...";
    vprint GeometricSolveVerbose,5:Sprint(answer);
    if not VerifyLF(answer) then
      PrintLF(answer);
      KroneckerBug("geosolve.mtx",243);
    end if;
    vprint GeometricSolveVerbose,2:"GeometricSolve: resolution OK";
    vprint GeometricSolveVerbose,2:"Irreducible decomposition";
    answer:=[ TrueFactorizationLF(z) : z in answer] ;
    vprint GeometricSolveVerbose,2:"Degrees of the components\n", [[ DegreeLF(z) : z in zz]: zz in answer];
  end for;
  return(answer);
end intrinsic;

// #Private
GeometricSolveForRationals:=function(equations, inequations);

  if IsEmpty(equations) then return([]); end if;

  R:=equations[1]`Parent;
  n:=Rank(R);
  F:=BaseRing(R);
  assert Type(F) eq FldRat;
  beginmodulartime:=Cputime();

  if IsVerbose("KroneckerMathMLVerbose") then
    R`MathMLsession:=OpenMathMLSession();
    R`MathMLsession:=R`MathMLsession;
    MathML_open_page(~R`MathMLsession, "index.html", "Kronecker session");
    if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "<h1> Kronecker: geometric resolution algorithm </h1> \n"); end if;
    if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "<HR>\n"); end if;
    MathML_tag_1(~R`MathMLsession,"hr","");
  end if;

  // Heuristic for choosing the prime number.
  h:=Max(UpperBoundHeight(equations));
  d:=Sort(TotalDegree(equations: Strategy:="UpperBound"));
  D:=&*d[Max(1,#d-n+1)..#d];
  H:=Min([h*D,2^100,1000*D]);
  if Ilog2(Max(1,H)) ge 100 then
    print "WARNING: cannot reach expected prime size";
  end if;

  K:=GF(NextPrime(H+Random(10^5)));
  vprint GeometricSolveVerbose,1:"GeometricSolveForRationals: computing in",K;

  if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "Modular resolution in " cat Sprint(K)); end if;
  if IsVerbose("KroneckerMathMLVerbose") then
    MathML_tag_1(~R`MathMLsession,"hr","");
  end if;
  
  llf:=GeometricSolve(equations,inequations,K:GenericLinearChangeOfVariables:=true);
  if Type(llf) eq MonStgElt then KroneckerBug("geosolve.mtx",291); end if;
  vprint GeometricSolveVerbose,1:"Equations Solved Modularly in",Cputime()-beginmodulartime,"s";

  FindSmallNoetherPositionLF(~llf);
  if Type(llf) eq MonStgElt then KroneckerBug("geosolve.mtx",295); end if;
  FindSmallPrimitiveElementLF(~llf);
  if Type(llf) eq MonStgElt then KroneckerBug("geosolve.mtx",297); end if;
  FindSmallMagicPointLF(~llf);
  if Type(llf) eq MonStgElt then KroneckerBug("geosolve.mtx",299); end if;

  if IsVerbose("KroneckerMathMLVerbose") then
    MathML_tag_1(~R`MathMLsession,"hr","");
    MathML_tag_1(~R`MathMLsession,"hr","");
  end if;
  if IsVerbose("KroneckerMathMLVerbose") then Put(R`MathMLsession`Pipe, "Resolution over the Rationals"); end if;
  if IsVerbose("KroneckerMathMLVerbose") then
    MathML_tag_1(~R`MathMLsession,"hr","");
    MathML_tag_1(~R`MathMLsession,"hr","");
  end if;

  LiftIntegersLF(~llf);
  if Type(llf) eq MonStgElt then KroneckerBug("geosolve.mtx",312); end if;

  if IsVerbose("KroneckerMathMLVerbose") then
    MathMLLLF(~R`MathMLsession,llf);
  end if;
  
  return llf;
end function;
// #Public
intrinsic GeometricSolve(equations::SeqEnum, inequations::SeqEnum) -> Rec
{
}
  if IsEmpty(equations) then return []; end if;
  R:=equations[1]`Parent;
  n:=Rank(R);
  F:=BaseRing(R);
  case Type(F):
    when FldRat:
      return(GeometricSolveForRationals(equations,inequations));
    else : return(GeometricSolve(equations,inequations,F));
  end case; 
end intrinsic;

intrinsic GeometricSolve(equations::SeqEnum) -> Rec
{
}
  return GeometricSolve(equations,[]);
end intrinsic;





