// # 1 "hensel.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: lun jun 10 21:05:09 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "hensel.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "hensel.mtx" 2
import "hensel_base.m": hensel_base__FromSeriesToRationalCoefficients, hensel_base__FromSeriesToRationalCoefficients_Pade, hensel_base__EvaluateAtOne, hensel_base__MultiplyByCommonDenominator, hensel_base__StopCriterion_Bound, hensel_base__StopCriterion_Integers, hensel_base__HeuristicCombineFactorized;// #import hensel_base.m


// #Public
// Hensel Lifting, Equidimensional Decomposition, Lifting fibers, Top

// #Package=Lifting

// This package provides an implementation of the global Newton
// lifting algorithm. Its verbose flag is HenselVerbose.

/////////////////////////////////
// #Section=Splitting before lifting
/////////////////////////////////

intrinsic LiftSplitLF(lf::Rec)-> SeqEnum
{
 
    lf, is an isolated lifting fiber.
 
 
   It returns a sequence of isolated lifting fibers corresponding to
   different subvarieties behaving differently wrt the the lifting process.
   This splitting must be achieved before performing any lifting.
 
}
  require HasShapeLemmaParametrizationLF(lf): "lf must be shape lemma";
  //TODO take care because of copying
//  if not IsMultipleLF(lf)// or assigned lf`GenericTrace
//  then ans:=LiftSplitLF_newton(lf);
//  else
  ans:= LiftSplitLF_fda(lf);
//  end if;
  return ans;
end intrinsic;

////////////////////
// #Section=Lift curves
////////////////////

// Common requirements for all the lifting functions:
//  
//    lf must be shape lemma.
//    The variety must be isolated wrt to the
//     lf`LiftingSystem.
//    lf must be irreducible with respect to the lifting
//     process.
//  itemize

intrinsic LiftCurveLF(~lf::Rec,destpoint::SeqEnum,precision::RngIntElt)
{
 
    lf, an isolated lifting fiber.
    destpoint, destination point.
    precision, integer.
 
 
   This procedure lifts the curve from the lifting point to the destination
   point destpoint and up to precision.
 
 
   cf. common requirements above.
 
}
  MakeShapeLemmaParametrizationLF(~lf);
  if IsMultipleLF(lf) then
    LiftCurveLF_fda(~lf,destpoint,precision);
  else
    LiftCurveLF_newton(~lf,destpoint,precision);
  end if;
end intrinsic; // LiftCurveLF

intrinsic LiftCurveLF(~llf::[],destpoint::SeqEnum,precision::RngIntElt)
{
  It iterates LiftCurveLF over each element of llf.
}
  answer:=[];
  for i:=1 to #llf do
    lf:=llf[i];
    LiftCurveLF(~lf,destpoint,precision);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    Append(~answer,lf);
  end for;
  llf:=answer;
end intrinsic;

intrinsic LiftLastFreeVariableLF(~lf::Rec)
{
 
    lf, an isolated lifting fiber.
 
 
   This procedure lifts the last free variable of lf.
   lf`ResolutionField becomes
   an univariate rational function field over the resolution field.
 
 
   cf. common requirements above.
 
}
  TranslateLastFreeVariableToZeroLF(~lf);
  hensel_base__HeuristicCombineFactorized(~lf);
  mp:=lf`MagicPoint;
  r:=#mp;
  LiftCurveLF(~lf,mp[1..r-1] cat [mp[r]+1],DegreeLF(lf)+1);
  if Type(lf) eq MonStgElt then return; end if;
  MakeKroneckerParametrizationLF(~lf);
  CombineLF(~lf:Parametrization:="Kronecker");
  hensel_base__FromSeriesToRationalCoefficients(~lf);
  //assert VerifyLF(lf);
end intrinsic;

intrinsic LiftLastFreeVariableLF(~llf::[])
{
  It iterates LiftLastFreeVariableLF over each element of llf.
}
  answer:=[];
  for i:=1 to #llf do
    lf:=llf[i];
    LiftLastFreeVariableLF(~lf);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    Append(~answer,lf);
  end for;
  llf:=answer;
end intrinsic;

///////////////////////////////
// #Section=Change the magic point
///////////////////////////////

intrinsic ChangeMagicPointLF(~lf::Rec,magicpoint)
{
  
     lf, an isolated lifting fiber.
     magicpoint, the destination magic point.
  
  
    Computes the lifting fiber for the magic point magicpoint
    and returns it in lf.
  
  
     #magicpoint must be equal to the dimension of lf
     cf. common requirements above.
  
}
  hensel_base__HeuristicCombineFactorized(~lf);
  LiftCurveLF(~lf,magicpoint,DegreeLF(lf)+1);
  if Type(lf) eq MonStgElt then return; end if;
  MakeKroneckerParametrizationLF(~lf);
  CombineLF(~lf:Parametrization:="Kronecker");
  hensel_base__EvaluateAtOne(~lf);
  // TODO faire des tests avec cette fonction
  // TODO description dans le tutorial
end intrinsic;

intrinsic ChangeMagicPointLF(~llf::[],magicpoint)
{
  It iterates ChangeMagicPointLF over each element of llf.
}
  answer:=[];
  for i:=1 to #llf do
    lf:=llf[i];
    ChangeMagicPointLF(~lf,magicpoint);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    Append(~answer,lf);
  end for;
  llf:=answer;
end intrinsic;


//////////////////////
// #Section=Check lifting
//////////////////////

intrinsic HasLiftingPointLF(lf::Rec)->BoolElt
{
  
     lf, an isolated fiber.
  
  
    It tells whether the magic point is a lifting point.
    If the variety is multiple wrt its lifting system then
    its generic trace must be known.
  
  
     cf. common requirements above.
  
}
  if IsMultipleLF(lf) then
    return(HasLiftingPointLF_fda(lf));
  else
    return(HasLiftingPointLF_newton(lf));
  end if;
end intrinsic;

intrinsic HasLiftingPointLF(llf::[])->BoolElt
{
  It iterates HasLiftingPointLF over each element of llf.
}
  for i:=1 to #llf do
    lf:=llf[i];
    if not HasLiftingPointLF(lf) then return false; end if;
  end for;
  return true;
end intrinsic;

// #Private
///////////////////////////////
// #Section=Change the coordinates
///////////////////////////////

swap:=function(lf,i,r)

  E:=lf`ParentBbp;
  n:=NumberOfVariablesLF(lf);
  D:=DegreeLF(lf);
  F:=lf`ResolutionField;
  M:=lf`ChangeOfVariables`LinearPart;
  b:=lf`ChangeOfVariables`AffinePart;
  
  // In order to deal with the ith free variable
  // we swap it with the last one.
  tmp:=lf`MagicPoint[i];
  lf`MagicPoint[i]:=lf`MagicPoint[r];
  lf`MagicPoint[r]:=tmp;
  N:=ScalarMatrix(n,F!1);
  N[i,i]:=0; N[r,r]:=0;
  N[i,r]:=1; N[r,i]:=1;
  lf`ChangeOfVariables`LinearPart *:= N;
  M:=lf`ChangeOfVariables`LinearPart;
  MP:=Parent(lf`PrimitiveElement);
  vars:=[MP| MP.j: j in [1..n]]; tmp:=vars[i];
  vars[i]:=vars[r]; vars[r]:=tmp;
  lf`PrimitiveElement:=Evaluate(lf`PrimitiveElement,vars);
  delete vars;
  delete tmp;
  assert VerifyLF(lf);
  // end of swap
  return lf;
end function;

intrinsic ChangeCoordinatesLF(~lf::Rec,i,argm,argp)
{
  
     lf, a lifting fiber.
     m, a sequence.
     p, a constant.
  
  
    TODO
    Note that the magic point is not modified and it is not checked that
    either the new variables are in Noether position and that the magic
    point is still a lifting point.
  
  
    TODO
    cf. common requirements above.
  
}
  vprint HenselVerbose,1:"--> ChangeCoordinatesLF";
  begintime:=Cputime(); 
  E:=lf`ParentBbp;
  r:=DimensionLF(lf);
  n:=NumberOfVariablesLF(lf);
  D:=DegreeLF(lf);
  F:=lf`ResolutionField;

  // Change the ith variable with the rth  
  lf:=swap(lf,i,r);
    
  M:=lf`ChangeOfVariables`LinearPart;
  b:=lf`ChangeOfVariables`AffinePart;
  
  m:=Eltseq(M[i]-Vector(argm)); m[i]:=0; m[r]:=1;
  p:=b[i]-argp;

  // Lifting the curve
  lc:=lf;
  LiftLastFreeVariableLF(~lc);
  if Type(lc) eq MonStgElt then lf:=lc; return; end if;
  // Intersection with the linear equation
  locm:= Matrix(1,n,m)*(M^(-1));
  locm:= locm[1];
  locp:= - &+[F| locm[j]*b[j] : j in [1..n]] + p;
  f:= Add(E,[ Cst(E,Integers()!locm[i])*Var(E.i): i in [1..n]]) + Cst(E,Integers()!locp);
  KroneckerIntersect(~lc,f);
  if Type(lc) eq MonStgElt then lf:=lc; return; end if;
  // Set f as the new last free variable  
  N:= ScalarMatrix(n,F!1);
  for j:=r+1 to n do N[r,j]:=m[j]; end for;
  N[r,r]:=1;

  v:= Vector([F| 0 : j in [1..n]]);
  v[r]:=p;

  // We return if the new fiber has been split
  if #lc ne 1 then lf:="FAILED"; return; end if;

  lc:=lc[1];
  lc`ChangeOfVariables`LinearPart:= M*N^(-1);
  lc`ChangeOfVariables`AffinePart-:=v*Transpose(M*N^(-1));
  lc`MagicPoint:=Append(Prune(lf`MagicPoint),F!0);
  Prune(~lc`LiftingSystem);
  MP:=Parent(lf`PrimitiveElement);
  lc`PrimitiveElement:= -&+[ m[j]*MP.j : j in [r+1..n] | j ne i] -p;
  lc`Parametrization:= [ [z[j] : j in [2..n-r+1]] : z in lc`Parametrization];
  lc`GenericTrace:=lf`GenericTrace;

  // Swap back the rth variable with the ith
  lf:=swap(lc,i,r);
  vprint HenselVerbose,1: "<-- ChangeCoordinatesLF, elapsed time: ", Cputime()-begintime;
end intrinsic;


