// # 1 "hensel_base.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 14:57:49 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "hensel_base.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "hensel_base.mtx" 2


/*
 *   Provides common functionnalities to the hensel_* algorithms.
 */

hensel_base__FromSeriesToRationalCoefficients:=procedure(~lf)
/*
  1. lf must have kronecker parametrization
  2. coerce the series into rational functions

  In return the resolution field of lf is a field of rational functions.
*/
  LS<y>:=lf`ResolutionField;
  F:=BaseRing(LS);
  precision:=lf`ResolutionField`Precision;
  UPy<y>:=PolynomialAlgebra(F);
  RFy<y>:=FieldOfFractions(UPy);
  f1:=hom<LS->UPy| x:->UPy![Coefficient(LS!x,i): i in [0..precision-1]]>;
  f2:=hom<UPy->RFy | RFy.1>;
  ChangeResolutionFieldLF(~lf,RFy,f1*f2);
end procedure;

hensel_base__FromSeriesToRationalCoefficients_Pade:=procedure(~lf,precision,num,den)
/*
  make pade approximant (num,den) on each coefficients and the change of variable
  In return the resolution field of lf is a field of rational functions.
*/
  LS<y>:=lf`ResolutionField;
  F:=BaseRing(LS);
  UP:=PolynomialAlgebra(F);
  RF<y>:=FieldOfFractions(UP);
  f:=hom<LS->RF| x:-> PadeApproximant(x,precision,num,den)>;
  ChangeResolutionFieldLF(~lf,RF,f);
end procedure;

hensel_base__EvaluateAtOne:=procedure(~lf)
/*
  Evaluate at 1 the parameter of the resolution field of lf.
*/
  K:=lf`ResolutionField;
  F:=BaseRing(K);
  f:=hom<K->F | x:->Evaluate(x,F!1)>;
  ChangeResolutionFieldLF(~lf,F,f);
end procedure;

hensel_base__MultiplyByCommonDenominator:=procedure(~lf)
/*
  lf`ResolutionField is a function field.
  the minimal polynomial, denominator and parametrization are multiplied by the
  least common denominator of all the rational functions.
*/
  RF:=lf`ResolutionField;
  listdenom:= // list of all the denominators
        &cat[ [Denominator(zz): zz in Coefficients(z)] : z in lf`MinimalPolynomial ]
    cat &cat[ [Denominator(zz): zz in Coefficients(z)] : z in lf`Denominator ]
    cat &cat[ &cat[ [Denominator(z) : z in Coefficients(zz) ] : zz in zzz]: zzz in lf`Parametrization];
  commondenom:=RF!Lcm(listdenom);
  lf`MinimalPolynomial := [commondenom*z : z in lf`MinimalPolynomial];
  lf`Denominator := [ commondenom*z : z in lf`Denominator];
  lf`Parametrization := [ [commondenom * z : z in w]: w in lf`Parametrization];
end procedure;

hensel_base__StopCriterion_Bound:=function(precision)
/*
  The bound is precision.
  To be removed... later.
*/
  precisions:=[];
  k:=precision;
  while k gt 1 do
    Append(~precisions,k);
    if k mod 2 eq 0 then k:= (k div 2); else k:= (k div 2) +1; end if;
  end while;
  Append(~precisions,1);
  vprint HenselVerbose,2: "Precisions:",precisions;
  stopcriterion:=function(lf)
    curprec:=lf`ResolutionField`Precision;
    if curprec ge precision then
      return true,lf;
    else 
      for i:=#precisions to 1 by -1 do
        if curprec lt precisions[i] then
          return false,(precisions[i]);
        end if;
      end for;
    end if;
  end function;
  return(stopcriterion);
end function;

hensel_base__StopCriterion_Integers:=function()
/*
*/
  stopcriterion:=function(lf)
    vprint HenselVerbose,2:"--> StopCriterion";
    loclf:=lf;
    // La recombinaison systematique coute trop cher.
    // On fait des tests simples avant.
    UP<T>:=Universe(lf`MinimalPolynomial);
    F:=BaseRing(UP);
    LS<T>:= PowerSeriesAlgebra(F,3);
    qs:=&*[ LS!z : z in lf`MinimalPolynomial];
    vtime HenselVerbose,3: b,_:=RationalReconstruction(Coefficients(qs));
    if not b then
      vprint HenselVerbose,3: "[0 early exit from stop criterion]";
      return false, 2*Precision(lf`ResolutionField); //// ici la stategie est bonne
    end if;

    vtime HenselVerbose,3: q:=&*lf`MinimalPolynomial;
    vtime HenselVerbose,3: b,q:=RationalReconstruction(Coefficients(q));
    if not b then
      vprint HenselVerbose,3: "[1 early exit from stop criterion]";
      return false, 2*Precision(lf`ResolutionField); //// revoir une bonne strategie !
    end if;

    // TODO : fixer et tester avec boon.
    // lc_denom:= Lcm([Denominator(z) : z in q]);
    // Il faut faire un truc qui recupere le denominateur du polynome minimal...
    // Multiplier la parametrization de loclf avec puis reconstruire....
    // En general les coefficients du polynome minimal sont un peu plus petits.
    //loclf`MinimalPolynomial[1] *:=lc_denom ; // Cela suffit car la parametrisation est shapelemma.
    vtime HenselVerbose,3: MakeKroneckerParametrizationLF(~loclf);
    vtime HenselVerbose,3: CombineLF(~loclf:Parametrization:="Kronecker");
    vtime HenselVerbose,3: b,ans:= RationalReconstructionLF(loclf);
        vtime HenselVerbose,3: if b then
        MakeMonicLF(~ans);
        if VerifyLF(ans:Strategy:="Probabilistic")
          then return true,ans;
          else return false, 2*Precision(lf`ResolutionField);
            //// revoir une bonne strategie ! car on y est presque.
        end if;
    end if;
    vprint HenselVerbose,2:"<-- StopCriterion";
    return false, 2*Precision(lf`ResolutionField); //// ici aussi on est pas tres loin.
  end function;

  return(stopcriterion);
end function;

hensel_base__HeuristicCombineFactorized:=procedure(~lf)
/*
// 
//    lf, a (factorized) geometric resolution.
// 
// 
//   In return lf is heuristically recombined.
// 
// 
//   lf must be shape lemma.
// 
*/
  if IsWholeSpaceLF(lf) then return; end if;
  
  UP<T>:=Universe(lf`MinimalPolynomial);

  combine:=function(qq,ww)
    s:=#qq;
    assert s gt 0;    
    fanin:=Fanin(qq:THRESHOLD:=1);
    ansq:=fanin[#fanin][1];
    ansp:=UP!1;
    c:=#ww[1];
    answ:=
      [UP| FastChineseRemainderTheorem([ww[j][i] :j in [1..s] ],qq):
      i in [1..c]];

    return ansq,ansp,answ;
  end function;

  s:=#lf`MinimalPolynomial;
  i:=1;
  begin:=1; locdeg:=0;
  TH:=14; // HACK!!!!!
  new_q:=[]; new_p:=[]; new_w:=[];
  while i le s do
    locdeg +:= Degree(lf`MinimalPolynomial[i]);
    if locdeg gt TH or i eq s then
      locq,locp,locw:=combine([lf`MinimalPolynomial[j]: j in [begin..i]],[lf`Parametrization[j]: j in [begin..i]]);
      Append(~new_q,locq);
      Append(~new_p,locp);
      Append(~new_w,locw);
      begin:=i+1; locdeg:=0;
    end if;
    i:=i+1;
  end while;

  lf`MinimalPolynomial:=new_q;
  lf`Parametrization:=new_w;
  lf`Denominator:=new_p;

end procedure;

