// # 1 "hensel_newton.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: lun jun 17 16:57:00 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "hensel_newton.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "hensel_newton.mtx" 2
import "hensel_base.m": hensel_base__FromSeriesToRationalCoefficients, hensel_base__FromSeriesToRationalCoefficients_Pade, hensel_base__EvaluateAtOne, hensel_base__MultiplyByCommonDenominator, hensel_base__StopCriterion_Bound, hensel_base__StopCriterion_Integers, hensel_base__HeuristicCombineFactorized;// #import hensel_base.m


// #Private

/******************************************************************************
                           Lifting Core
******************************************************************************/

UpdateJacobianInverse:=procedure(~ls,requestedprecision)
/*
  Improves the precision of the inverse of the Jacobian matrix
  up to requested precision, if possible.
*/
  reachableprecision:=Min([2*ls`JacobianInversePrecision,ls`JacobianPrecision,requestedprecision]);
  if reachableprecision lt requestedprecision then
    print("WARNING: can not reach requested precision");
  end if;
  vprint HenselVerbose,3: "Updating Jacobian inverse to precision", reachableprecision;
  R,H:=ChangePrecision(BaseRing(Parent(ls`JacobianInverse)),reachableprecision);
  I:=ScalarMatrix(NumberOfColumns(ls`JacobianInverse),R!1);
  iJ:=MyChangeRing(ls`JacobianInverse,R,H);
  if ls`JacobianInversePrecision lt reachableprecision then
    J:=MyChangeRing(ls`JacobianValue,R,H);
    ls`JacobianInverse:= iJ-iJ*(J*iJ-I);
  end if;
  ls`JacobianInversePrecision:=reachableprecision; 
end procedure;

UpdateSolution:=procedure(~ls, f, fprecision)
/*
  Improves the solution of ls, using the value of the equations f,
  known at precision fprecision.
*/
  reachableprecision:= Min(
    Min(2*ls`JacobianInversePrecision,ls`JacobianPrecision)
      +ls`SolutionPrecision,
    fprecision
  );
  if reachableprecision lt fprecision then
    print("WARNING: can not reach requested precision");
  end if;
  vprint HenselVerbose,3: "Updating solution to precision", reachableprecision;
  // set precision
  R,h:=ChangePrecision(Universe(f),reachableprecision);
  jv:=Transpose(MyChangeRing(ls`JacobianValue,R,h));
  ji:=Transpose(MyChangeRing(ls`JacobianInverse,R,h));
  floc:=Vector([R| h(z) : z in f]);
  // computations
  tmp:= floc*ji; // sub expression sharing
  tmp:= 2*tmp-(tmp*jv)*ji;
  ls`Solution := [R| h(ls`Solution[i]) - tmp[i] : i in [1..#f]];
  ls`SolutionPrecision:=reachableprecision;
end procedure;

InitializeEvaluationOfEquationsAndJacobian:=procedure(~lf)
/*
*/
  trace:=[ z[1] : z in lf`GenericTrace`Trace];
  tmp:= [ lf`LiftingSystem[z]: z in trace ];
  InitializeEvaluation(tmp cat &cat[Gradient(z): z in tmp]);
end procedure;

EvaluateEquations:=function(lfloc,point)
/*
*/
  trace:=[ z[1] : z in lfloc`GenericTrace`Trace];
  return EvaluateLF(lfloc,[lfloc`LiftingSystem[z] : z in trace],point);
end function;

EvaluateJacobian:=function(lf,x)
/*
*/
  n:=NumberOfVariablesLF(lf);
  assert #x eq n;
  r:=RankLF(lf);
  s:=n-r;
  R:=Universe(x);
  ans:=[R|];
  trace:=[ z[1] : z in lf`GenericTrace`Trace];
  assert #trace eq s;
  
  // Scanning what there really is to evaluate.
  todo:={};
  for i:=1 to s do
    for k:=1 to s do
      for j:=1 to n do
        if lf`ChangeOfVariables`LinearPart[j,r+k] ne 0 then
          Include(~todo,[i,j]);
        end if;
      end for;
    end for;
  end for;

  todo:= [ z : z in todo];
  values:= [ [ R| 0 : j in [1..n]] : i in [1..s] ];
  tmp:= EvaluateLF(lf,[Derivative(lf`LiftingSystem[trace[z[1]]],z[2]): z in todo],x);
  for i:=1 to #todo do values[todo[i][1]][todo[i][2]]:=tmp[i]; end for;
  delete tmp,todo;

  for i:=1 to s do
    for k:=1 to s do
      tmp:=R!0;
      for j:=1 to n do
        c:=lf`ChangeOfVariables`LinearPart[j,r+k];
        if c ne 0 then
          tmp+:=(R!c)*values[i][j];
        end if;
      end for;
      Append(~ans,tmp);
    end for;
  end for;

  return(Matrix(R,s,ans));
end function;

EvaluateEquationsAndJacobian:=function(lf,ls,requestedprecision)
/*
  Evaluates the equations and the jacobian matrix of the fiber lf
  at precision requestedprecision.
*/
  currentprecision:=ls`SolutionPrecision;
  s:=#lf`LiftingSystem;
  MP:=ParentBbpLF(lf);
  A,H:=ChangePrecision(Universe(ls`Solution),requestedprecision);
  if IsWholeSpaceLF(lf) then return [A|], Matrix(A,0,[]); end if;
  sol:=[A|H(z): z in ls`Solution];
  mp:=[A|H(Universe(ls`Solution)!z): z in ls`MagicPoint];
  point:= mp cat sol;
  lfloc:=lf;
  lfloc`ChangeOfVariables:=ls`ChangeOfVariables;

  vprint HenselVerbose,3: "Initializing evaluation...";
  vtime HenselVerbose,3: InitializeEvaluationOfEquationsAndJacobian(~lf);

  vprint HenselVerbose,3: "Evaluating the equations...";
  vtime HenselVerbose,3: equationsvalue:=EvaluateEquations(lfloc,point);

  // Diminution de la prcision dans la table de remember  currentprecision
  A,H:= ChangePrecision(Universe(MP`Dag`Values[1]),currentprecision);
  vprint HenselVerbose,3: "Updating RememberTable...";
  vtime HenselVerbose,3: MP`Dag`Values:= [ [A| H(zz): zz in z] : z in MP`Dag`Values];
  point:=[H(z):z in point];
  
  vprint HenselVerbose,3: "Evaluating the jacobian...";
  vtime HenselVerbose,3: jacobianvalue:=EvaluateJacobian(lfloc,point);

  ClearRememberTableValues(~MP);
  return equationsvalue,jacobianvalue;
end function;

EvaluateJacobian_l:= function(lf,l)
/*
  Returns the value of the Jacobian corresponding to the l th
  factor of lf.
*/
  s:=CodimensionLF(lf);
  q:=lf`MinimalPolynomial[l];
  A<a>:=quo<Parent(q)|q>;
  point:=ChangeUniverse(lf`MagicPoint,A) cat ChangeUniverse(lf`Parametrization[l],A);
  J:=EvaluateJacobian(lf,point);
  return J;
end function;

GlobalizeSolution:= procedure(~ls)
/*
  Globalization deformation trick: improves precision
  of the minimal polynomial using the primitive element.
*/
  A<a>:=Universe(ls`Solution);
  UP<T>:=PreimageRing(A);
  q:=Modulus(A);
  point:=ChangeUniverse(ls`MagicPoint,A) cat ls`Solution;
  du:=a-Evaluate(ls`PrimitiveElement, point);
  Q:= q + UP!((A!Derivative(q))*du);
  B<a>:=quo<UP|Q>;
  ls`Solution:= [B| UP!(z + (A!Derivative(UP!z))*du) : z in ls`Solution];
  ls`JacobianValue:=MyChangeRing(ls`JacobianValue,UP);
  ls`JacobianValue:=MyChangeRing(ls`JacobianValue,B);
  ls`JacobianInverse:=MyChangeRing(ls`JacobianInverse,UP);
  ls`JacobianInverse:=MyChangeRing(ls`JacobianInverse,B);
end procedure;

/******************************************************************************
                           Lifting Public
******************************************************************************/


LiftingState:=function()
/*
  Returns the record format used to store lifting states.
*/
  return(recformat<
    Algorithm,
    MagicPoint,
    PrimitiveElement,
    Solution,
    SolutionPrecision,
    JacobianValue,
    JacobianPrecision,
    JacobianInverse,
    JacobianInversePrecision,
    ChangeOfVariables,
    LocalVariables
  >);
end function;

LocalVariables:=function(lf)
/*
  Returns a change of coordinates to be used during the lifting process.
*/
  c:=CodimensionLF(lf);
  Q:=lf`ResolutionField;
  return ScalarMatrix(c,Q!1);
end function;

LiftInitialization:= function(lf,i,localvariables)
/*
  Returns a lifting state for the i th factor of lf.
*/
  q:= lf`MinimalPolynomial[i];
  A<a>:= quo<Parent(q)|q>;
  jacobianvalue:=EvaluateJacobian_l(lf,i);
  // TODO: handled the situation where A is not a field.
  // The problem araises in HasLiftingPoint from MoveCoordinates
  d:=Determinant(jacobianvalue);
  if not IsUnit(d) then
    return "Not invertible";
  end if;
  if IsField(A) then
    jacobianinverse:= jacobianvalue^(-1);
  else
    invd:=InverseMod(Parent(q)!d,q);
    MA:=Parent(jacobianvalue);
    jacobianinverse:=MA!invd*Adjoint(jacobianvalue);
  end if;    
  return rec<LiftingState()|
    Algorithm:="$filename$",
    MagicPoint:=lf`MagicPoint,
    PrimitiveElement:=lf`PrimitiveElement,
    Solution:=ChangeUniverse(lf`Parametrization[i],A),
    SolutionPrecision:=1,
    JacobianValue:= jacobianvalue,
    JacobianPrecision:= 1,
    JacobianInverse:= jacobianinverse,
    JacobianInversePrecision:= 1,
    ChangeOfVariables:= lf`ChangeOfVariables,
    LocalVariables:=localvariables
  >;

end function;

ChangeBaseRingLS:=procedure(~ls,R)
/*
  Changes base ring of the lifting state ls to R.
*/
  A<a>:=Universe(ls`Solution);
  q:=Modulus(A);
  UP<T>:=PreimageRing(A);
  UPR<T>:=PolynomialRing(R);
  q:=UPR![R| z: z in Coefficients(q)];
  AR<a>:=quo<UPR|q>;
  ChangeUniverse(~ls`Solution,UP);
  ChangeUniverse(~ls`Solution,UPR);
  ChangeUniverse(~ls`Solution,AR);
  ChangeUniverse(~ls`MagicPoint,R);
  ls`PrimitiveElement:=ChangeRing(Parent(ls`PrimitiveElement),R)!ls`PrimitiveElement;
  MyChangeRing(~ls`JacobianValue,UP);
  MyChangeRing(~ls`JacobianValue,UPR);
  MyChangeRing(~ls`JacobianValue,AR);
  MyChangeRing(~ls`JacobianInverse,UP);
  MyChangeRing(~ls`JacobianInverse,UPR);
  MyChangeRing(~ls`JacobianInverse,AR);
  MyChangeRing(~ls`ChangeOfVariables`LinearPart,R);
  MyChangeRing(~ls`ChangeOfVariables`AffinePart,R);
  MyChangeRing(~ls`LocalVariables,R);
end procedure;

FiberFromLLS:=function(lf,lls)
/*
  Build a fiber from lf and the solutions contained in lls.
*/
  assert #lls gt 0;
  ans:=lf;
  A<a>:= Universe(lls[1]`Solution);
  UP<T>:= PreimageRing(A);
  ans`ResolutionField:=BaseRing(UP);
  ans`MinimalPolynomial:=[Modulus(Universe(Z`Solution)) : Z in lls];
  ans`Denominator:=[UP!1 : Z in lls];
  ans`Parametrization:= [ [UP| UP!z : z in Z`Solution] : Z in lls ];
  ans`MagicPoint:= lls[1]`MagicPoint;
  ans`PrimitiveElement:= lls[1]`PrimitiveElement;
  ans`ChangeOfVariables:= lls[1]`ChangeOfVariables;
  return(ans);
end function;

Lift:=procedure(~lf,~lls,stopcriterion)
  while true do
    lfloc:=FiberFromLLS(lf,lls);
    b, answer:= stopcriterion(lfloc);
    if b then lf:=answer; return; end if;
    requestedprecision:=answer;
    vprint HenselVerbose,2:"Lifting to precision", requestedprecision;
    for l:=1 to #lls do
      vprint HenselVerbose,3:"Dealing with factor",l,"/",#lls,"of degree",Degree(lf`MinimalPolynomial[l]);
      ls:=lls[l];
      vprint HenselVerbose,3:"Updating Jacobian inverse";
      vtime HenselVerbose,3 : UpdateJacobianInverse(~ls,Min(2*ls`JacobianInversePrecision,ls`JacobianPrecision));
      f,J:= EvaluateEquationsAndJacobian(lf,ls,requestedprecision);
      ls`JacobianValue:=J;
      ls`JacobianPrecision:=ls`SolutionPrecision;
      vprint HenselVerbose,3:"Updating solution";
      vtime HenselVerbose,3 : UpdateSolution(~ls,f,requestedprecision);
      vprint HenselVerbose,3:"Globalizing solution";
      vtime HenselVerbose,3 : GlobalizeSolution(~ls);
      lls[l]:=ls;
    end for;
  end while;
end procedure;

Split:=function(lf)
  return [lf];
end function;

/*****************************************************************************
                                Intrinsics
******************************************************************************/

// renames intrinsics




// # 1 "hensel_common.h" 1
// -*-magma-*-
// $Header$
// $Log$

/*****************************************************************************
                                Intrinsics
******************************************************************************/

intrinsic HasLiftingPointLF_newton(lf::Rec)->BoolElt
{
 Returns whether any lifting process on lf will be ok or not.
}
  if IsWholeSpaceLF(lf) then return(true); end if;
  localvariables:=LocalVariables(lf);
  for l:=1 to NumberOfFactorsLF(lf) do
    ls:= LiftInitialization(lf,l,localvariables);
    if Type(ls) ne Rec then return false; end if;
  end for;
  return true;  
end intrinsic;

intrinsic LiftCurveLF_newton(~lf::Rec,destpoint,precision)
{
  See documentation in hensel.mtx.
}
  IndentPush();
  vprint HenselVerbose,1:"--> LiftCurveLF";
  begintime:=Cputime();
  vprint HenselVerbose,2:"Degrees of the factors:",Sprint([Degree(z) : z in lf`MinimalPolynomial]);
  r:=RankLF(lf);
  require #destpoint eq r: "Wrong size of 2nd argument";
  F:=lf`ResolutionField;
  LS:=LaurentSeriesRing(F); AssignNames(~LS,[Sprint(ParentBbpLF(lf).r)]);
  LS`Precision:=1;
  lls:=[];
  localvariables:=LocalVariables(lf);
  for l:=1 to NumberOfFactorsLF(lf) do
    ls:=LiftInitialization(lf,l,localvariables);
    if Type(ls) ne Rec then lf:=ls; IndentPop(); return; end if;
    ChangeBaseRingLS(~ls,LS);
    ls`MagicPoint := [ LS| LS!ls`MagicPoint[j]
      +LS.1*(LS!destpoint[j]-LS!ls`MagicPoint[j]) : j in [1..r]];
    Append(~lls,ls);
  end for;
  Lift(~lf, ~lls, hensel_base__StopCriterion_Bound(precision));
  vprint HenselVerbose,1:"<-- LiftCurveLF, elapsed time:", Cputime()-begintime;
  IndentPop();
end intrinsic;

intrinsic LiftIntegersLF_newton(~lf::Rec)
{
  See documentation in hensel.mtx.
}
  MakeMonicLF(~lf);
  if IsWholeSpaceLF(lf) then return; end if;
  require HasShapeLemmaParametrizationLF(lf): "ShapeLemma parametrization required";
  IndentPush();
  vprint HenselVerbose,1:"--> LiftIntegersLF";
  vprint HenselVerbose,2:"Degrees of the factors:",Sprint([Degree(z) : z in lf`MinimalPolynomial]);
  begintime:=Cputime();
  // initialization
  K:=lf`ResolutionField;
  PA:=pAdicIntegers(#K,1);
  PA,h:=ChangePrecision(PA,1);
  lls:=[];
  localvariables:=LocalVariables(lf);
  for l:=1 to NumberOfFactorsLF(lf) do
    ls:=LiftInitialization(lf,l,localvariables);
    if Type(ls) ne Rec then lf:=ls; IndentPop(); return; end if;
    ChangeBaseRingLS(~ls,PA);
    Append(~lls,ls);
  end for;
  stopcriterion:=hensel_base__StopCriterion_Integers();
  Lift(~lf, ~lls, stopcriterion);
  vprint HenselVerbose,1:"<-- LiftIntegersLF, elapsed time:", Cputime()-begintime;
  IndentPop();
end intrinsic;

intrinsic LiftSplitLF(lf) -> SeqEnum
{
  See documentation in hensel.mtx.
}
  IndentPush();
  vprint HenselVerbose: "--> LiftSplitLF";
  tmp:= Split(lf);
  vprint HenselVerbose: "<-- LiftSplitLF";
  IndentPop();
  return tmp;
end intrinsic;

// # 340 "hensel_newton.mtx" 2

