// # 1 "intersect.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: lun jun 24 17:56:11 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "intersect.mtx" 2
// $Log$

// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 6 "intersect.mtx" 2


PROFILE:=false;
//PROFILE:=true;

//NODE_INTERSECT

// #Package=Kronecker Intersection

// #Public

// This Package implements Kronecker's intersection algorithm. 
// The verbose level flag of this package is IntersectionVerbose.

/***************************************************************************
                       Low Level Functionalities
***************************************************************************/
// #Private

IsCorrectForIntersection:=function(lc)
/*
  lc is a lifted curve.
*/
  if Type(lc`ResolutionField) ne FldFunRat or
    Type(IntegerRing(lc`ResolutionField)) ne RngUPol then
    return false; 
  end if;

  if #lc`MinimalPolynomial ne 1 then
    return false;
  end if;
  q:=lc`MinimalPolynomial[1];
  p:=lc`Denominator[1];
  w:=lc`Parametrization[1];
  n:=#w;
  delta:=Degree(q);

  UPT:=Parent(q);
  RFy:=lc`ResolutionField;
  UPy:=IntegerRing(RFy);
  F:=BaseRing(RFy);
  
  // Last coordinate of the magic point must be zero
  // in order to improve the computations
  for z in lc`MagicPoint do
    if Evaluate(z,0) ne 0 then return false; end if;
  end for;
  
  // check that the resolution is polynomial in y
  //   and compute the degree in y
  dy:=0; // dy will contain the maximum degree in y
  for i:= 0 to delta do
    if Denominator(Coefficient(q,i)) ne UPy!1 then return false; end if;
    if Denominator(Coefficient(p,i)) ne UPy!1 then return false; end if;
    dy:=Max(dy,Degree(Numerator(Coefficient(q,i))));
    for j:= 1 to n do
      if Denominator(Coefficient(w[j],i)) ne UPy!1 then return false; end if;
      dy:=Max(dy,Degree(Numerator(Coefficient(w[j],i))));
    end for;
  end for;

  // check that the degree in y is less than the degree in T
  if dy gt delta then return false; end if;
  return(true);

end function;

IntersectIncrementalInterpolationBasis:=procedure(sy,~base_de_lagrange)
/*
  Input:
    sy, sequence of integers.
    base_de_lagrange, sequence of univariate polynomials.

  Description:
    It is assumed that the given base_de_lagrange is an interpolating basis for
    the sequence Prune(sy) given.
    So base_de_lagrange is just updated in order to be a basis for sy.

  Other:
    This procedure should be called by Intersect only.
    TODO << improve for huge cases.
*/
  UP:=Universe(base_de_lagrange);
  T:=UP.1;
  for i:=1 to #base_de_lagrange do
    base_de_lagrange[i]*:=(T-sy[#sy])/(sy[i]-sy[#sy]);
  end for;
  Append(~base_de_lagrange, &*[UP|T-sy[i] : i in [1..#sy-1]]
    / &*[Universe(sy)|sy[#sy]-sy[i] : i in [1..#sy-1]]);
end procedure;

KroneckerIntersectResultant:=function(lc,g)
/*
  See the documentation of KroneckerIntersect.

  returned messages are:

    INTERSECT_GENERICITY
    INTERSECT_CHARACTERISTIC
    INTERSECT_NOETHER
*/
  assert #lc`MinimalPolynomial eq 1;

  q:=lc`MinimalPolynomial[1];
  d:=TotalDegree(g);
  p:=lc`Denominator[1];
  mp:=lc`MagicPoint;
  w:=lc`Parametrization[1];
  n:=#w;
  delta:=DegreeLF(lc);

  RFy:=lc`ResolutionField;
  UPy:=IntegerRing(RFy);
  F:=BaseRing(RFy);

  //~~~~~~~~~~~~~~~ begining of the computations ~~~~~~~~~~~~~~~

  UPT<T>:=PolynomialAlgebra(F);

  sy:=[F|]; // sequence of interpolating points
  srq:=[F|];         // sequence of the values of the new minimal polynomial
  srw:=[];           // sequence of the values of the new parametrization

  yi:=-1;
  base_de_lagrange:=[UPT|]; // sequence of Lagrange's interpolation basis
  qy:=UPT!0;
  IndentPush();
  vprint IntersectionVerbose,1 : "--> KroneckerIntersectResultant in degrees", delta, "and", d;

  begintime:=Cputime();

  ;
    
  // interpolating loop
  found:=0;
  while 10*#srq le 13*d*delta+5 and yi lt 2*(d*delta+1)+10 do   // 5 and 10 are arbitrary values
    yi+:=1;
    vprint IntersectionVerbose,2 : "Intersecting at point",yi;

    // the intersection can not be performed because
    // of bad genericity in the sub resultant algorithm
    if  yi ge 2*(d*delta+1)+10 then
      ans:= "Not Enough Genericity For The Intersection";
      IndentPop();
      return ans,ans;
    end if;

    if Characteristic(F) gt 0 and yi ge Characteristic(F) then
      ans:= "Bad Characteristic Of The Resolution Field";
      IndentPop();
      return ans,ans;
    end if;

    // TODO << improve with fast multipoint evaluation
    vprint IntersectionVerbose,3 :"Specializing curve...";
    vtime IntersectionVerbose,4 : qyi:= UPT![F| Evaluate(z, yi) : z in Coefficients(q)];
    vtime IntersectionVerbose,4 : pyi:= UPT![F| Evaluate(z, yi) : z in Coefficients(p)];
    vtime IntersectionVerbose,4: wyi:= [UPT| UPT![F| Evaluate(z, yi) : z in Coefficients(zz)]: zz in w];

    tmp,ipyi:= Xgcd(pyi,qyi);

    if Degree(tmp) ne 0 then
      vprint IntersectionVerbose,3 : "WARNING: bad interpolating point";
      continue;
    end if;
    ipyi:=ipyi*(Coefficient(tmp,0))^(-1);

    vprint IntersectionVerbose,3 : "Evaluation of the equation...";
    A<a>:=quo<UPT|qyi>;
    mp:= [A| A!UPT!F!Evaluate(z,yi) : z in lc`MagicPoint];
    point:= mp cat [A!(z*ipyi) : z in wyi];
    vtime IntersectionVerbose,4: val_g:=A!EvaluateLF(lc,g,point);
    val_g:= UPT!val_g;

    vprint IntersectionVerbose,3 : "Resultant...";
    vtime IntersectionVerbose,4 : r:= Resultant(qyi,val_g);

    Append(~sy,F!yi);
    Append(~srq,r);
    
    // Here is a small hack to improve big intersections
    if yi le 30 then
      IntersectIncrementalInterpolationBasis(sy,~base_de_lagrange);
      old_qy:=qy;
      qy:=&+[UPT| UPT!srq[i]*base_de_lagrange[i] : i in [1..#srq]];
      if old_qy eq qy then found +:=1; else found:=0; end if;
    else
      if 10*#srq ge 13*Degree(qy) then
        old_qy:=qy;
        qy:=UPT!Interpolation(sy,srq);
        if old_qy eq qy then found +:=5; else found:=0; end if;	 
      end if;
    end if;
    if Degree(qy) ge delta*d or found ge 5 then break; end if; // 5 is also arbitrary.
  end while;

  if Degree(qy) gt d*delta+5 then IndentPop(); return "Bad Previous Noether Position","Bad Previous Noether Position"; end if;

  vprint IntersectionVerbose,1 : "<-- exit KroneckerIntersectResultant, elapsed time: ",Cputime()-begintime;

  ;
  IndentPop();
  return qy;

end function;

Xgcd_withtrace:=function(q,p)
/*
  This function returns the extended gcd of q and p, that is the gcd
  and the inverse of p. It also returns a sequence of integers called
  trace and corresponding to the degrees of the successive remainders.
  It simply runs Euclide's algorithm >> TODO.

  We assume that deg q > deg p !!!
*/
  UP:=Parent(q);
  u:=q; v:=p; iv1:=UP!0; iv2:=UP!1; 
  du:=Degree(u);
  dv:=Degree(v);
  trace:=[du,dv];
  while v ne 0 do
    tmp:= v; q,v := Quotrem(u,v); u:= tmp; du:= dv; dv:= Degree(v);
    tmp:= iv2; iv2:= iv1-q*iv2; iv1:= tmp;
    Append(~trace,dv);
  end while;
  return u,iv1,trace;
end function;

incrementalcombine:=procedure(~l,q,v,m,trace)
/*
  l is a sequence of tuples, containing a minimal polynomial,
  a parametrization, the multiplicity and the trace corresponding
  to the way it has been obtained.
  If there is an element in l with multiplicity m and trace trace, we
  combine q and v to it. If not we append a new element to l.
*/

  qp:=Derivative(q);
  /* on cherche si l contient un machin avec la meme trace et meme multiplicite */
  for i:=1 to #l do
    if l[i][4] eq trace and l[i][3] eq m then
      Append(~l[i][1],q);
      Append(~l[i][2],v);
      return;
    end if;
  end for;
  Append(~l,<[q],[v],m,trace>);

end procedure;

SubstituteNewParametrizationInOldOne:=function(lc,g,q,p,w,qy,y0,RemoveMultipleComponents);
/*
  We have: q(xr,T)=0, p(xr,T) xi = wi(xr,T) and qy(xr)=0, g(x)=0

  Output: a sequence of parametrizations W such that qy'(xr) xi = Wi(xr),
    together with the corresponding minimal polynomial in y.

  RemoveMultipleComponents is a boolean, when true all the multiple components
  are removed.

  FAILED is returned if the substitution fails.
  
  We assert that y0=0!
*/

  IndentPush();
  vprint IntersectionVerbose,1 : "--> SubstituteNewParametrizationInOldOne";
  assert IsZero(y0);
  assert #lc`MinimalPolynomial eq 1; // Car ca a ete recombine a l'issue de l'intersection !

  begintime:=Cputime();

  ;

  UPy<y>:=Parent(qy);

  olddelta:=Degree(q);

  vprint IntersectionVerbose,3: "Factorizing characteristic polynomial...";
  vtime IntersectionVerbose,4: l_fqy:=Factorization(qy);
  vprint IntersectionVerbose,3: "Degrees of factors: ",[Degree(z[1]):z in l_fqy];
  done:=[];

  i:=1;
  for fqy in l_fqy do
    vprint IntersectionVerbose,2 : "Dealing with factor",i,"/",#l_fqy,"of degree",Degree(fqy[1]);
    i+:=1;
    e:=fqy[2];
    if RemoveMultipleComponents and e gt 1 then continue; end if;

    A<a>:=quo<UPy|fqy[1]>;
    icanA:=hom<A->UPy | y>;

    UPa<T>:=PolynomialAlgebra(A);

    subs_y:= function(P) // UPy -> UPa
      return(UPa![ A!Numerator(z): z in Coefficients(P)]); 
    end function;

    vprint IntersectionVerbose,3: "Computing trace...";
    pa:= subs_y(p);
    qa:= subs_y(q);

    vtime IntersectionVerbose,4: ma,ipa,trace:=Xgcd_withtrace(qa,pa);  // attention  l'ordre des arguments

    qa:=qa div ma;
    B<b>:=quo<UPa|qa>;

    vprint IntersectionVerbose,3: "Parametrization...";
    tmp:=[B| ((subs_y(z) div ma)*ipa) : z in w ];

    vprint IntersectionVerbose,3: "Evaluation...";
    v:=[B| A!Evaluate(z,a) : z in lc`MagicPoint] cat tmp;
    vtime IntersectionVerbose,4: val_g:= UPa!EvaluateLF(lc,g,v);
    
    vprint IntersectionVerbose,3: "Linear factor...";
    vtime IntersectionVerbose,4: linear_form:= Gcd(val_g,qa);

    vprint IntersectionVerbose,3: "Substituting back...";

    if Degree(linear_form) ne 1 then IndentPop(); return("Can not substitute back"); end if;
    Ta:= -Coefficient(linear_form,0)/Coefficient(linear_form,1);
    //TODO << Patterson StockMeyer
    vtime IntersectionVerbose,4: powers_Ta:= [A| i eq 0 select A!1 else Ta*Self(i) : i in [0..olddelta] ];
    vtime IntersectionVerbose,4: v:= [UPy| y mod fqy[1]] cat [UPy | &+[ Coefficient(z,i)*powers_Ta[i+1]: i in [0..olddelta]]: z in tmp];

    incrementalcombine(~done,fqy[1],v,fqy[2],trace);

  end for;

  vprint IntersectionVerbose,1 : "<-- SubstituteNewParametrizationInOldOne, elapsed time: ", Cputime()- begintime;

  ;

  IndentPop();
  return(done);

end function;

/////////////////////////////////
// #Section=Kronecker's Intersection
/////////////////////////////////

intrinsic KroneckerIntersect(~lc::Rec,g:RemoveMultipleComponents:=false)
{
   
      lc, a lifted curve.
      g, a blackbox polynomial.
   
   
     Let y be the parameter of the curve lc, and u its primitive element.
     The procedure computes the fiber intersection
     of the curve C described by lc with g.
     The new primitive element taken for C is y.
     If the parameter RemoveMultipleComponents is set,
     the multiple components are thrown away.
   
   
      lc must be shape lemma.
     
       lc may contain in return:
        -
          "Bad lifted curve" if lc
           does not look like a lifted curve.
          "Bad Characteristic Of The Resolution Field" if the characteristic of
           the ResolutionField is positive and too small to perform
           the interpolation.
          "Not Enough Genericity For The Intersection" if there is not enough
           genericity to perform the subresultant algorithm.
          "Bad Previous Noether Position" if the previous Noether position was
           wrong, the lifted curve is false and thus the intersection fails.
        itemize
      lc`ResolutionField must be a Rational Function Field in one variable
       over a field (lc should come from LiftLastFreeVariableLF).
   
}
  IndentPush();
  vprint IntersectionVerbose,1 : "--> KroneckerIntersect";
  begintime:=Cputime();

  if not IsCorrectForIntersection(lc) then
    IndentPop();
    lc:="Bad lifted curve";
    return;
  end if;

  // TODO << leave it here or not?
  // Compute and remember the gradient of g
  _:=Gradient(g);

  //~~~~~~~~~~~~~~~ Checking Entries ~~~~~~~~~~~~~~~~
  require #lc`MinimalPolynomial eq 1: "lc can not be factorized";
  q:=lc`MinimalPolynomial[1];
  d:=TotalDegree(g);
  
  p:=lc`Denominator[1];
  w:=lc`Parametrization[1];
  n:=#w;
  delta:=Degree(q);

  UPT:=Parent(q);
  RFy:=lc`ResolutionField;
  UPy:=IntegerRing(RFy);
  F:=BaseRing(RFy);

  // we raise an error when the characteristic is too small
  if Characteristic(F) gt 0 and Characteristic(F) le d*delta+1 then
    lc:="Bad Characteristic Of The Resolution Field";
    IndentPop();
    return;
  end if;

  qy:=KroneckerIntersectResultant(lc,g);

  if Type(qy) eq MonStgElt then lc:=qy; IndentPop(); return; end if;
  UP<T>:=Parent(qy);
  if qy eq 0 then
    IndentPop();
    lc:="Bad lifted curve";
    return;
  end if;

  assert Coefficient(Numerator(lc`MagicPoint[#lc`MagicPoint]),1) eq 1;
  y0:=F!Evaluate(lc`MagicPoint[#lc`MagicPoint],F!0);
  assert y0 eq 0;
  
  vprint IntersectionVerbose,2 : "KroneckerIntersect: Deducing the new parametrization";
  newparametrizations:= SubstituteNewParametrizationInOldOne(lc,g,q,p,w,qy,y0,RemoveMultipleComponents);

  if Type(newparametrizations) eq MonStgElt then
    lc:="Bad Previous Noether Position";
    IndentPop();
    return;
  end if;
    
  // updating the fiber
  answer:=[];
  for param in newparametrizations do
    loclc:=lc;
    if assigned loclc`GenericTrace then delete loclc`GenericTrace; end if;
    loclc`ResolutionField := F;
    loclc`MinimalPolynomial := param[1];
    if param[3] gt 1 then loclc`IsMultiple:=true; end if;
    loclc`Denominator := [UP| UP!1 : z in param[1]];
    loclc`Parametrization := param[2];
    // TODO << should not rebuild the following domain.
    MP:=PolynomialAlgebra(F,NumberOfVariablesLF(lc));
    loclc`PrimitiveElement := MP.RankLF(loclc);
    Append(~loclc`LiftingSystem,g);
    f:=function(x) return Evaluate(x,F!0); end function;
    ChangeUniverse(~loclc`MagicPoint,F,f);
    Prune(~loclc`MagicPoint);
    MyChangeRing(~loclc`ChangeOfVariables`LinearPart,F,f);
    MyChangeRing(~loclc`ChangeOfVariables`AffinePart,F,f);
    loclc`Stamp cat:= "![" cat Sprint(g`Address[1]) cat "," cat Sprint(g`Address[2]) cat "]^[" cat Sprint(param[3]) cat "]";
    Append(~answer,loclc);
  end for;
  lc:=answer;
  vprint IntersectionVerbose,1: "<-- KroneckerIntersect, elapsed time",Cputime()-begintime;
  IndentPop();
end intrinsic;

intrinsic KroneckerIntersect(~llc::SeqEnum,g:RemoveMultipleComponents:=false)
{
  It iterates KroneckerIntersect over each element of llc.
}
  answer:=[];
  for i:=1 to #llc do
    lc:=llc[i];
    KroneckerIntersect(~lc,g:RemoveMultipleComponents:=RemoveMultipleComponents);
    if Type(lc) eq MonStgElt then llc:=lc; return; end if;
    answer cat:=lc;
  end for;
  llc:=answer;
end intrinsic;
