// # 1 "mseries.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun  8 14:57:49 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "mseries.mtx" 2
// $Log$

debug:=false;
debug:=true;

forward MultiplicationTable,DerivativeTable;

/**************************************************************
*           Multivariate Power Series Algebra                 *
**************************************************************/

intrinsic MultivariatePowerSeriesAlgebra(R::Rng,S:PrettyPrint:=false, VariableName:="dx") -> AlgAss
{
  Multivariate power series ring over R with support S.
}
  require #S gt 0: "Support can not be empty";
  n:=#S[1];
  AddAttribute(AlgAss,"Support");
  AddAttribute(AlgAss,"Rank");
  AddAttribute(AlgAss,"PrettyPrint");
  AddAttribute(AlgAss,"VariableName");
  AddAttribute(AlgAss,"DerivativeTable");
  AddAttribute(AlgAss,"IndexVariables");
  AddAttribute(AlgAss,"Variables");
  AddAttribute(AlgAss,"One");
  AddAttribute(AlgAss,"IndexOne");

  table:= MultiplicationTable(S);

  A:=AssociativeAlgebra<R, #S| table :Rep:="Sparse">;

  A`Support:=S;
  A`DerivativeTable:= DerivativeTable(S);
  A`Rank:=n;
  A`PrettyPrint:=PrettyPrint;
  A`VariableName:=VariableName;

  vars:=[ ];
  for i:=1 to n do
    e:=[ j eq i select 1 else 0: j in [1..n]];
    Append(~vars,Position(S,e));
  end for;

  A`IndexVariables:=vars;
  A`Variables:=[ A.l : l in vars];
  A`IndexOne:= Position(S,[0 : i in [1..n]]);
  A`One:= A.(A`IndexOne);

  return(A);
end intrinsic;

// Elementary Queries
intrinsic MpsRank(A::AlgAss) -> RngIntElt
{
  Rank of the power series algebra A.
}
  return(A`Rank);
end intrinsic

intrinsic MpsBaseRing(A::AlgAss) -> Rng
{
  Base Ring of A.
}
  return(BaseRing(A));
end intrinsic

intrinsic MpsVariable(A::AlgAss,i) -> AlgAssElt
{
  Return the ith variable of A.
}
  return(A`Variables[i]);
end intrinsic;

intrinsic MpsVariables(A::AlgAss) -> []
{
  Return the sequence of the variables of A.
}
  return(A`Variables);
end intrinsic;

intrinsic MpsIndexVariable(A::AlgAss,i) -> AlgAssElt
{
  Return the index of the ith variable in the support of A.
}
  return(A`IndexVariables[i]);
end intrinsic;

intrinsic MpsIndexVariables(A::AlgAss) -> []
{
  Return the sequence of the index of the variables in the support of A.
}
  return(A`IndexVariables);
end intrinsic;

intrinsic MpsOne(A::AlgAss) -> AlgAssElt
{
  Return A!1.
}
  return(A`One);
end intrinsic;

intrinsic '*:='(~x::AlgAssElt, y::AlgAssElt)
{
  Replace x with x * y.
}
  x:=x*y;
end intrinsic;

intrinsic '+:='(~x::AlgAssElt, y::AlgAssElt)
{
  Replace x with x + y.
}
  x:=x+y;
end intrinsic;

/**************************************************************
*            Support, Gradient, Look up tables                *
**************************************************************/

intrinsic MpsSupportProduct(m) -> .
{
  It returns the elements of [0..m[1]-1] x ... x [0..m[#m]-1].
}
  s:=#m;
  if s eq 0 then return({@ [] @}); end if;
  supp:=$$(Prune(m));
  ans:= {@ [ 0 : i in [1..s]]@};
  for i in [0..m[s]-1] do
    ans join:= {@ e cat [i] : e in supp @};
  end for;
  return(ans);
end intrinsic;

intrinsic MpsSupportTotalDegree(m,n) -> .
{
  It returns all the monomials of degree at most m in n variables
}
  if n eq 0 or m lt 0 then return({@ @}); end if;
  if m eq 0 then return({@ [0 : i in [1..n] ] @}); end if;
  supp0:= $$(m-1,n);
  supp:=supp0;
  for i:=1 to n do
    supp join:= {@ [ e[j]: j in [1..i-1]]
                            cat [e[i]+1]
                            cat [ e[j]: j in [i+1..n]] : e in supp0 @};
  end for;
  return(supp);
end intrinsic;

intrinsic MpsIdealGradient(support) -> .
{
  It returns the gradient of support.
}

  s:=#support;
  if s eq 0 then return(support); end if;
  n:=#support[1];

  ans:={@ @};
  for e in support do
    Include(~ans,e);
    for i:=1 to n do
      f:=e; f[i] +:=1;
      Include(~ans,f);
    end for;
  end for;
  return(ans);
end intrinsic;

MultiplicationTable:=function(supp)
/*
  It returns the multiplication table for supp.
*/
  s:=#supp;
  if s eq 0 then return(supp); end if;
  n:=#supp[1];
  table:=[];
  for e in supp do
    for f in supp do
      g:=[e[j]+f[j] : j in [1..n]];
      l:=Position(supp,g);
      Append(~table,[ j eq l select 1 else 0 : j in [1..s]]);
    end for;
  end for;
  return table;
end function;

DerivativeTable:=function(supp)
/*
  Returns the derivative look up tables.
*/
  N:=#supp;
  assert N gt 0;
  D:=[];
  n:=#supp[1];
  one:= Position(supp,[0 : i in [1..n]]);
  for i:= 1 to n do
    Di:=[];
    for j:=1 to N do
      e:=supp[j];
      if e[i] eq 0 then
        Append(~Di,<one,0>);
      else
        de:= [i eq k select e[k]-1 else e[k] : k in [1..#e]];
        Append(~Di,<Position(supp,de),e[i]>);
      end if;
    end for;
    Append(~D,Di);
  end for;
  return(D);
end function;

/**************************************************************
*                        ChangeRing                           *
**************************************************************/

intrinsic MpsChangeRing(A::AlgAss,R) -> AlgAss,.
{
  It returns B,h: h is the coercion map from A to B.
}
  B:= MultivariatePowerSeriesAlgebra(R,A`Support:
                PrettyPrint:=A`PrettyPrint,
                VariableName:=A`VariableName);
  n:= #B`Support;
  Bzero:= B!0;
  h:= function(x)
        y:=Bzero;
        for i:=1 to n do y:=y+R!x[i]*B.i; end for;
        return(y);
      end function;
  return B,h;
end intrinsic;

intrinsic MpsChangeRing(A::AlgAss,R,f) -> AlgAss,.
{
  It returns B,h: h is the coercion map from A to B.
}
  B:= MultivariatePowerSeriesAlgebra(R,A`Support:
                PrettyPrint:=A`PrettyPrint,
                VariableName:=A`VariableName);
  n:= #B`Support;
  Bzero:= B!0;
  h:= function(x)
        y:=Bzero;
        for i:=1 to n do y:=y+f(x[i])*B.i; end for;
        return(y);
      end function;
  return B,h;
end intrinsic;

intrinsic CoerceInteger(A::AlgAss,x) -> AlgAssElt
{
  Miscellaneous function for embedding integers into
  Mps algebras.
}
  return x*MpsOne(A);
end intrinsic;

intrinsic CoerceInteger(A,x) -> .
{
  Returns A!x.
}
  return A!x;
end intrinsic;

/**************************************************************
*                         MpsElt ops                          *
**************************************************************/

intrinsic MpsCoefficient(f::AlgAssElt,e::SeqEnum)-> .
{
  Coefficient of the monomial of f with exponent e.
}
  A:= Parent(f);
  i:= Position(A`Support,e);
  if i gt 0 then return(f[i]); else return(BaseRing(A)!0); end if;
end intrinsic;

intrinsic MpsConstantCoefficient(f::AlgAssElt)-> .
{
  Constant coefficient of f.
}
  A:= Parent(f);
  return(f[A`IndexOne]);
end intrinsic;

intrinsic MpsDerivative(f::AlgAssElt,i::RngIntElt)-> AlgAssElt
{
  The derivative of f with respect to the i-th variable.
}
  A:= Parent(f);
  require i gt 0 and i le MpsRank(A): "bad second argument";

  support:=A`Support;
  T:= A`DerivativeTable[i];
  ans:= A!0;
  for i:=1 to #support do
    tmp:=T[i];
    ans:= ans+(f[i]*tmp[2])*A.tmp[1];
  end for;
  return(ans);
end intrinsic;

intrinsic MpsDerivative(f::AlgAssElt,i::RngIntElt,k::RngIntElt)-> AlgAssElt
{
  The k-th derivative of f with respect to the i-th variable.
}
  A:= Parent(f);
  require i gt 0 and i le MpsRank(A): "bad second argument";
  ans:=f;
  for j:=1 to k do
    ans:=MpsDerivative(ans,i);
  end for;
  return(ans);
end intrinsic;

intrinsic MpsDerivative(f::AlgAssElt,m::SeqEnum)-> AlgAssElt
{
  d^(&+m) f/ d x1^m1...xn^mn, where n=Rank(Parent(f)).
}
  require #m eq MpsRank(Parent(f)): "bad second argument";
  ans:=f;
  for i:=1 to #m do
    for j:=1 to m[i] do
      ans:=MpsDerivative(ans,i);
    end for;
  end for;
  return(ans);
end intrinsic;

intrinsic MpsCoefficient(f::AlgAssElt,v::SeqEnum,e::SeqEnum)-> .
{
  Coefficients of f the variables in v with exponent e.
}
  A:= Parent(f);
  n:= MpsRank(A);
  require #e eq #v: "sizes of the sequences must be equal";
  support:= A`Support;
  ans:= A!0;
  for j:=1 to #support do
    z:= support[j];
    if [ z[k] : k in v] eq e then
      z:=  [ i in v select 0 else z[i] : i in [1..n]];
      ans:= ans + f[j]*A.Position(support,z);
    end if;
  end for;
  return(ans);
end intrinsic;

intrinsic MpsGradient(f::AlgAssElt)-> []
{
  Sequence of the partial derivatives of f.
}
  return([ MpsDerivative(f,i) : i in [1..MpsRank(Parent(f))]]);
end intrinsic;
 
intrinsic MpsPrint(f::AlgAssElt:NewLine:=true)
{
  Print f.
}
  A:=Parent(f);
  support:=A`Support;
  PP:=A`PrettyPrint;
  X:=A`VariableName;
  if f eq 0 then
    printf "0";
    if NewLine then printf "\n"; end if;
    return;
  end if;
  first:=true;
  for i:=1 to #support do
    e:=support[i];
    if f[i] ne 0 then
      if PP then
        first1:=true;
        fs:=Sprint(f[i]);
        if fs[1] ne "-" and not first then
          printf " + ";
        end if;
      if fs[1] eq "-" then
        printf " - ";
        c:=-f[i];
      else
        c:=f[i];
      end if;
      if not IsOne(c) or (IsOne(c) and &+e eq 0) then
        printf "(%o)",c;
        first1:=false;
      end if;

        for j:=1 to #e do
          if e[j] gt 0 then
            if not first1 then printf "*"; else first1:=false; end if;
            if e[j] gt 1   then printf "%o%o^%o",X,j,e[j];
            elif e[j] eq 1 then printf "%o%o",X,j;
            end if;
          end if;
        end for;
      else
          if not first then printf " + "; else first:=false; end if;
        printf "%o*%o",e,f[i];
      end if;
    first:=false;
    end if;
  end for;
  if NewLine then printf "\n"; end if;
end intrinsic;

intrinsic MpsPrint(f::SeqEnum: NewLine:=true)
{
  Print f.
}
  first:=true;
  printf "[ ";
  IndentPush();
  for z in f do
    if not first then printf ", "; else first:=false; end if;
    MpsPrint(z: NewLine:=false);
  end for;
  IndentPop();
  printf "]";
  if NewLine then printf "\n"; end if;
end intrinsic;
 
