// # 1 "posdim.mtx"
// $Header$
// # 1 "header.h" 1
/*
///////////////////////////////////////////////////

   Kronecker Package version 0.166-9 for Magma

///////////////////////////////////////////////////

Date: sam jun 22 18:44:15 UTC 2002
Copyright (C) 1999--2002, G. Lecerf, Centre National de la Recherche Scientifique
Location: 
          Laboratoire de mathmatiques, LAMA - UMR 8100
          Universit de Versailles St-Quentin-en-Yvelines
          45 avenue des Etats-Unis - Btiment Fermat
          78035 Versailles, France
Url: http://kronecker.medicis.polytechnique.fr/
Bug report to: Gregoire.Lecerf@math.uvsq.fr

!!! WARNING, this package comes with NO WARRANTY !!!
*/

// # 3 "posdim.mtx" 2
// $Log$
// # 1 "macros.h" 1
// $Header$
// $Log$


















declare verbose BbpVerbose,5;
GlobalBbpVerbose:=[0,0,0,0,0,5];
declare verbose FdaVerbose,5;
GlobalFdaVerboseVerbose:=[0,0,0,1,2,3];
declare verbose FiberVerbose,5;
GlobalFiberVerbose:=[0,1,2,3,4,5];
declare verbose GeometricSolveVerbose,5;
GlobalGeometricSolveVerbose:=[0,1,2,3,4,5];
declare verbose HenselVerbose,5;
GlobalHenselVerboseVerbose:=[0,1,2,3,3,4];
declare verbose IntersectionVerbose,5;
GlobalIntersectionVerbose:=[0,1,2,2,3,4];
declare verbose NonUniversalVerbose,5;
GlobalNonUniversalVerbose:=[0,1,2,3,4,5];
declare verbose AlgebraicApproximantVerbose,5;
GlobalAlgebraicApproximantVerbose:=[0,0,0,1,2,3,5];




















// MathML support

declare verbose KroneckerMathMLVerbose,5;









// # 5 "posdim.mtx" 2
import "hensel_base.m": hensel_base__FromSeriesToRationalCoefficients, hensel_base__FromSeriesToRationalCoefficients_Pade, hensel_base__EvaluateAtOne, hensel_base__MultiplyByCommonDenominator, hensel_base__StopCriterion_Bound, hensel_base__StopCriterion_Integers, hensel_base__HeuristicCombineFactorized;// #import hensel_base.m
import "random.m": random_RandomInvertibleMatrix, random_RandomInvertibleUpperTriangularMatrix, random_RandomVector, random_RandomPrimeNumber, random_RandomSequence;// #import random.m



/*
$Header$

Module: lift all free variables

History:

 o Originaly, this file was written by Eric Schost.
   (see \cite{Schost2000} at http://tera.medicis.polytechnique.fr)
   See also Schost's Phd thesis also available at tera.

 o Integrated into Kronecker 0.16 since 21/01/2000.

--- CVS management ---
$Log$

*/

PrimeSize:=30; // Size of the prime used for modular lifting



// #Public

// #Package=Lifting the Free Variables

// #Section=Introduction
/////////////////////

//  This part provides functionalities to write in an expanded form
//  the parametrization of a geometric
//  resolution of a positive dimensional variety.
//  
//  This section is mainly due to Eric Schost Schost2000.

// #Private
Recursive_LiftAllFreeVariables:=function(lf,point,MP)

  r:=RankLF(lf);
  s:=#point;
  delta:=DegreeLF(lf);
  n:=NumberOfVariablesLF(lf);

  if s eq r-1 then
    direction:=point cat [1];
    loclf:=lf;
    hensel_base__HeuristicCombineFactorized(~lf);
    LiftCurveLF(~loclf,direction, DegreeLF(lf)+1);
    if Type(loclf) eq MonStgElt then return "Failed" ; end if;
    MakeKroneckerParametrizationLF(~loclf);
    CombineLF(~loclf:Parametrization:="Kronecker");
    assert #loclf`MinimalPolynomial eq 1;
    return( [
         [ MP| &+[Coefficient(Coefficient(loclf`MinimalPolynomial[1],i),j)*(MP.r)^j : j in [0..delta]] : i in [0..delta] ]
      ]
      cat
      [
         [MP| &+[Coefficient(Coefficient(w,i),j)*(MP.r)^j:  j in [0..delta]]
         : i in [0..delta] ]
      : w in loclf`Parametrization[1]
      ]);
  end if;

  ans:=[];
  for i:=0 to delta do
    Append(~ans,$$(lf,point cat [i],MP));
  end for;

  return([
    [MP| Interpolation([BaseRing(MP)| k : k in [0..delta]],
                       [ans[i][l][k] : i in [1..delta+1]],
                       s+1)
       : k in [1..delta+1] ]

     : l in [1..n-r+1] ]);

end function;

//////////////////////////////// INTRINSICS //////////////////////////

// #Section=Lifting
////////////////

LiftAllFreeVariablesLF_generic:=procedure(~lf)
/*
// It lifts all the free variables of lf.
*/
  IndentPush();
  vprint HenselVerbose: "--> LiftAllFreeVariablesLF_generic";

  F:=lf`ResolutionField;
  r:=RankLF(lf);
  if r eq 0 then IndentPop(); return; end if;

  TranslateAllFreeVariablesToZeroLF(~lf);

  MP:=PolynomialAlgebra(F,r);

  E:=ParentBbpLF(lf);
  ans:=Recursive_LiftAllFreeVariables(lf,[],MP);
  MF:=FieldOfFractions(MP);
  h:=hom<MP->MF | x:->Evaluate(x,[MF.i/MF.r: i in [1..r-1]] cat [MF.r])>;

  UP<T>:=PolynomialAlgebra(MF);
  lf`ResolutionField:=MF;
  lf`MinimalPolynomial:=[UP![ h(z): z in ans[1]]];
  lf`Denominator:=[Derivative(lf`MinimalPolynomial[1])];
  lf`Parametrization:=[[ UP![MF | h(z) : z in ans[i]]: i in [2..#ans]]];
  lf`MagicPoint:=[MF | MF!lf`MagicPoint[i]+MF.i : i in [1..r]];
  vprint HenselVerbose: "<-- exit LiftAllFreeVariablesLF_generic";
  IndentPop();
end procedure;


// #Private
RecursiveChineseRemainderTheorem:=function(A,B,p,q)
/*
  This is a chinese remainder for the structures we use.
*/

  OUT:=Integers(p*q);
  if Type(A) eq RngIntElt and Type(B) eq RngIntElt then 
    return(OUT!ChineseRemainderTheorem([A,B],[p,q]));

  elif Type(A) eq RngIntResElt and Type(B) eq RngIntResElt then 
    return(OUT!$$(Integers()!A,Integers()!B,p,q));

  elif Type(A) eq SeqEnum and Type(B) eq SeqEnum and #A eq #B then
    if #A eq 0 then return [Universe(A)|]; end if;
    U:=Parent($$(A[1],B[1],p,q));
    tmp:=[U| U!$$(A[i],B[i],p,q): i in [1..#A]];
    return(tmp );

  elif Type(A) eq RngUPolElt and Type(B) eq RngUPolElt then
    Ka:=BaseRing(Parent(A));
    Kb:=BaseRing(Parent(B));
    d:=Maximum(Degree(A),Degree(B));
    OUT:=Parent($$(Ka!0,Kb!0,p,q));
    ans:=$$([Ka|Coefficient(A,i): i in [0..d]],[Kb|Coefficient(B,i): i in [0..d]],p,q);
    UP:=PolynomialAlgebra(OUT);
    return(UP!ans);

  elif Type(A) eq RngMPolElt and Type(B) eq RngMPolElt  then

    MPint:=ChangeRing(Parent(A),Integers());
    MPint:=ChangeRing(Parent(B),Integers());
    intA:=MPint!A; intB:=MPint!B;
    monomes:={z : z in Monomials(intA)} join { z: z in Monomials(intB)};
    MPout:=ChangeRing(MPint,OUT);
    tmp:=&+[MPout| $$(MonomialCoefficient(intA,x),MonomialCoefficient(intB,x),p,q)*MPout!x: x in monomes];
    return(tmp);

  elif Type(A) eq FldFunRatMElt and Type(B) eq FldFunRatMElt  then
    num:=$$(Numerator(A),Numerator(B),p,q);
    denom:=$$(Denominator(A),Denominator(B),p,q);
    MFout:=FieldOfFractions(Parent(num));
    return(MFout!num/MFout!denom);
  else print(Type(A)); print(Type(B)); error "Unexpected types";
  end if;
end function;

CombineLF_ChineseRemainder:=procedure(~intlf,modlf)
/*
  It returns the chinese remainder of intlf and modlf.
*/
  MF:=modlf`ResolutionField;
  F:=BaseRing(MF);
  q:=#F;

  if Type(intlf) eq MonStgElt then intlf:=modlf; return; end if;

  p:=#BaseRing(intlf`ResolutionField);  
  intlf`MinimalPolynomial:=RecursiveChineseRemainderTheorem(intlf`MinimalPolynomial,modlf`MinimalPolynomial,p,q);
  intlf`Denominator:=RecursiveChineseRemainderTheorem(intlf`Denominator,modlf`Denominator,p,q);
  intlf`Parametrization:=RecursiveChineseRemainderTheorem(intlf`Parametrization,modlf`Parametrization,p,q);
  intlf`ResolutionField:=BaseRing(Universe(intlf`MinimalPolynomial));

  // Et le reste... >> TODO
end procedure;

LiftAllFreeVariablesLF_Integers:=procedure(~lf)
/*
// It lifts all the free variables when then resolution field of lf
// is the rational numbers field.
*/
  IndentPush();
  vprint HenselVerbose: "--> LiftAllFreeVariablesLF_Integers";

  // remontee des variables libres en faisant du reste chinois
  intlf:="begin";

  p:=NextPrime(10^PrimeSize);
  while true do
    p:=NextPrime(p);
    vprint HenselVerbose:  "LiftAllFreeVariablesAndIntegersLF, prime", p;
    modlf:=lf;
    ChangeResolutionFieldLF(~modlf,GF(p));
    MakeShapeLemmaParametrizationLF(~modlf);
    if Type(modlf) eq MonStgElt then continue; end if;
    LiftAllFreeVariablesLF_generic(~modlf);
    MF:=modlf`ResolutionField;
    MP:=IntegerRing(MF);
    h:=map<MF->MP| x:-> Numerator(x)>;
    ChangeResolutionFieldLF(~modlf,MP,h);
    CombineLF_ChineseRemainder(~intlf,modlf);

    b, ratlf:= RationalReconstructionLF(intlf);
    if not b then continue; end if;
    MP:=ratlf`ResolutionField;
    MF:=FieldOfFractions(MP);
    h:=Bang(MP,MF);
    ChangeResolutionFieldLF(~ratlf,MF,h);
    /// TODO: BUG ici au 4/5/2002 avec f744.mgm
    if VerifyLF(ratlf) then break; end if;
  end while;

  lf:=ratlf;

  vprint HenselVerbose: "<-- LiftAllFreeVariablesLF_Integers";
  IndentPop();
end procedure;

// #List
intrinsic LiftAllFreeVariablesLF(~lf::Rec)
{
 It lifts all the free variables of lf.
}
  IndentPush();
  vprint HenselVerbose: "--> LiftAllFreeVariablesLF";
  if DimensionLF(lf) eq 0 then IndentPop(); return; end if;
  if Type(lf`ResolutionField) eq FldRat then
    LiftAllFreeVariablesLF_Integers(~lf);
  else
    LiftAllFreeVariablesLF_generic(~lf);
  end if;
  vprint HenselVerbose: "<-- LiftAllFreeVariablesLF";
  IndentPop();
end intrinsic;

intrinsic LiftAllFreeVariablesLF(~llf::SeqEnum)
{
 It lifts all the free variables of all the elements of llf.
}
  if IsEmpty(llf) then return; end if;
  answer:=[];
  for i:=1 to #llf do
    lf:=llf[i];
    LiftAllFreeVariablesLF(~lf);
    if Type(lf) eq MonStgElt then llf:=lf; return; end if;
    Append(~answer,lf);
  end for;
  llf:=answer;
end intrinsic;

// #EndList
